/* {{{ preamble
 * hivescan.c
 *
 * Hive scanner
 *
 * Author: Sean Loaring
 *
 * Copyright (c) 2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */
// {{{ includes
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "../lib/ntreg.h"
#include "../lib/libhive.h" //win32-style registry interface
#include "../hivetools/nstdreg.h" //my attempt at a posix-like interface to libhive
#include "../hivetools/sam.h" // sam-specific functions (set user password, etc)
#include "../misc/tools.h" //
#include "../misc/uerr.h" // like errno and strerror...kindof
// }}}

#define TMODE_USAGE 0
#define TMODE_NK    1
#define TMODE_HBIN  2

static int mode = TMODE_USAGE;

// {{{ forward defs
int usage(); 
int parse_args( int *argc, char ***argv ); 
int load_hives( int argc, char **argv );
int list_software();
int list_users();
int list_users_long();
int run_nk_scan(); 
// }}}
int main(int argc, char **argv ) // {{{
{
	//rl_hkey hkey;

	//init
        unicode_init();
        libhive_init();
        ns_init();

	//operate
	if( 0 != parse_args( &argc, &argv ) ) { //sets mode
		fprintf(stderr, "Error parsing arguments: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	if( 0 != load_hives( argc, argv ) ) { 
		fprintf(stderr, "Error loading hives: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	switch( mode ) {
		case TMODE_HBIN:
			WARN("Not running hbin scan...");
			break;
		case TMODE_NK:
			run_nk_scan();
			break;
		case TMODE_USAGE:
			usage();
			break;
		default:
			DIE("Unknown mode.");
	}
	//shutdown

	return 0;

MAINERR:
	return 1;
} // }}}
int parse_args( int *argc, char ***argv ) // {{{ parse leading arguments
{
        extern int optind;
        extern char* optarg;
        char c;

        char *options = "hlLsD:";

        while((c=getopt(*argc,*argv,options)) > 0) {
                switch(c) {
                        case 'h': mode = TMODE_USAGE;      break;
                        case 's': mode = TMODE_NK;      break;
			case 'D':
			{
				int dbval;
				if( 1 != sscanf(optarg, "%d", &dbval ) ) {
					fprintf(stderr, "Couldn't set debug level to \"%s\", is it an integer?\n", optarg );
					exit(1);
				}
				DEBUG_LEVEL = dbval;
				break;
			}
                        default: 
				  fprintf(stderr, "Unrecognized option %c.  Run with -h for help.\n", c);
				  exit(1);
                }
        }

	*argc -= optind;
	*argv += optind;

	return 0;
} // }}}
int usage() // {{{
{
	fprintf(stderr,
		"Usage: Don't\n"
		"Scans hives (more to test my assumptions then to check anything).\n"
		"\n"
		"options:\n"
                " -s          nk scan\n"
                " -h          This message\n"
        );

	fprintf(stderr, "%s", hl_mount_help() );

	/*
	fprintf(stderr, 
                "See readme file on how to extract/read/write the NT's SAM file\n"
                "if it's on an NTFS partition!\n");
		*/

	return 0;
} // }}}
int load_hives( int argc, char **argv ) // {{{ 
{
//        if( !argv[optind] ) {
//                fprintf(stderr, "No hive files specified.\n");
//                return 1;
//        }

	//HERE
        if( 0 != hl_mount_list( argc, argv  ) ) {
                fprintf(stderr, "There was a problem loading the given hives: %s\n", uerr_str( uerr_get() ) );
                return 1;
	}

	return 0;
} // }}}
int list_software() // {{{
{
	NREGDIR* sh;
        struct ns_dirent *d;
        char *path = "/HKEY_LOCAL_MACHINE/software";

        if( NULL == ( sh = ns_opendir( path) ) ) {
		fprintf(stderr, "Error opening key: %s, %s\n", path, uerr_str( uerr_get() ) );
                return -1;
        }

        while( ( d = ns_readdir( sh ) ) ) {
                printf("%s\n", d->filename);
        }

        ns_closedir( sh );

	return 0;
} // }}}
int list_users() // {{{
{
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int *ridlist;
        int i;

        if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error reading user RID list: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

        ridlist = (int*)vb->b;

        for( i=0; i<vb->dlen; i++ ) {
                if( 0 == sam_get_user_name( nbuf, ridlist[i] ) ) {
                        printf("%08x %s\n", ridlist[i], nbuf->b );
                }
                else {
                        fprintf(stderr, "Error reading data for RID %x: %s\n", ridlist[i], uerr_str( uerr_get() ) );
                }
        }

	return 0;
} // }}}
int list_users_long() // {{{
{

	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int *ridlist;
        int i;

        if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error reading user RID list: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

        ridlist = (int*)vb->b;

	printf("Excessivly verbose user list:\n");
        for( i=0; i<vb->dlen; i++ ) {
		//TODO: return value
		nbuf->dlen = 0;
		if( 0 != sam_print_user( nbuf, ridlist[i] ) ) {
			fprintf(stderr, "Error printing user data for RID %x: %s\n", ridlist[i], uerr_str( uerr_get() ) );
			return -1;
		}
		printf("------------------------------------------------\n%s", nbuf->b );
        }

	return 0;
} // }}}

int run_nk_scan() // {{{
{
	int i;
	for( i=0; i<hl_mount_count(); i++ ) {
		rl_hkey hkey = hl_mount_get_hkey( i );
		nk_scan( hkey->hd, get_root_key( hkey->hd ) );
	}
} // }}}

// vim600: set foldlevel=0 foldmethod=marker:
