/* {{{ preamble
 * hiveshell.c
 *
 * Interactive registry access
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *		Jeremy Allison (from pwdump).
 *		Jonathan Wilkins (from NTCrack)
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * TODO: Jeremy Allison (pwdump), Jonathan Wilkins (NTCrack) <- their stuff is
 * probably not in this file anymore, they should be mentioned elsewhere
 *
 }}} */

// {{{ Includes & defines
#include <stdio.h>
#include <wchar.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <iconv.h>
#include <errno.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <signal.h>

#include <openssl/des.h>
#include <openssl/md4.h>
#define uchar u_char
#define MD4Init MD4_Init
#define MD4Update MD4_Update
#define MD4Final MD4_Final

#define CRAP_ERROR printf("FIXME: This is a meaningless error.");

//#define CRAP_ERROR DIE("This is a meaningless error.");

#include "../misc/tools.h"
#include "../hivetools/hivetools.h"
#include "../lib/regtypes.h"
#include "../hivetools/sam.h"
#include "../hivetools/nstdreg.h"
#include "../misc/globals.h"
#include "../lib/ntreg.h"

//command flags
#define CMD_VERBOSE 1
#define CMD_RECURSE 2

//prototypes
void find_n_change(char *username); 
//void print_user( FILE *fp, int rid ); 
long get_realloc( rl_hkey hk, char *path, long *type, char **buf, long *blen ); 
int user_get_password( char *buf, int len ); 
void editor(void);
//}}}

//const char chntpw_version[] = "chntpw version 0.99.2 040105, (c) Petter N Hagen";

int syskeyreset = 0;

/*
 * of user with RID 500, because silly MS decided
 * to localize the bloody admin-username!! AAAGHH!
 */
char admuser[129]="Administrator";

int done=0; //set to 1 to quit program

int change_pw(int rid, int stat) // {{{ TODO: Clean up
{
	/*
	 * rid - the users RID, required for the DES decrypt stage
	 * Some of this is ripped & modified from pwdump by Jeremy Allison
	 */

	int pl;
	char newp[20];
	struct uv *userv;
	struct uf *userf;

	VBUF_STATIC( uname, 10 );
	VBUF_STATIC( uvbuf, 10 );
	VBUF_STATIC( ufbuf, 10 );
	VBUF_STATIC( vb, 100 );

	if( 0 != sam_get_user_name( uname, rid ) ) return -1;

	if( NULL == ( userv = sam_get_user_V( rid, uvbuf ) ) ) return -1;
	if( NULL == ( userf = sam_get_user_F( rid, ufbuf ) ) ) return -1;

	if (stat) {
		//acb = handle_F(rid,0);
		printf("RID: %04x, Username: <%s>%s\n",
		rid, uname->b,  (  userf->acb & 0x8000 ? ", *disabled or locked*" : (userv->head.e[UV_NTPW].len < 16) ? ", *BLANK password*" : "")  );
		return 0;
	}

	if( 1 || gverbose ) {
		printf("Current passwords:\n");
		vbuf_strset(vb, "" );
		pass_print( vb, rid, userv );
		printf("%s\n", vb->b );
	}
		
	if( 0 >= ( pl = user_get_password( newp, 16 ) ) ) return -1;

	if (pl == 1 && *newp == '@') { //TODO: should this be here?
		//promote_user(rid);
		DIE("I commented out the above line because the function does not compile at the moment.  ");
		printf("Changes saved!\n");
		return -1;
	}

	if( pl > 0 ) {
		if( 0 == strcmp( newp, "*" ) ) {
			printf("Blanking password!\n");
			sam_set_nt_password( rid, "" );
			sam_set_lm_password( rid, "" );
		}
		else {
			sam_set_nt_password( rid, newp );
			sam_set_lm_password( rid, newp );
		}

		if( 1 || gverbose ) { //TODO: is this necessary
			if( NULL == ( userv = sam_get_user_V( rid, uvbuf ) ) ) return -1;
			if( NULL == ( userf = sam_get_user_F( rid, ufbuf ) ) ) return -1;
			printf("New passwords:\n");
			vbuf_strset(vb, "" );
			pass_print( vb, rid, userv);
			printf("%s\n", vb->b );
		}
	} else {
		printf("Nothing entered, not setting passwords.\n");
	}

	printf("\n");
	return 0;
} // }}}

void usage(void) // {{{
{ 
	fprintf(stderr, "hiveshell: change password of a user in a NT SAM file, or invoke registry editor.\n"
		"hiveshell [OPTIONS] <samfile> [systemfile] [securityfile] [otherreghive] [...]\n"
		" -h          This message\n"
		//" -u <user>   Username to change, Administrator is default\n"
		//" -i          Interactive. List users (as -l) then ask for username to change\n"
		//" -e          Registry editor. Now with full write support!\n"
		//" -d          Enter buffer debugger instead (hex editor), \n"
		" -v          Be a little more verbose (for debuging)\n"
		//" -L          Write names of changed files to /tmp/changed\n"
		//" -N          No allocation mode. Only (old style) same length overwrites possible\n"
		"See readme file on how to extract/read/write the NT's SAM file\n"
		"if it's on an NTFS partition!\n"
		"Source/binary freely distributable. See README/COPYING for details\n"
		"NOTE: This program is somewhat hackish! You are on your own!\n"
	  );

	fprintf(stderr, "%s", hl_mount_help() );	
} // }}}
void doexit( void ) // {{{
{
	//fprintf(stderr, "Meow!");
} // }}}

int main(int argc, char **argv) // {{{
{
	//int i=0;
	/* int vofs; */
	//int dodebug = 0, inter = 0,edit = 0,il,d = 0, dd = 0, logchange = 0, noalloc = 0, export = 0;
	extern int /* opterr, */ optind;
	extern char* optarg;
	//char *filename,c;
	char c;
	char *who = "Administrator";
	char *regpath;
	char iwho[100];
	FILE *ch;     /* Write out names of touched files to this */

	char *options = "LNhvurp:";

	libhive_init();
	ns_init();
	unicode_init();

	while((c=getopt(argc,argv,options)) > 0) {
		switch(c) {
			case 'v': gverbose = 1; break;
			case 'h': usage(); exit(0); break;
			default: usage(); exit(1); break;
		}
	}

	if( 0 != hl_mount_list( argc - optind, argv + optind ) ) {
		fprintf(stderr, "hl_mount_list() had issues (%s),  I am going to continue...\n",
				uerr_str( uerr_get() ));
	}

	atexit( &doexit );
	editor();

	hl_close_all();

	return 0;

} // }}}
// {{{ Command Vector
struct cmds {
	char str[12];
	char parms[100];
	char desc[100];
	char help[300];
	void (*cmd)(FILE *fp, int argc, char **argv);
};

void cmd_dquit( FILE *fp, int argc, char **argv );
void cmd_rm( FILE *fp, int argc, char **argv );
void cmd_cp( FILE *fp, int argc, char **argv );
void cmd_quit( FILE *fp, int argc, char **argv );
void cmd_undef( FILE *fp, int argc, char **argv );
void cmd_cat( FILE *fp, int argc, char **argv );
void cmd_hd( FILE *fp, int argc, char **argv );
void cmd_find( FILE *fp, int argc, char **argv ); 
void cmd_nv( FILE *fp, int argc, char **argv );
void cmd_mkdir( FILE *fp, int argc, char **argv ); 
void cmd_lsusers( FILE *fp, int argc, char **argv );
void cmd_userinfo( FILE *fp, int argc, char **argv );
void cmd_samstuff( FILE *fp, int argc, char **argv );
void cmd_passwd( FILE *fp, int argc, char **argv );

void cmd_ls_hives( FILE *fp, int argc, char **argv ); 
void cmd_scan_hive( FILE *fp, int argc, char **argv ); 
void cmd_hive_alloc_graph( FILE *fp, int argc, char **argv );
void cmd_hive_dump_unallocd( FILE *fp, int argc, char **argv ); 

void cmd_cd( FILE *fp, int argc, char **argv );
void cmd_ls( FILE *fp, int argc, char **argv );
void cmd_recurse( FILE *fp, int argc, char **argv );

void cmd_help( FILE *fp, int argc, char **argv );
struct cmds maincmds[] = {

	{ "SEP", "Simple registry editor commands", "", "", NULL },
	{ "cd" , "key", "change current working key",
		"" "e.g. cd HKEY_LOCAL_MACHINE/software/Microsoft/\"Windows NT\"/\n" , &cmd_cd
	},
	{ "ls" , "<path>", "list conents of key", "", &cmd_ls },
	{ "find", "<path>", "recursive ls (for now)", "note: this command is rather broken at the moment.", &cmd_find },
	{ "cat", "value_name", "output value data", "", &cmd_cat },
	{ "recurse", "<path>", "testing recursion", "", &cmd_recurse },
	{ "rm", "path", "delete values or recursively delete keys", "", &cmd_rm },
	{ "cp", "src dest", "recursive copy", "", &cmd_cp },
	{ "hd", "value_name", "hex dump value data", "", &cmd_hd },
	{ "nv",
		"value_name value_format [value_type] value_data",
		"set a value",
		""
			"By setting the optional value_type parameter you can force a differnt value type.\n"
			"e.g.\n"
			"   nv ValName1 DWORD 00FF00FF\n"
			"   nv \"Val Name 2\" SZ \"Some string data with an escaped \\\\ backslash.\"\n"
			"   nv ValName3 DWORD BINARY 00,00,ae,00\n",
		&cmd_nv
	} ,
	{ "nk", "key", "create a new key", "", &cmd_mkdir },

	{ "SEP", "SAM", "", "", NULL },
	{ "lsusers", "", "list users", "", &cmd_lsusers },
	{ "userinfo", "username", "detailed user information", "", &cmd_userinfo },
	{ "samstuff", "", "Load some data from SAM and dump it (for testing things)", "", &cmd_samstuff },
	{ "passwd", "username", "Set a user's password", "", &cmd_passwd },

	{ "SEP", "DEBUG", "", "", NULL },
	{ "lshives", "", "list loaded hives.", "", &cmd_ls_hives },

	{ "uadump","hive_no. [file]", "dump unallocated blocks", "", &cmd_hive_dump_unallocd },
	{ "allocgraph","hive_no. [file]", "dump allocation graph", "", &cmd_hive_alloc_graph },
	{ "scanhive", "hive_no.", "hbin level hive scan.", "", &cmd_scan_hive },

	{ "SEP", "Misc", "", "", NULL },
	{ "help", "<cmd>", "display help", "", &cmd_help },
	{ "q", "", "quit the program", "", &cmd_quit } ,
	{ "dquit", "", "call dmalloc_shutdown and crash", "", &cmd_dquit } ,
	{ "", "", "", "", &cmd_undef }
};


void cmd_dquit( FILE *fp, int argc, char **argv ) // {{{
{
	VBUF_STATIC_FREE();
#ifdef WITH_DMALLOC
	dmalloc_shutdown();
#endif
	DIE("dquit is shooting self in the head.");
} // }}}

void cmd_undef( FILE *fp, int argc, char **argv ) // {{{
{
	printf("Sorry, this command (%s) is undefined at this time.\n", argv[0]);
} // }}}
void cmd_help( FILE *fp, int argc, char **argv ) // {{{
{
	struct cmds *c = maincmds;
	char pstr[] = "  %-10s %-30s %-10s\n";
	if( argc == 2 ) {
		for( c = maincmds; c->str[0] != '\0'; c++ ) {
			if( 0 == strcmp( c->str, argv[1] ) ) {
				printf("Command: %s\n", argv[1] );

				printf("Takes parameters: %s\n\n", c->parms );

				printf("%s\n\n", c->desc );
				printf(c->help );
				printf("\n");
				return;
			}
		}
		printf("Nothing found for command %s\n", argv[1] );
		return;
	}
	printf("Simple registry editor help:\n");
	printf(pstr, "Command", "Parameters", "Description" );
	for( c = maincmds; c->str[0] != '\0'; c++ ) {
		if( 0 == strcmp( c->str, "SEP") ) { //this is gross
			printf("--------------------------------------------------------\n");
			printf("%s:\n", c->parms);
		}
		else
			printf(pstr, c->str, c->parms, c->desc );
	}
} // }}}
// }}}
// {{{ Command functions
void cmd_cd( FILE *fp, int argc, char **argv ) // {{{
{
	NREGDIR *n;
	char *path = "";

	if( argc > 1 ) path = argv[1];

	if( 0 != ns_chdir( path ) ) {
		fprintf(stderr, "cd: %s\n", uerr_str( uerr_get() ) );
	}
	
	return;
} // }}}
void cmd_ls( FILE *fp, int argc, char **argv ) // {{{
{
	NREGDIR* sh;
	struct ns_dirent *d;
	char *path = "";

	if( argc > 1 ) path = argv[1];

	if( NULL == ( sh = ns_opendir( path ) ) ) {
		fprintf(stderr, "cmd_ls(): %s\n", uerr_str(uerr_get()) );
		return;
	}

	while( ( d = ns_readdir( sh ) ) ) {
		printf("%s\n", d->filename);
	}

	ns_closedir( sh );
} // }}}
int prt_bleh( char *base, char *rel, long type, void *opt ) // {{{
{
	printf("b == %s, r = %s type == %d\n", base, rel, type );
	return 0;
} // }}}
void cmd_recurse( FILE *fp, int argc, char **argv ) // {{{
{
	char *path = "";

	if( argc > 1 ) path = argv[1];

#ifdef WITH_DMALLOC
	dmalloc_message("Pre-fulldir");
        unsigned long mark = dmalloc_mark();
#endif
	if( 0 != ns_recurse( path, NULL, &prt_bleh, 0, NULL ) ) {
		printf("ns_recurse() returns with error: %s\n", uerr_str( uerr_get() ) );
	}

#ifdef WITH_DMALLOC
        dmalloc_log_changed( mark, 1, 1, 1 );
#endif
} // }}}

int cp_item( char *base, char *rel, long type, void *opt ) // {{{
{
	int r=0;
	long vtype;
	char *src = ((char**)opt)[0];
	char *dest = ((char**)opt)[1];
	VBUF_STATIC( buf, 10 );
	VBUF_STATIC( cpath, 10 );
	VBUF_STATIC( cdest, 10 );

	if( 0 != vb_pathjoin( cpath, base, rel ) ) return -1;
	if( 0 != vb_pathjoin( cdest, dest, rel ) ) return -1;

	//WARN("cp_item(%s) -> %s", cpath->b, cdest->b );
	switch( type ) {
		case RTYPE_KEY:
			//WARN("Should I? nk_mkdir(%s)\n", cdest->b );
			ns_mkdir( cdest->b, 0 );
			//ns_mkdir( dest, 0 );
			//WARN("ns_mkdir(%s) + %s ^ %s", dest, src, p );
			break;
		case RTYPE_VALUE:
			ns_read_valueT( cpath->b, &type, buf );
			WARN("cp_item( %s -> %s ):", cpath->b, cdest->b );
			hexdump( buf->b, 0, buf->dlen, 1 );
			ns_set_valueT( cdest->b, type, buf->b, buf->dlen );
			//printf("cp value %s to %s\n", cpath->b, cdest->b );

			break;
		default:
			DIE("cp_item(): Unknown typepp_item called on %s, type=%d", cpath->b, type );
	}

	if( 0 != r ) {
		fprintf(stderr, "cp_item(): %s\n", uerr_str( uerr_get() ) );
		return r;
	}

	return 0;
} // }}}
void do_cp( char *src, char *dest ) // {{{
{
	VBUF_STATIC( cp, 10 );

	int type=-1;
	//vb_pathjoin( cp, ns_getcwd(), src );
	ASSERT( cp->dlen <= MAX_PATH, "Path is too long (should not be possible).");
	//memcpy( path, cp->b, cp->dlen );
	char *parms[2];

	type = ns_gettype( src );
	//printf("Trying to remove %s which is of type %d\n", cp->b, type );
	switch( type ) {
		case RTYPE_VALUE:
			/*
			if( 0 != ns_unlink( cp->b ) ) {
				fprintf(stderr, "rm: %s\n", uerr_str( uerr_get()) );
			}
			*/
			break;
		case RTYPE_KEY:
			parms[0] = src;
			parms[1] = dest;
			ns_recurse( src, NULL, &cp_item, 0, parms );
			break;
		default:
			fprintf(stderr, "do_cp: Unknown type %d from %s\n", type , src);
	}
		/*
	if( 0 != ns_recurse( cp->b, &rm_item, 0, NULL ) ) {
		printf("ns_recurse() returns with error: %s\n", uerr_str( uerr_get() ) );
	}
	*/
} // }}}
void cmd_cp( FILE *fp, int argc, char **argv ) // {{{
{
	VBUF_STATIC( dirents, 10 ); 
	VBUF_STATIC( srcpath, 10 );
	VBUF_STATIC( dstpath, 10 );
	PRO_START;

	int i;
	char *dest;
	VBUF_STATIC( vbdirents, 10 );

	if( argc < 3 ) {
		fprintf(stderr, "cp command requires at least two parameters.\n");
		return;
	}

	dest = argv[ argc - 1 ];

	if( argc > 3 && !ns_test( dest, "-d" ) ) {
		fprintf(stderr, "destination must be a key when copying multiple keys/values\n");
		return;
	}

	if( argc == 3 && !ns_test( dest, "-e" ) ) {
		char *src = argv[ 1 ];
		//cp -r "foo" "foo2": We are copying a single item directly to the destination
		if( ns_test( src, "-d" ) ) {
			int nents = 0;
			struct ns_dirent *ents;
			if( 0 != ns_mkdir( dest, 0 )  ) {
				fprintf(stderr, "Could not create destination: %s, %s\n", dest, uerr_str( uerr_get() ) );
			}

			NREGDIR *dir;
			if( NULL == ( dir = ns_opendir( src ) ) ) {
				fprintf(stderr, "Couldn't open source directory %s, %s: ", src, uerr_str( uerr_get() ) );
				return;
			}
			if( 0 > ( nents =  ns_readfulldir( vbdirents, dir ) ) ) {
				fprintf(stderr, "Error reading key contents: %s", uerr_str( uerr_get() ) );
				ns_closedir( dir );
				return;
			}

			ents = (struct ns_dirent*)vbdirents->b;

			for( i=0; i< nents; i++ ) {
				vb_pathjoin( srcpath, src, ents[i].filename );
				vb_pathjoin( dstpath, dest, ents[i].filename );
				//fprintf(stderr, "You should write this part %s/%s -> %s, %s -> %s\n", src, ents[i].filename, dest, srcpath->b, dstpath->b);
				do_cp( srcpath->b, dstpath->b );
			}
			return;
		} else {
			WARN("Copying non-directory: %s %s", src, dest);
			do_cp( src, dest );
			WARN("  done.");
		}

	}

	WARN("here.");

	for( i=1; i<argc - 1; i++ ) {
		do_cp( argv[i], dest );
	}

	PRO_END("cmd_cp");

} // }}}
int rm_item( char *p, char *rel, long type, void *opt ) // {{{
{
	int r=0;

	VBUF_STATIC( cpath, 10 );
	//fprintf(stderr, "%s\n", p );

	if( 0 != vb_pathjoin( cpath, p, rel ) ) return -1;

	fprintf(stderr, "rm_item: %s\n", cpath->b );

	switch( type ) {
		case RTYPE_KEY:
			r = ns_rmdir( cpath->b );
			break;
		case RTYPE_VALUE:
			r = ns_unlink( cpath->b );
			break;
		case RTYPE_NOT_FOUND:
			WARN("rm_item called on %s: not found!", cpath->b );
			return -1;
		default:
			DIE("rm_item called on %s: unknown type=%d", cpath->b, type );
	}

	if( 0 != r ) {
		fprintf(stderr, "rm_item(): %s\n", uerr_str( uerr_get() ) );
		return r;
	}
	//printf("rm( p == %s, type == %d\n", p, type );
	return 0;
} // }}}
void cmd_rm( FILE *fp, int argc, char **argv ) // {{{ Delete a value or recursively delete a key
{
	VBUF_STATIC( cp, 100 );

	PRO_START;

	int i;
	int type=-1;
	//char path[MAX_PATH];

	if( argc < 2 ) {
		printf("rm command requires a parameter.\n");
		return;
	}

	for( i=1; i<argc; i++ ) {
		vb_pathjoin( cp, ns_getcwd(), argv[i] );
		ASSERT( cp->dlen <= MAX_PATH, "Path is too long (should not be possible).");
		//memcpy( path, cp->b, cp->dlen );

		type = ns_gettype( cp->b );
		//printf("Trying to remove %s which is of type %d\n", cp->b, type );
		switch( type ) {
			case RTYPE_VALUE:
				if( 0 != ns_unlink( cp->b ) ) {
					fprintf(stderr, "rm: %s\n", uerr_str( uerr_get()) );
				}
				break;
			case RTYPE_KEY:
				ns_recurse( cp->b, NULL, &rm_item, NSR_FLAG_PARENT_LAST, NULL );
				break;
			default:
				fprintf(stderr, "rm: Unknown type %d\n", type );
		}
			/*
		if( 0 != ns_recurse( cp->b, &rm_item, 0, NULL ) ) {
			printf("ns_recurse() returns with error: %s\n", uerr_str( uerr_get() ) );
		}
		*/
	}

	PRO_END("cmd_rm");

} // }}}
void cmd_cat( FILE *fp, int argc, char **argv ) // {{{
{
	NREGH* sh;
	char *path;
	int i;

	if( argc < 2 ) {
		printf("cat command requires a parameter.\n");
		return;
	}

	for( i=1; i<argc; i++ ) {
		path = argv[i];

		if( NULL == ( sh = ns_open( path ) ) ) {
			WARN("cmd_cat(): ns_open_relative() has returned, and failed.");
			fprintf(stderr, "ns_open_relative() failed: %s\n", uerr_str( uerr_get() ) );
			return;
		}

		printf("%s:", get_type_str( sh->type ) );

		switch( sh->type ) {
			default:
				fprint_val_nice( stdout, sh->buf, sh->type, sh->blen, 0 );
				break;
		}

		putchar('\n');

		ns_close( sh );
	}

} // }}}
void cmd_hd( FILE *fp, int argc, char **argv ) // {{{
{
	NREGH* sh;
	char *path;
	int i;

	if( argc < 2 ) {
		printf("cat command requires a parameter.\n");
		return;
	}

	for( i=1; i<argc; i++ ) {
		path = argv[i];

		if( NULL == ( sh = ns_open( path ) ) ) {
			fprintf(stderr, "ns_open_relative() failed: %s\n", uerr_str( uerr_get() ));
			return;
		}

		/*
		printf("Value <%s> of type %s, data length %ld [0x%lx]\n", path,
			get_type_str( sh->type ), sh->blen, sh->blen);
			*/

		printf("type: %s\n", get_type_str( sh->type ) );

		hexdump( sh->buf, 0, sh->blen, 1 );

		putchar('\n');

		ns_close( sh );
	}

} // }}}
void cmd_find_old( FILE *fp, char *path ) // {{{
{
	NREGDIR* sh;
	struct ns_dirent *d;
	vbuf *tpath = vbuf_alloc(100);
	vbuf *vpath = vbuf_alloc(100);

	//char lbuf[SZ_MAX];

	//WARN("cmd_find_old running on >%s<", path );
	WARN("cmd_find_old() disabled for update to ns_recurse()");
	return;

	/*
	if( NULL == ( sh = ns_opendir( path ) ) ) {
		fprintf(stderr, "Error opening directory (%s, %s): %s\n", uerr_str( uerr_get() ), ns_getcwd(), path);
		vbuf_free( tpath );
		vbuf_free( vpath );
		return;
	}

	while( ( d = ns_readdir( sh ) ) ) {
		//this is nice
		
		vbuf_printf( tpath,"./%s/%s", path, d->filename);

		//WARN("Processing directory... %s", tpath->b );
		vb_process_path( vpath, tpath->b );
		switch( ns_gettype( sh, d->filename ) ) {
			case RTYPE_KEY: {
				//WARN("HERE1");
				printf("K ");
				printf("%s\n", vpath->b );
				//printf("Recursing with: <%s>\n", vpath->b );
				//snprintf(lbuf, SZ_MAX, "./%s/%s", path, d->filename );
				cmd_find_old(fp, tpath->b );
				break;
			}
			case RTYPE_VALUE:
				//WARN("HERE2");
				printf("V ");
				printf("%s\n", vpath->b );
					break;
			case RTYPE_NOT_FOUND: {
				WARN("RTYPE_NOT_FOUND: %s", tpath->b );
				//WARN("HERE3");
				//printf("RTYPE_NOT_FOUND: %s\n", lbuf);
				break;
			}
			case -1: {
				 WARN("There was an error reading type file type: %s", uerr_str( uerr_get() ) );
				 break;
			}

			default: { 
				WARN("HERE4");
				//printf("RTYPE_NOT_KEY: %s\n", lbuf);
				break;
			}
		}
	}

	vbuf_free( tpath );
	vbuf_free( vpath );

	ns_closedir( sh );
	*/
} // }}}
void cmd_find( FILE *fp, int argc, char **argv ) // {{{
{
	int i;
	if( argc > 1 ) {
		for( i=1; i<argc; i++ ) {
			cmd_find_old( fp, argv[i] );
		}
	}
	else
		cmd_find_old( fp, "" );
} // }}}
void cmd_quit( FILE *fp, int argc, char **argv ) // {{{
{
	done=1;
} // }}}
void cmd_nv( FILE *fp, int argc, char **argv ) // {{{
{
	char *vpath;
	char *stfmt;
	char *sttype = NULL;
	char *dataa;
	vbuf *vbdata = vbuf_alloc( 0 );
	int fmt;
	int type;
	struct varbuf *vb = vbuf_alloc(0);

	//TODO: uglyness

	switch( argc ) {
		case 4:
			//printf("Format 1\n");
			vpath = argv[1];
			stfmt = argv[2];
			dataa = argv[3];
			break;
		case 5:
			//printf("Format 2\n");
			vpath = argv[1];
			stfmt = argv[2];
			sttype = argv[3];
			dataa = argv[4];
			break;
		default:
			printf("Invalid number of parameters.  Try \"help nv\".\n");
			return;
	}

	vbuf_strset( vbdata, dataa );

	//printf("nv vpath=%s, stfmt=%s, sttype=%s, dataa=%s\n", vpath, stfmt, sttype, dataa );

	if( 0 > ( fmt = str2valtype( stfmt ) ) ) {
		printf("Couldn't get valid value type from %s\n", stfmt );
		return;
	}

	if( NULL != sttype ) {
		if( 0 > ( type = str2valtype( sttype ) ) ) {
			printf("Couldn't get valid value type from %s\n", sttype );
			return;
		}
	}
	else {
		type = fmt;
	}

	if( 0 != regp_udata( vb, fmt, vbdata ) ) {
		printf("parse_data(%s) failed: %s\n", vbdata->b, uerr_str( uerr_get() ) );
		return;
	}

	//WARN("Calling ns_set_valueT w/ %s, %d, %s", vpath, type, vb->b );

	if( -1 == ns_set_valueT( vpath, type, vb->b, vb->dlen ) ) {
		printf("Couldnt' set value: %s\n", uerr_str( uerr_get() ) );
		return;
	}
} // }}}
void cmd_mkdir( FILE *fp, int argc, char **argv ) // {{{
{
	int i;
	if( argc < 2 ) {
		printf("Command mkdir requires a parameter.\n");
		return;
	}
	for( i=1; i<argc; i++ ) {
		if( 0 != ns_mkdir( argv[i], 0 ) ) {
			printf("cmd_mkdir(): ns_mkdir_relative(): %s\n", uerr_str( uerr_get() ) );
			return;
		}
	}
} // }}}
void cmd_lsusers( FILE *fp, int argc, char **argv ) // {{{
{
	static struct varbuf *vb = NULL;
	static struct varbuf *nbuf = NULL;
	int *ridlist;
	int i;

	if( NULL == vb ) vb = vbuf_alloc( 10 );
	if( NULL == nbuf ) nbuf = vbuf_alloc( 10 );

	if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error: %s\n", uerr_str( uerr_get() ) );
		return;
	}

	ridlist = (int*)vb->b;

	for( i=0; i<vb->dlen; i++ ) {
		if( 0 == sam_get_user_name( nbuf, ridlist[i]  ) ) {
			printf("%08x %s\n", ridlist[i], nbuf->b );
		}
		else {
			printf("Error reading data for RID %x: %s\n", ridlist[i], uerr_str( uerr_get() ) );
		}
	}

} // }}}
void cmd_userinfo( FILE *fp, int argc, char **argv ) // {{{
{
	int rid;

	VBUF_STATIC(buf, 100);

	if( argc != 2 ) {
		fprintf(fp, "userinfo: please specify a user name\n");
		return;
	}

	if( 0 > ( rid = sam_uname2rid( argv[1] ) )  ) {
		printf("Error converting user name to RID: %s\n", uerr_str( uerr_get() ) );
		return;
	}
	//WARN("cmd_userinfo() sam_uname2rid returns: %x\n", rid );
	sam_print_user( buf, rid );
	printf("%s\n", buf->b );

	return;

} // }}}
void cmd_samstuff( FILE *fp, int argc, char **argv ) // {{{
{
	struct samkeyf *sf;
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( buf, 100 );
	if( NULL == (sf = sam_get_samkeyf(vb)) ) {
		fprintf(fp, "sam_get_samkeyf() returned NULL.\n");
		return;
	}
	samkeyf_dump( buf, sf );
	printf("%s\n", buf->b);
	return;

} // }}}
void cmd_passwd( FILE *fp, int argc, char **argv ) // {{{
{
	int rid;

	if( argc != 2 ) {
		fprintf(fp, "userinfo: please specify a user name\n");
		return;
	}

	if( 0 > ( rid = sam_uname2rid( argv[1] ) ) ) {
		WARN("Failed to set password: %s\n", uerr_str( uerr_get() ) );
		return;
	}
	if( 0 != change_pw( rid, 0 ) ) {
		WARN("Failed to set password: %s\n", uerr_str( uerr_get() ) );
		return;
	}

	return;
} // }}}
void cmd_ls_hives( FILE *fp, int argc, char **argv ) // {{{
{
	char *path;
	int i;
	VBUF_STATIC( mp, 100 );

	for(i=0; i<hl_mount_count(); i++ ) {
		hl_mount_get_point( mp, i );
		fprintf(stderr, "% 2d %s\n", i, mp->b );
	}
} // }}}
void cmd_scan_hive( FILE *fp, int argc, char **argv ) // {{{
{
	char *path;
	int i;

	if( argc < 2 ) {
		printf("rm requires a parameter\n");
		return;
	}
	for( i=1; i<argc; i++) {
		int no;
		if( 1 != sscanf( argv[i], "%d", &no ) ) {
			printf("I was expecting an integer, got this: %s\n", argv[i] );
			return;
		}
		if( no >= hl_mount_count() ) {
			printf("%d > mount count.\n", no, hl_mount_count() );
			return;
		}
		rl_hkey hkey = hl_mount_get_hkey( no );

		hive_scan( hkey->hd );
	}
} // }}}
void cmd_hive_dump_unallocd( FILE *fp, int argc, char **argv ) // {{{
{
	char *path;
	int no;
	FILE *out = stdout;
	VBUF_STATIC( outbuf, 100 );

	if( argc < 2 ) {
		printf("uadump requires a parameter\n");
		return;
	}
	if( argc == 3 ) {
		if( NULL == ( out = fopen( argv[2], "w" ) ) ) {
			fprintf(stderr, "Couldn't open file %s for write: %s\n",
					argv[2], strerror( errno ) );
			return;
		}
	}
	if( 1 != sscanf( argv[1], "%d", &no ) ) {
		printf("I was expecting an integer, got this: %s\n", argv[1] );
		return;
	}
	if( no >= hl_mount_count() ) {
		printf("%d > mount count.\n", no, hl_mount_count() );
		return;
	}
	rl_hkey hkey = hl_mount_get_hkey( no );

	struct hbin_page *p = hive_get_first_hbinp( hkey->hd );
	do {
		struct hbin_block *b = hbinp_first_hbinb( p );
		do {
			if( !hbinb_is_allocd( b ) )  {
				hbinb_vdump( outbuf, hkey->hd, b );
				fprintf(out, "%s", outbuf->b );
			}
		}while( b = hbinb_get_next( hkey->hd, p, b ) );
	} while( p = hbinp_get_next_hbinp( hkey->hd, p ) );
} // }}}
void cmd_hive_alloc_graph( FILE *fp, int argc, char **argv ) // {{{
{
	char *path;
	int no;
	FILE *out = stdout;
	VBUF_STATIC( outbuf, 100 );

	if( argc < 2 ) {
		printf("uadump requires a parameter\n");
		return;
	}
	if( argc == 3 ) {
		if( NULL == ( out = fopen( argv[2], "w" ) ) ) {
			fprintf(stderr, "Couldn't open file %s for write: %s\n",
					argv[2], strerror( errno ) );
			return;
		}
	}
	if( 1 != sscanf( argv[1], "%d", &no ) ) {
		printf("I was expecting an integer, got this: %s\n", argv[1] );
		return;
	}
	if( no >= hl_mount_count() ) {
		printf("%d > mount count.\n", no, hl_mount_count() );
		return;
	}
	rl_hkey hkey = hl_mount_get_hkey( no );

	struct hbin_page *p = hive_get_first_hbinp( hkey->hd );
	do {
		struct hbin_block *b = hbinp_first_hbinb( p );
		fprintf(out, "== Page F:%#8x\n", get_file_offset( hkey->hd, p));
		int tc=0;
		do {
			fprintf(out, "[%c %4d/%#5x] ",
					 (hbinb_is_allocd( b ))?'O':'.', 
					hbinb_size( b ), hbinb_size( b ) );
			if( 8 < tc++ ) {
				fprintf(out, "\n");
				tc=0;
			}
		}while( b = hbinb_get_next( hkey->hd, p, b ) );

		fprintf(out, "\n");
		b = hbinp_first_hbinb( p );

		tc=0;
		do {
			int bc=0;
			for( bc=0; bc < hbinb_size( b ); bc+= 8 ) {
				tc++;
				if( 0 == (tc % 60) )
					fprintf(out, "\n");
				if( 0 == bc )
					fprintf(out, " ");
				else
					fprintf(out, ".");
				fprintf(out,"%c", (hbinb_is_allocd( b ))?'O':'.' );
			}
		}while( b = hbinb_get_next( hkey->hd, p, b ) );

		fprintf(out,"\n");
	} while( p = hbinp_get_next_hbinp( hkey->hd, p ) );
} // }}}
// }}}
// Interactive Input {{{
// Readline -> argc,argv, etc
//Due to license issues I have to write the following. It is crap and should be
//replaced with something torn from bash or wherever
//TODO: escapes should be observed
// [statement] := [rawarg][rawarg]...
// [rawarg] := \s*[cnstr]\s*
// [cnstr] := \S+ || [quotestr]
// [quotestr] := "[^\"]*"
#define CTOK_ERROR 1
#define CTOK_EMPTY 2
#define CTOK_STR 3
int is_quote( int c ) { // {{{ if c == " || ' returns 1, else returns 0
	switch( c ) {
		case '"':
		case '\'':
			return 1;
		default:
			return 0;
	}
}// }}}
int read_cmd_token( char **b, struct varbuf *vb, char **next ) // {{{ in: b == str: out: b == start of token, len == length of token, returns: token type
{
	int len=0;
	vbuf_set( vb, "", 1 );
	while ( isspace( (*b)[0] ) ) (*b)++; //dump initial white space

	if( (*b)[0] == '\0' ) return CTOK_EMPTY;

	while( !isspace( (*b)[len] ) && (*b)[len] != 0 ) {
		if( is_quote( (*b)[len] ) ) {
			int qc = (*b)[len]; //grab quote character
			int q2off=-1;

			//printf("pre-quote: %s\n", vb->b );
			len++;
			char *qv = index( *b + len, qc );
			if( NULL == qv ) {
				fprintf(stderr, "Untermainted quote string\n");
				return CTOK_ERROR;
			}

			q2off = qv - (*b + len);

			//string append
			vb->dlen--;
			vbuf_append( vb, *b + len, q2off );
			vbuf_append( vb, "", 1 );

			len += qv - ( *b + len ) + 1;

			*next = *b + len + 1;
		}
		else {
			vb->dlen--;
			vbuf_append( vb, (*b) + len, 1 );
			vbuf_append( vb, "", 1 );

			(len)++;
		}

	}

	*next = *b + len;

	//printf("Returning %d, %s\n", len, vb->b );

	return CTOK_STR;
} // }}}
#define MAX_ARGS 20
int read_cmd( char *prompt, int *argc, char **argv ) // {{{ read a user command, populate argc and argv; NOTE: you must free() argv; returns -1 on empty
{
	char *rlbuf=NULL, *bp;
	int i,t; //len;
	char *next;

	static struct varbuf *cmd_buf = NULL;
	if( cmd_buf == NULL ) cmd_buf = vbuf_alloc( 0 );

	*argc = 0;

	rlbuf = readline( prompt );
	add_history( rlbuf );
	bp = rlbuf;

	while( CTOK_EMPTY != ( t = read_cmd_token( &bp, cmd_buf, &next ) ) ) {

		//printf("read_cmd_token(): returns: %d\n", len );

		//ASSERT( len + 1 < SZ_MAX, "slen > SZ_MAX, sorry, buffer is too small."); 

		switch( t ) {
			case CTOK_STR:
				break;

			case CTOK_ERROR:
				fprintf(stderr, "Tokenizer error.\n");
				return -3;
			default:
				DIE("Unknown result from read_cmd_token: >%d<\n", t ); 
		}

		argv[*argc] = F_MALLOC( cmd_buf->dlen + 1 );
		strncpy( argv[*argc], cmd_buf->b, cmd_buf->dlen );
		argv[*argc][ cmd_buf->dlen ] = '\0';

		//printf("Result: %d: \"%s\", %s\n", *argc, argv[*argc], next);

		(*argc)++;

		ASSERT( *argc < MAX_ARGS, "More arguments than MAX_ARGS, sorry." );

		bp = next;
	}

	if( rlbuf != NULL ) { free( rlbuf ); rlbuf = NULL; }

	if( 0 == strcmp( argv[0], "SEP" ) ) return -1;
	if( *argc > 0 && argv[0] ) { //find the given command
		for(i=0; maincmds[i].str[0] != '\0' ; i++ ) {
			if ( 0 == strcmp(argv[0], maincmds[i].str ) ) {
				return i;
			}
		}
	}
	else {
		return -2;
	}

	return -1;
} // }}}
void editor(void) // {{{
{
	/* Simple interactive command-parser
	* Main loop for manually looking through the registry
	*/
	char prompt[SZ_MAX];
	char *myargv[MAX_ARGS];
	int myargc;
	int i;
	int cmdno=-1;

	//char p[] = "/HKEY_LOCAL_MACHINE";
	char p[] = "/";
	if( 0 != ns_chdir( p ) ) 
		DIE("Couldn't chdir to initial working directory %s: %s", p, uerr_str( uerr_get() ) );

	printf("Simple registry editor. 'help' for help.\n");

	while (!done) {

		snprintf(prompt, SZ_MAX, "%s> ", ns_getcwd() );
		cmdno = read_cmd( prompt, &myargc, myargv );

		if( cmdno >= 0 )
			(maincmds[cmdno].cmd)(stdout, myargc, myargv );
		else {
			switch( cmdno ) {
				case -1 :
					printf("Unknown command.\n");
					break;
				case -2:
					//printf("You entered nothing.\n");
					break;
				case -3:
					printf("Parser error.\n");
					break;
				default:
					DIE("I don't know what this might be all about.");
			}
		}

		for( i=0; i<myargc; i++ )
			free( myargv[i] );
		myargc = 0;
	}
} // }}}
int user_get_password( char *buf, int len ) // {{{ Prompt the user for a password, dump it in buf, return length
{
	char yn[4];
	printf("* = blank the password (This may work better than setting a new password!)\n");
	/*   printf("@ = promote user to administrator\n"); */
	printf("Enter nothing to leave it unchanged\n");
	int pl = fmyinput("Please enter new password: ",buf,len);

	return pl;
} // }}}
// }}}

#undef LSADATA

// Pending -- stuff from chntpw that I haven't done anything with yet {{{
void sz_edit( struct valh *vh ) // {{{ Print a REG_SZ*, allow user to modify, store result
{

	DIE("This is being executed.");
//	R_SZ8( vh, &dbuf );
//	dbuf = (char *)&vh->buf;
//	cheap_uni2ascii(dbuf,origstring,len);
//	n = 0; i = 0;
//	while (i < (len>>1)-1) {
//		printf("[%2d]: %s\n",n,origstring+i);
//		i += strlen(origstring+i) + 1;
//		n++;
//	}
//
//	printf("\nNow enter new strings, one by one.\n");
//	printf("Enter nothing to keep old.\n");
//	if (type == REG_MULTI_SZ) {
//		printf("'--n' to quit (remove rest of strings)\n");
//		printf("'--i' insert new string at this point\n");
//		printf("'--q' to quit (leaving remaining strings as is)\n");
//		printf("'--Q' to quit and discard all changes\n");
//		printf("'--e' for empty string in this position\n");
//	}
//	n = 0; i = 0; in = 0; go = 0; done = 0;
//
//	while (i < (len>>1)-1 || !done) {
		
//		char inbuf[SZ_MAX];
//		printf("[%2d]: %s\n",n, insert == 1 ? "[INSERT]" : ((i < (len>>1)-1 ) ? origstring+i : "[NEW]"));
//		if (insert) insert++;
//		if (!go) fmyinput("-> ",inbuf, 500);
//		else *inbuf = 0;
//		if (*inbuf && strcmp("--q", inbuf)) {
//			if (!strcmp("--n", inbuf) || !strcmp("--Q", inbuf)) { /* Zap rest */
//				i = (len>>1) +1; done = 1;
//			} else if (strcmp("--i", inbuf)) {  /* Copy out given string */
//				if (!strcmp("--e",inbuf)) *inbuf = '\0';
//				if (newstring) newstring = realloc(newstring, in+strlen(inbuf)+1);
//				else newstring = malloc(in+strlen(inbuf)+1);
//				strncpy(newstring+in, inbuf, 499);
//				in += strlen(inbuf)+1;
//			} else {
//				insert = 1;
//			}
//		} else {  /* Copy out default string */
//			if (newstring) newstring = realloc(newstring, in+strlen(origstring+i)+1);
//			else newstring = malloc(in + strlen(origstring+i) + 1);
//			strncpy(newstring+in, origstring+i, 499);
//			in += strlen(origstring+i)+1;
//			if (!strcmp("--q", inbuf)) { go = 1; done = 1; }
//		}
//		if (!insert) i += strlen(origstring+i) + 1;
//		if (insert != 1) n++;
//		if (insert == 2) insert = 0;
//		if (type != REG_MULTI_SZ) {
//			i = (len<<1);
//			done = 1;
//		}
//	}
//
//			if (strcmp("--Q", inbuf)) {  /* We didn't bail out */
//				if (newstring) newstring = realloc(newstring, in+1);
//				else newstring = malloc(in+1);
//				in++;
//				*(newstring+in) = '\0';  /* Must add null termination */
//
//				ALLOC(newkv,1,(in<<1)+sizeof(int));
//				newkv->len = in<<1;
//				cheap_ascii2uni(newstring, (char *)&(newkv->data), in);
//
//				d = 1;
//			}
//			FREE(kv);
//			kv = newkv;
//			break;
} // }}}
#if 0
void edit_val(rl_hkey hk, char *path ) //{{{
{
	/* Edit value: Invoke whatever is needed to edit it
	 * based on its type
	 */
	//struct keyval *kv, *newkv;
	int d = 0;
	//int type,done , insert = 0;
	char inbuf[SZ_MAX+4];
	//char origstring[SZ_MAX+4];
	//char *newstring = NULL;
	//char *dbuf;
	struct valh *vh;

	if( NULL == ( vh = vh_open( hk, path ) ) ) {
		CRAP_ERROR;
		return;
	}
//	type = get_val_type(h, nkofs, path);
//	if (type == -1) {
//		printf("Value <%s> not found!\n",path);
//		return;
//	}
//
//	kv = get_val2buf(h, NULL, nkofs, path, type);
//	if (!kv) {
//		printf("Unable to get data of value <%s>\n",path);
//		return;
//	}
//	len = kv->len;

	printf("EDIT: <%s> of type %s with length %ld [0x%lx]\n", path,
		get_type_str( vh->type ),
		vh->blen, vh->blen);

	switch(vh->type) {
		case REG_DWORD:
		{
			ulong v = vh_get_dword(vh);
			printf("DWORD: Old value %ld [0x%lx], ", v, v );
			fmyinput("enter new value (prepend 0x if hex, empty to keep old value)\n-> ",
				inbuf, 12);
			if (*inbuf) {
				sscanf(inbuf,"%li", &v);
				d = 1;
			}
			printf("DWORD: New value %ld [0x%lx], ", v, v);
			vh_set_dword(vh, v);
			break;
		}
		case REG_SZ:
		case REG_EXPAND_SZ:
		case REG_MULTI_SZ:
			sz_edit( vh );
			break;

		case REG_BINARY:
			fmyinput("New length (ENTER to keep same): ",inbuf,90);
			vh_set_buf( vh, vh->buf, atoi(inbuf) ); //TODO: yuck, check input
			/*
			if (*inbuf) {
				newsize = atoi(inbuf);
				ALLOC(newkv,1,newsize+sizeof(int)+4);
				bzero(newkv,newsize+sizeof(int)+4);
				memcpy(newkv, kv, (len < newsize) ? (len) : (newsize) +sizeof(int));
				FREE(kv);
				kv = newkv;
				kv->len = newsize;
			}
			*/
			d = debugit((char *)vh->buf, vh->blen);
			break;

		default:
			printf("Type not handeled (yet), invoking hex editor on data!\n");
	}

//	if (d) {
//		if (!(put_buf2val(h, kv, nkofs, path, type))) {
//			printf("Failed to set value!?\n");
//		}
//	}
	vh_close( vh );
} //}}}
#endif
void disable_syskey( struct samkeyf *ff ) // {{{
{
} // }}}

void handle_syskey(void) // {{{
{

	/* Check for presence of syskey and possibly disable it if
	* user wants it.
	* This is tricky, and extremely undocumented!
	* See docs for more info on what's going on when syskey is installed
	*/

#ifdef LSADATA
	/* SYSTEM\CurrentControlSet\Control\Lsa\Data, only on NT5?? */
	/* Probably contains some keyinfo for syskey. Byte 0x34 seems to be mode */
	struct lsadata {
		char keydata[0x34];        /* Key information */
		int  syskeymode;           /* Is this what we're looking for? */
	};
#endif

	/* void *fdata; */
	struct samkeyf *ff = NULL;
	struct secpoldata *sf = NULL;
	/* struct lsadata *ld = NULL; */
	int /* len, */ i,secboot, samfmode, secmode /* , ldmode */ ;
	//struct keyval *secpol, *samf; /* *lsad */ ;
	//struct keyval samff;
	char *syskeytypes[4] = { "off", "key-in-registry", "enter-passphrase", "key-on-floppy" }; 
	char yn[5];

	printf("\n---------------------> SYSKEY CHECK <-----------------------\n");

	/*
	if (H_SAM < 0) {
		printf("ERROR: SAM hive not loaded!\n");
		return;
	}
	samf = get_val2buf(hive[H_SAM], NULL, 0, "\\SAM\\Domains\\Account\\F", REG_BINARY);
	*/

	long stsize = 0;
	long type;
	char *afbuf = NULL;
	ASSERT( rlERROR_MORE_DATA == rlRegQueryValueEx( rlHKEY_LOCAL_MACHINE, "\\SAM\\SAM\\Domains\\Account\\F", &type, NULL, &stsize ), "Failed to query size of F struct" );
	afbuf = F_MALLOC( stsize );
	ASSERT( rlERROR_SUCCESS == rlRegQueryValueEx( rlHKEY_LOCAL_MACHINE, "\\SAM\\SAM\\Domains\\Account\\F", &type, afbuf, &stsize ), "Failed to query size of F struct" );

	DIE("TODO: replace the following...");
	/*
	samf = (struct keyval *)&afbuf;

	samf = &samff;

	if (samf && samf->len > 0x70 ) {
		ff = (struct samkeyf *)&samf->data;
		samfmode = ff->syskeymode;
	} else {
		samfmode = -1;
	}

	secboot = -1;
	if (H_SYS >= 0) {
		secboot = get_dword(hive[H_SYS], 0, "\\ControlSet001\\Control\\Lsa\\SecureBoot");
	}

	*/


	DIE("TODO: replace the following.");
//	secmode = -1;
//	if (H_SEC >=0) {
//		secpol = get_val2buf(hive[H_SEC], NULL, 0, "\\Policy\\PolSecretEncryptionKey\\@", REG_NONE);
//		if (secpol) {     /* Will not be found in NT 4, take care of that */
//			sf = (struct secpoldata *)&secpol->data;
//			secmode = sf->syskeymode;
//		}
//	}

#ifdef LSADATA
	DIE("TODO: replace the following.");
	/*
	lsad = get_val2buf(hive[H_SYS], NULL, 0, "\\ControlSet001\\Control\\Lsa\\Data\\Pattern", REG_BINARY);

	if (lsad && lsad->len >= 0x38) {
		ld = (struct lsadata *)&lsad->data;
		ldmode = ld->syskeymode;
	} else {
		ldmode = -1;
	}
	*/
#endif

	printf("SYSTEM   SecureBoot            : %d -> %s\n", secboot,
	(secboot < 0 || secboot > 3) ? "Not Set (not installed, good!)" : syskeytypes[secboot]);
	printf("SAM      Account\\F             : %d -> %s\n", samfmode,
	(samfmode < 0 || samfmode > 3) ? "Not Set" : syskeytypes[samfmode]);
	printf("SECURITY PolSecretEncryptionKey: %d -> %s\n", secmode,
	(secmode < 0 || secmode > 3) ? "Not Set (OK if this is NT4)" : syskeytypes[secmode]);

#ifdef LSADATA
	printf("SYSTEM   LsaData               : %d -> %s\n\n", ldmode,
	(ldmode < 0 || ldmode > 3) ? "Not Set (strange?)" : syskeytypes[ldmode]);
#endif

	if (secboot != samfmode && secboot != -1) {
		printf("WARNING: Mismatch in syskey settings in SAM and SYSTEM!\n");
		printf("WARNING: It may be dangerous to continue (however, resetting syskey\n");
		printf("         may very well fix the problem)\n");
	}

	if (secboot > 0 || samfmode > 0) {
		printf("\n***************** SYSKEY IS ENABLED! **************\n");
		printf("This installation very likely has the syskey passwordhash-obfuscator installed\n");
		printf("It's currently in mode = %d, %s-mode\n",secboot,
		(secboot < 0 || secboot > 3) ? "Unknown" : syskeytypes[secboot]);

		if (no_hives < 2) {
			printf("\nSYSTEM (and possibly SECURITY) hives not loaded, unable to disable syskey!\n");
			printf("Please start the program with at least SAM & SYSTEM-hive filenames as arguments!\n\n");
			return;
		}
		printf("SYSKEY is on! However, DO NOT DISABLE IT UNLESS YOU HAVE TO!\n");
		printf("This program can change passwords even if syskey is on, however\n");
		printf("if you have lost the key-floppy or passphrase you can turn it off,\n");
		printf("but please read the docs first!!!\n");
		printf("\n** IF YOU DON'T KNOW WHAT SYSKEY IS YOU DO NOT NEED TO SWITCH IT OFF!**\n");
		printf("NOTE: On WINDOWS 2000 it will not be possible\n");
		printf("to turn it on again! (and other problems may also show..)\n\n");
		printf("NOTE: Disabling syskey will invalidate ALL\n");
		printf("passwords, requiring them to be reset. You should at least reset the\n");
		printf("administrator password using this program, then the rest ought to be\n");
		printf("done from NT.\n");

		fmyinput("\nDo you really wish to disable SYSKEY? (y/n) [n] ",yn,2);
		if (*yn == 'y') {
			/* Reset SAM syskey infostruct, fill with zeroes */
			if (ff) { 
				DIE("syskeymode is gone, long live something else.");
				//ff->syskeymode = 0;

				for (i = 0; i < 0x3b; i++) {

					DIE("syskeyflags is gone, long live something else.");
					//ff->syskeyflags1[i] = 0;
				}

				DIE("TODO: replace following.");
				//put_buf2val(hive[H_SAM], samf, 0, "\\SAM\\Domains\\Account\\F", REG_BINARY);

			}
			/* Reset SECURITY infostruct (if any) */
			if (sf) { 

				DIE("TODO: replace following.");
				//memset(sf, 0, secpol->len);
				sf->syskeymode = 0;

				DIE("TODO: replace following.");
				//put_buf2val(hive[H_SEC], secpol, 0, "\\Policy\\PolSecretEncryptionKey\\@", REG_BINARY);

			}

#if LSADATA
			if (ld) { 

				ld->syskeymode = 0;

				DIE("TODO: replace following.");
				//put_buf2val(hive[H_SYS], lsad, 0, "\\ControlSet001\\Control\\Lsa\\Data\\Pattern", REG_BINARY);

			}
#endif

			/* And SYSTEM SecureBoot parameter */

			DIE("TODO: replace following.");
			//put_dword(hive[H_SYS], 0, "\\ControlSet001\\Control\\Lsa\\SecureBoot", 0);

			dirty = 1;
			syskeyreset = 1;
			printf("Updating passwordhash-lengths..\n");
			DIE("list_users() doesn't exist at the moment.");
			//list_users(1);
			printf("* SYSKEY RESET!\nNow please set new administrator password!\n");
		} else {

			syskeyreset = 1;
		}
	} else {
		printf("Syskey not installed!\n");
		return;
	}

} // }}}
void useredit(void) // {{{ /* Interactive user edit */
{
	char iwho[100];
	int il;

	printf("\n\n===== Hiveshell Edit User Info & Passwords ====\n\n");

	/*
	if (H_SAM < 0) {
		printf("ERROR: SAM registry file (which contains user data) is not loaded!\n\n");
		return;
	}
	*/


	while (1) {
		DIE("list_users() does not exist at the moment.");
		//list_users(1);

		printf("\nSelect: ! - quit, . - list users, 0x<RID> - User with RID (hex)\n");
		printf("or simply enter the username to change: [%s] ",admuser);
		il = fmyinput("",iwho,32);
		if (il == 1 && *iwho == '.') {
			DIE("list_users() does not exist at the moment.");
			//list_users(1);
			continue; }
		if (il == 1 && *iwho == '!') return;
		if (il == 0) strcpy(iwho,admuser);
		find_n_change(iwho);
	}

} // }}}
void recoveryconsole() // {{{
{
	int cmd = 0;
	int sec = 0;
	char yn[5];
	DIE("TODO: replace the folloiwng.");
	//static char *scpath = "\\Microsoft\\Windows NT\\CurrentVersion\\Setup\\RecoveryConsole\\SetCommand";
	//static char *slpath = "\\Microsoft\\Windows NT\\CurrentVersion\\Setup\\RecoveryConsole\\SecurityLevel";


	if (H_SOF < 0) {
		printf("\nERROR: SOFTWARE-hive not loaded, and there's where RecoveryConsole settings are..\n");
		return;
	}

	DIE("Replace following.");
	//cmd = get_dword(hive[H_SOF],0,scpath);
	//sec = get_dword(hive[H_SOF],0,slpath);

	if (cmd == -1 && sec == -1) {
		printf("\nDid not find registry entries for RecoveryConsole. Strange unless this is pre-win2k\n");
		return;
	}

	printf("\nRecoveryConsole:\n- Extended SET command is \t%s\n", cmd>0 ? "ENABLED (1)" : "DISABLED (0)");
	printf("- Administrator password login: %s\n", sec>0 ? "SKIPPED (1)" : "ENFORCED (0)");

	fmyinput("\nDo you want to change it? (y/n) [n] ",yn,2);
	if (*yn == 'y') {
		cmd ^= 1;
		sec ^= 1;
		DIE("Replace following");
		//if (!put_dword(hive[0], 0, scpath, cmd)) printf("Update of SET level failed registry edit\n");
		//if (!put_dword(hive[0], 0, slpath, sec)) printf("Update of login level failed registry edit\n");
		printf("Done!\n");
	}

} // }}}
void interactive(void) // {{{ /* Interactive menu system */
{
	int il;
	char inbuf[20];


	while(1) {
		printf("\n\n<>========<> Hiveshell Main Interactive Menu <>========<>\n\n"
		"Loaded hives:");
		WARN("TODO: replace the following?");
		/*
		for (il = 0; il < no_hives; il++) {
			printf(" <%s>",hive[il]->filename);
		}
		*/
		printf("\n\n  1 - Edit user data and passwords\n"
		"  2 - Syskey status & change\n"
		"  3 - RecoveryConsole settings\n"
		"      - - -\n"
		"  9 - Registry editor, now with full write support!\n"
		"  q - Quit (you will be asked if there is something to save)\n"
		"\n\n");

		il = fmyinput("What to do? [1] -> ", inbuf, 10);

		if (!il) useredit();
		if (il) {
			switch(inbuf[0]) {
				case '1': useredit(); break;
				case '2': handle_syskey(); break;
				case '3': recoveryconsole(); break;
				case '9': editor(); break;
				case 'q': return; break;
			}
		}
	}
} // }}}
void find_n_change(char *username) // {{{
{

	/* Find a username in the SAM registry, then get it's V-value,
	* and feed it to the password changer.
	*/
	char s[200];
	//struct vk_key *vkkey;
	struct keyval *v;
	int rid = 0;

	if (!username) return;
	if (*username == '0' && *(username+1) == 'x') sscanf(username,"%i",&rid);

	rid = sam_uname2rid( username );

//	if (!rid) { /* Look up username */
//		/* Extract the unnamed value out of the username-key, value is RID  */
//		snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\Names\\%s\\@",username);
//		rid = get_dword(hive[H_SAM],0,s);
//		if (rid == -1) {
//			printf("Cannot find value <%s>\n",s);
//			return;
//		}
//	}

	/*
	printf("Username: %s, RID = %d (0x%0x)\n",username,rid,rid);
	*/

	/* Now that we have the RID, build the path to, and get the V-value */
	long blen=0;
	char *buf;
	DIE("I commented out the below line");
	//ASSERT( NULL == sam_get_user_V( rid, NULL, &blen ), "Got strange return from sam_get_user_V." );
	buf = F_MALLOC( blen );

	DIE("I commented out the below line");
	//ASSERT( NULL != sam_get_user_V( rid, buf, &blen ), "Problem with sam_get_user_V." );

//	snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\%08X\\V",rid);
//	v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY);
//	if (!v) {
//		printf("Cannot find value <%s>\n",s);
//		return;
//	}

	v = (struct keyval *)buf;
	if (blen < 0xcc) {
		printf("Value <%s> is too short (only %ld bytes) to be a SAM user V-struct!\n",
		s, blen);
	} else {
		change_pw( rid, 0);
		if (dirty) {
			WARN("TODO: WRITE STUFF TO REGISTRY HERE");
//			if (!(put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY))) {
//				printf("Failed to write updated <%s> to registry! Password change not completed!\n",s);
//			}
		}
	}
	FREE(buf);
} // }}}
// }}}
// {{{ Almost Dumpster
#if 0
	// {{{ Old
	WARN("TODO: replace the following");
	//if (dodebug) debugit(hive[0]->buf,hive[0]->size);
	if(0) DIE("Hi!  I'm impossible!.");
	else {
		//check_get_samdata();
		//DIE("I commented out the above because it no longer exists.");
		if (list && !edit && !inter && !export) {
			DIE("list_users() does not exist at the moment");
			//if ( list_users(1) ) edit = 1;
		}

		if (edit) editor();
		else if (!export && who) { handle_syskey(); find_n_change(who); }


		if (inter) interactive();
		//     if (export) doexport( regpath );
	}

	hl_close_all();

	if( gverbose ) {
		printf("\nHives that have changed:\n #  Name\n");
		for (il = 0; il < no_hives; il++) {
			DIE("TODO: replace me");
			/*
			if (hive[il]->state & HMODE_DIRTY) {
				if (!logchange) printf("%2d  <%s>\n",il,hive[il]->filename);
				d = 1;
			}
			*/
		}
	}
	if (d) {
		/* Only prompt user if logging of changed files has not been set */
		/* Thus we assume confirmations are done externally if they ask for a list of changes */
		if (!logchange) fmyinput("Write hive files? (y/n) [n] : ",iwho,3);
		if (*iwho == 'y' || logchange) {
			if (logchange) {
				ch = fopen("/tmp/changed","w");
			}
			for (il = 0; il < no_hives; il++) {
				DIE("TODO: replace me.");
				/*
				if (hive[il]->state & HMODE_DIRTY) {
					printf("%2d  <%s> - ",il,hive[il]->filename);
					if (!writeHive(hive[il])) {
						printf("OK\n");
						if (logchange) fprintf(ch,"%s ",hive[il]->filename);
						dd = 2;
					}
				}
				*/
			}
			if (logchange) {
				fprintf(ch,"\n");
				fclose(ch);
			}
		} else {
			printf("Not written!\n\n");
		}
	} else {
		if( gverbose ) printf("None!\n\n");
	}
	return(dd);
	// }}}
#endif
#if 0
char *change_pw_orig(char *buf, int rid, int vlen, int stat) // {{{
{
	/* Decode the V-struct, and change the password
	* vofs - offset into SAM buffer, start of V struct
	* rid - the users RID, required for the DES decrypt stage
	*
	* Some of this is ripped & modified from pwdump by Jeremy Allison
	* 
	*/

	uchar x1[] = {0x4B,0x47,0x53,0x21,0x40,0x23,0x24,0x25};
	int pl;
	char *vp;
	static char username[128],fullname[128];
	char comment[128],homedir[128],md4[32],lanman[32];
	char newunipw[34], newp[20], despw[20], newlanpw[16], newlandes[20];
	char yn[4];
	int username_offset,username_len;
	int fullname_offset,fullname_len;
	int comment_offset,comment_len;
	int homedir_offset,homedir_len;
	int ntpw_len,lmpw_len,ntpw_offs,lmpw_offs,i;
	int dontchange = 0;
	struct user_V *v;

	des_key_schedule ks[2];
	des_cblock deskey1, deskey2;

	MD4_CTX context;
	unsigned char digest[16];
	unsigned short acb;

	v = (struct user_V *)buf;
	vp = buf;

	username_offset = v->username_ofs;
	username_len    = v->username_len; 
	fullname_offset = v->fullname_ofs;
	fullname_len    = v->fullname_len;
	comment_offset  = v->comment_ofs;
	comment_len     = v->comment_len;
	homedir_offset  = v->homedir_ofs;
	homedir_len     = v->homedir_len;
	lmpw_offs       = v->lmpw_ofs;
	lmpw_len        = v->lmpw_len;
	ntpw_offs       = v->ntpw_ofs;
	ntpw_len        = v->ntpw_len;

#if 0
	printf("lmpw_offs: 0x%x, lmpw_len: %d (0x%x)\n",lmpw_offs,lmpw_len,lmpw_len);
	printf("ntpw_offs: 0x%x, ntpw_len: %d (0x%x)\n",ntpw_offs,ntpw_len,ntpw_len);
#endif

	*username = 0;
	*fullname = 0;
	*comment = 0;
	*homedir = 0;

	if(
			username_len	<= 0 ||
			username_len	> vlen ||
			username_offset <= 0 ||
			username_offset >= vlen ||
			comment_len	< 0 ||
			comment_len	> vlen   ||
			fullname_len	< 0 ||
			fullname_len	> vlen ||
			homedir_offset	< 0 ||
			homedir_offset	>= vlen ||
			comment_offset	< 0 ||
			comment_offset	>= vlen ||
			lmpw_offs	< 0 ||
			lmpw_offs	>= vlen)
	{
		if (stat != 1) printf("Not a legal struct? (negative struct lengths)\n");
		return(NULL);
	}

	/* Offsets in top of struct is relative to end of pointers, adjust */
	username_offset += 0xCC;
	fullname_offset += 0xCC;
	comment_offset += 0xCC;
	homedir_offset += 0xCC;
	ntpw_offs += 0xCC;
	lmpw_offs += 0xCC;

	cheap_uni2ascii(vp + username_offset,username,username_len);
	cheap_uni2ascii(vp + fullname_offset,fullname,fullname_len);
	cheap_uni2ascii(vp + comment_offset,comment,comment_len);
	cheap_uni2ascii(vp + homedir_offset,homedir,homedir_len);

// #if 0
//	/* Reset hash-lengths to 16 if syskey has been reset */
//	if (syskeyreset && ntpw_len > 16 && !stat) {
//	ntpw_len = 16;
//	lmpw_len = 16;
//	ntpw_offs -= 4;
//	(unsigned int)*(vp+0xa8) = ntpw_offs - 0xcc;
//	*(vp + 0xa0) = 16;
//	*(vp + 0xac) = 16;
//	}
// #endif

	if (stat) {
		acb = handle_F(rid,0);
		printf("RID: %04x, Username: <%s>%s\n",
		rid, username,  (  acb & 0x8000 ? ", *disabled or locked*" : (ntpw_len < 16) ? ", *BLANK password*" : "")  );
		return(username);
	}

	printf("RID     : %04d [%04x]\n",rid,rid);
	printf("Username: %s\n",username);
	printf("fullname: %s\n",fullname);
	printf("comment : %s\n",comment);
	printf("homedir : %s\n\n",homedir);

	acb = handle_F(rid,2);

	if (lmpw_len < 16) {
		printf("** LANMAN password not set. User MAY have a blank password.\n** Usually safe to continue\n");
	}

	if (ntpw_len < 16) {
		printf("** No NT MD4 hash found. This user probably has a BLANK password!\n");
		if (lmpw_len < 16) {
			printf("** No LANMAN hash found either. Sorry, cannot change. Try login with no password!\n");
			dontchange = 1;
		} else {
			unsigned int *i = (unsigned int*)(vp+0xa8);
			printf("** LANMAN password IS however set. Will now install new password as NT pass instead.\n");
			printf("** NOTE: Continue at own risk!\n");
			ntpw_offs = lmpw_offs;
			*i  = ntpw_offs - 0xcc;
			ntpw_len = 16;
			lmpw_len = 0;
		}
	}

	if (!rid) {
		printf("No RID given. Unable to change passwords..\n");
		return(0);
	}

	if (gverbose) {
		printf("Crypted NT pw: ");
		hexprnt(stdout, (vp+ntpw_offs),16);
		printf( "Crypted LM pw: ");
		hexprnt(stdout,(vp+lmpw_offs),16);
	}

	/* Get the two decrpt keys. */
	sid_to_key1(rid,(unsigned char *)deskey1);
	des_set_key((des_cblock *)deskey1,ks[0]);
	sid_to_key2(rid,(unsigned char *)deskey2);
	des_set_key((des_cblock *)deskey2,ks[1]);

	/* Decrypt the NT md4 password hash as two 8 byte blocks. */
	des_ecb_encrypt((des_cblock *)(vp+ntpw_offs ),
		(des_cblock *)md4, ks[0], DES_DECRYPT);
	des_ecb_encrypt((des_cblock *)(vp+ntpw_offs + 8),
		(des_cblock *)&md4[8], ks[1], DES_DECRYPT);

	/* Decrypt the lanman password hash as two 8 byte blocks. */
	des_ecb_encrypt((des_cblock *)(vp+lmpw_offs),
		(des_cblock *)lanman, ks[0], DES_DECRYPT);
	des_ecb_encrypt((des_cblock *)(vp+lmpw_offs + 8),
		(des_cblock *)&lanman[8], ks[1], DES_DECRYPT);

	if (gverbose) {
		printf("MD4 hash     : ");
		hexprnt(stdout, md4,16);
		printf("LANMAN hash  : ");
		hexprnt(stdout, lanman,16);
	}

	printf("\n* = blank the password (This may work better than setting a new password!)\n");
	/*   printf("@ = promote user to administrator\n"); */
	printf("Enter nothing to leave it unchanged\n");
	pl = fmyinput("Please enter new password: ",newp,16);

	/*   printf("password: [%s] have length %d\n",newp,pl); */

	if (!pl) { printf("Nothing changed.\n"); return(0); }

	if (pl == 1 && *newp == '*') {
	printf("Blanking password!\n");
	} else if (pl == 1 && *newp == '@') {
		promote_user(rid);
		printf("Changes saved!\n");
		return(username);
	} else {

	cheap_ascii2uni(newp,newunipw,pl);

	make_lanmpw(newp,newlanpw,pl);

	/*   printf("Ucase Lanman: %s\n",newlanpw); */

	MD4Init (&context);
	MD4Update (&context, newunipw, pl<<1);
	MD4Final (digest, &context);

	if (gverbose) {
		printf("\nNEW MD4 hash    : ");
		hexprnt(stdout,digest,16);
	}

	E1(newlanpw,   x1, lanman);
	E1(newlanpw+7, x1, lanman+8);

	if (gverbose) {
		printf("NEW LANMAN hash : ");
		hexprnt(stdout,lanman,16);
	}

	/* Encrypt the NT md4 password hash as two 8 byte blocks. */
	des_ecb_encrypt((des_cblock *)digest,
	     (des_cblock *)despw, ks[0], DES_ENCRYPT);
	des_ecb_encrypt((des_cblock *)(digest+8),
	     (des_cblock *)&despw[8], ks[1], DES_ENCRYPT);

	des_ecb_encrypt((des_cblock *)lanman,
	     (des_cblock *)newlandes, ks[0], DES_ENCRYPT);
	des_ecb_encrypt((des_cblock *)(lanman+8),
	     (des_cblock *)&newlandes[8], ks[1], DES_ENCRYPT);

	if (gverbose) {
		printf("NEW DES crypt   : ");
		hexprnt(stdout,despw,16);
		printf("NEW LANMAN crypt: ");
		hexprnt(stdout,newlandes,16);
	}

	} /* blankit check */

	fmyinput("\nDo you really wish to change it? (y/n) [n] ",yn,2);

	if (*yn == 'y') {
		if (pl == 1 && *newp == '*') {
			/* Setting hash lengths to zero seems to make NT think it is blank
			* However, since we cant cut the previous hash bytes out of the V value
			* due to missing resize-support of values, it may leak about 40 bytes
			* each time we do this.
			*/
			v->ntpw_len = 0;
			v->lmpw_len = 0;
		} else if (pl > 1 && *newp != '@') {
			if (!dontchange) {
				/* Reset hash length to 16 if syskey enabled, this will cause
				* a conversion to syskey-hashes upon next boot */
				if (syskeyreset && ntpw_len > 16) { 
					 unsigned int *i = (unsigned int *)(vp+0xa8);
					ntpw_len = 16;
					lmpw_len = 16;
					ntpw_offs -= 4;
					*i = ntpw_offs - 0xcc;
					*(vp + 0xa0) = 16;
					*(vp + 0xac) = 16;
				}

				for (i = 0; i < 16; i++) {
					 unsigned char *cfm = (unsigned char*)(vp+ntpw_offs+i);
					*cfm = despw[i];
					if (lmpw_len >= 16) *cfm = newlandes[i];
				}
			} else {
				printf("Unable to set since it is blank.\n");
			}
		}

//#if 0
		hexprnt("Pw in buffer: ",(vp+ntpw_offs),16);
		hexprnt("Lm in buffer: ",(vp+lmpw_offs),16);
//#endif
		dirty = 1;
		printf("Changed!\n");
	} else {
		printf("Password not changed.\n");
	}

	printf("\n");
	return(username);
} // }}}
#endif

#if 0
short handle_F(int rid, int mode) // {{{
{
	/* Try to decode and possibly change account lockout etc
	 * This is \SAM\Domains\Account\Users\<RID>\F
	 * It's size seems to always be 0x50.
	 * Params: RID - user ID, mode - 0 silent, 1 silent, 2 edit.
	 * Returns: ACB bits with high bit set if lockout count is >0
	 */
	struct varbuf *ufbuf;
	struct uf *userf;

	if( NULL == ( userf = sam_get_user_F( rid, ufbuf ) ) ) {
		//TODO: error
		printf("handle_F(): Couldn't read user F record.\n" );
		return -1;
	}

	if (mode) {
		acb_print( stdout, userf->acb );

		fprintf(stdout, "\nFailed login count: %u, while max tries is: %u\n",
		userf->bad_login_count,max_sam_lock);
		fprintf(stdout, "Total  login count: %u\n",userf->no_login);

		if (mode > 1) acb_edit( userf );
	}

	return userf->acb;

	/* The following seems to only set fields in the ACB that have no known effect (???)
	return (userf->acb |
		(
			(userf->bad_login_count > 0 && userf->bad_login_count >= max_sam_lock)<<15
		) |
		(userf->acb & ACB_AUTOLOCK)<<15 |
		(userf->acb & ACB_DISABLED)<<15);
		*/
} // }}}
#endif

//TODO: this should be trash soon
long get_realloc( rl_hkey hk, char *path, long *type, char **buf, long *blen ) // {{{ realloc() buf (if needed), set blen to new size and retrieve value into buffer
{
	//TODO: return errors rather than crashing and burning
	long nblen;
	ASSERT( rlERROR_MORE_DATA == rlRegQueryValueEx( hk, path, type, NULL, &nblen ), "This is a bad error string" );
	if( nblen < *blen ) { *buf = F_REALLOC( *buf, *blen ); }
	*blen = nblen;
	ASSERT( rlERROR_SUCCESS == rlRegQueryValueEx( hk, path, type, *buf, blen ), "This is another bad error string" );

	return nblen;
} // }}}
// }}}
// vim600: set foldlevel=0 foldmethod=marker:
