/* {{{ preamble
 * proftool.c
 *
 * Profile migration tool.  Finds a user's SID-RID, replaces old user SID-RIDs
 * in user registry hive.
 *
 * Author: Sean Loaring
 *
 * Copyright (c) 2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */
// {{{ includes
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>

#include "../lib/libhive.h" //win32-style registry interface
#include "../hivetools/nstdreg.h" //my attempt at a posix-like interface to libhive
#include "../hivetools/sam.h" // sam-specific functions (set user password, etc)
#include "../misc/tools.h" //
#include "../misc/uerr.h" // like errno and strerror...kindof
// }}}

#define TMODE_USAGE 0
#define TMODE_ULIST 1
#define TMODE_ULIST_LONG 2
#define TMODE_SLIST 3

// There are a whole bunch of pre-defined accounts and groups
// see here: http://www.beginningtoseethelight.org/ntsecurity/index.php
#define RID_ADMINISTRATOR 	0x01f4
#define RID_GUEST		0x01f5
#define RID_DOMAIN_USER		0x0201

static int mode = TMODE_USAGE;

// {{{ forward defs
int usage(); 
int parse_args( int *argc, char ***argv ); 
int load_hives( int argc, char **argv );
int list_software();
int list_users();
int list_users_long();
// }}}
struct binsid {
	char typeid[12];
	char machineid[12];
	uint32_t rid;
};
int dump_binsid( vbuf *out, struct binsid *sidrid ) // {{{
{
	vbuf_catprintf( out, "[" );
	vbuf_cathex( out, sidrid->typeid, sizeof( sidrid->typeid ) );
	vbuf_catprintf( out, "] [" );
	vbuf_cathex( out, sidrid->machineid, sizeof( sidrid->machineid ) );
	vbuf_catprintf( out, "] [" );
	vbuf_catprintf( out, "%08lx", sidrid->rid );
	//vbuf_cathex( out, sidrid->rid, sizeof( sidrid->rid ) );
	vbuf_catprintf( out, "]" );

	return 0;
} // }}}
int main(int argc, char **argv ) // {{{
{
	char *uname;
	char *path;
	int rid;
	struct uv *userv;
	VBUF_STATIC( uvbuf, 10 );
	VBUF_STATIC( mvbuf, 10 );
	VBUF_STATIC( vbdump, 10 );
	VBUF_STATIC( ntdatpath, 10 );
	struct sam_v *mv;
	struct v_perms_pent *ne;
	FILE *ntuserdat = NULL;
	struct stat statbuf;
	VBUF_STATIC( ntdbuf, 10 );
	char sidridprefix[] = { 0x01, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x15, 0x00, 0x00, 0x00 };
	//01 05 00 00 00 00 00 05 15 00 00 00 81 77 D9 74 8A 5A 41 60 82 8B A6 28 EC 03 00 00 00 00 18 00
	int sidlen = 28;
	struct binsid localsid;
	// machine == 12
	// RID = 4
	//
	// pref == 12
	//
	//
	// 12 + 16: 28


	//init
        unicode_init();
        libhive_init();
        ns_init();

	if( argc < 3 ) {
		usage();
		exit(1);
	}

	uname = argv[1];
	path = argv[2];

	vbuf_strset( ntdatpath, path );
	vbuf_strcat( ntdatpath, "/ntuser.dat");

	argc -= 3;
	argv += 3;

	if( 0 != load_hives( argc, argv ) ) { 
		fprintf(stderr, "Error loading hives: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	rid = sam_uname2rid( uname );

	if( 0 > rid) {
		fprintf(stderr, "Error reading user RID: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	/*
	ASSERT( NULL != sam_get_machine_V(mvbuf), "Bah: %s", uerr_str( uerr_get() ) );

	vbuf_strset( vbdump, "" );
	if( 0 != sam_vmachine_dump( mvbuf, vbdump ) ) {
		fprintf(stderr, "Error reading user V struct: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	fprintf(stderr, "Dumping Machine V Struct:\n%s\n", vbdump->b );
	*/

	if( NULL == (userv = sam_get_user_V( rid, uvbuf ) ) ) {
		fprintf(stderr, "Error reading user V struct: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	if( NULL == ( ne = sam_v_get_perm( uvbuf, 1, 0 ) ) ) {
		fprintf(stderr, "Couldn't read user SID-RID: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}


	vbuf_strset( vbdump, "" );
	vbuf_cathex( vbdump, ne->data, ne->len - 4 );

	fprintf(stderr, "Here it is [%d]:\n%s\n", ne->len, vbdump->b );

	if( ne->len < sizeof( struct binsid ) ) {
		fprintf(stderr, "Attempted buffer overrun. :P, %d < %d\n", ne->len, sizeof( struct binsid ) );
		goto MAINERR;
	}

	memcpy( &localsid, ne->data, sizeof( localsid ) );

	fprintf(stderr, "Time for file munging %s!\n", ntdatpath->b );

	if( -1 == stat( ntdatpath->b, &statbuf ) ) {
		fprintf(stderr, "Couldn't stat file \"%s\": %s\n", ntdatpath->b, strerror( errno ) );
		goto MAINERR;
	}

	fprintf(stderr, "stat worked on %s\n", ntdatpath->b );

	if( NULL == ( ntuserdat = fopen( ntdatpath->b, "r+b" ) ) ) {
		fprintf(stderr, "Couldn't open file \"%s\": %s\n", ntdatpath->b, strerror( errno ) );
		goto MAINERR;
	}


	vbuf_grow( ntdbuf, statbuf.st_size );

	fprintf(stderr, "Reading %d bytes.\n", statbuf.st_size );

	if( 1 != fread( ntdbuf->b, statbuf.st_size, 1, ntuserdat ) ) {
		fprintf(stderr, "Error loading file: %s\n", strerror( errno ) );
		goto MAINERR;
	}

	ntdbuf->dlen = statbuf.st_size;

	fprintf(stderr, "File loaded, seeking SID-RIDs\n");

	char *p = ntdbuf->b;
	while( NULL != ( p = (char*)memmem( p, ntdbuf->dlen - (p - ntdbuf->b), sidridprefix, sizeof( sidridprefix ) ) ) ) {
		struct binsid *bsid = (struct binsid *)p;
		printf("Found at 0x%x\n", (p - ntdbuf->b) );

		switch( bsid->rid ) {
			case RID_ADMINISTRATOR:
			case RID_DOMAIN_USER:

				//Keep the user bits, replace the other stuff

				printf("------------\nSwap SID Only\n");
				vbuf_strset( vbdump, "");
				dump_binsid( vbdump, bsid );
				printf("BINSID: %s\n", vbdump->b );

				memcpy( bsid, &localsid, sizeof( struct binsid ) - sizeof( bsid->rid ) );

				printf("Swap SID-RID\n");
				vbuf_strset( vbdump, "");
				dump_binsid( vbdump, bsid );
				printf("BINSID: %s\n", vbdump->b );

				break;
			default:

				printf("------------\nSwap SID-RID\n");
				vbuf_strset( vbdump, "");
				dump_binsid( vbdump, bsid );
				printf("BINSID: %s\n", vbdump->b );
				memcpy( bsid, &localsid, sizeof( struct binsid ) );
				printf("Swap SID-RID\n");
				vbuf_strset( vbdump, "");
				dump_binsid( vbdump, bsid );
				printf("BINSID: %s\n", vbdump->b );
				/*
				dump_binsid( vbdump, bsid );
				printf("BINSID: %s\n", vbdump->b );
				*/
				break;
		}



		//hexdump( p, 0, sidlen, 1 );
		p+=1;
	}

	rewind( ntuserdat );

	if( 1 != fwrite( ntdbuf->b, ntdbuf->dlen, 1, ntuserdat ) ) {
		fprintf(stderr, "Write error: %s\n", strerror( errno ) );
		goto MAINERR;
	}
	fclose( ntuserdat );
	ntuserdat = NULL;

	fprintf(stderr, "Done.\n");

	/*
	vbuf_strset( vbdump, "" );
	if( 0 != sam_vuser_dump( uvbuf, vbdump ) ) {
		fprintf(stderr, "Error reading user V struct: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}


	fprintf(stderr, "Dumping User V Struct:\n%s\n", vbdump->b );

	//hexdump( userv, 0, uvbuf->dlen, 1 );

	if( NULL == userv ) {
		fprintf(stderr, "Error reading user V struct: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}
	*/

	return 0;

MAINERR:

	if( ntuserdat ) fclose( ntuserdat );
	return 1;
} // }}}
int parse_args( int *argc, char ***argv ) // {{{ parse leading arguments
{
        extern int optind;
        extern char* optarg;
        char c;

        char *options = "hlLsD:";

        while((c=getopt(*argc,*argv,options)) > 0) {
                switch(c) {
                        case 'h': mode = TMODE_USAGE;      break;
                        case 'l': mode = TMODE_ULIST;      break;
                        case 'L': mode = TMODE_ULIST_LONG; break;
                        case 's': mode = TMODE_SLIST;      break;
			case 'D':
			{
				int dbval;
				if( 1 != sscanf(optarg, "%d", &dbval ) ) {
					fprintf(stderr, "Couldn't set debug level to \"%s\", is it an integer?\n", optarg );
					exit(1);
				}
				DEBUG_LEVEL = dbval;
				break;
			}
                        default: 
				  fprintf(stderr, "Unrecognized option %c.  Run with -h for help.\n", c);
				  exit(1);
                }
        }

	*argc -= optind;
	*argv += optind;

	return 0;
} // }}}
int usage() // {{{
{
	fprintf(stderr,
		"!!! THIS PROGRAM IS EXPERIMENTAL !!!\n"
		"Backup your ntuser.dat file before you run this!\n\n"
		"Usage: proftool user_name user_profile_path\n"
		"       - Updates access control in a user hive file.\n\n"
        );

	//fprintf(stderr, "%s", hl_mount_help() );

	/*
	fprintf(stderr, 
                "See readme file on how to extract/read/write the NT's SAM file\n"
                "if it's on an NTFS partition!\n");
		*/

	return 0;
} // }}}
int load_hives( int argc, char **argv ) // {{{ 
{
//        if( !argv[optind] ) {
//                fprintf(stderr, "No hive files specified.\n");
//                return 1;
//        }

	//HERE
        if( 0 != hl_mount_list( argc, argv  ) ) {
                fprintf(stderr, "There was a problem loading the given hives: %s\n", uerr_str( uerr_get() ) );
                return 1;
	}

	return 0;
} // }}}
int list_software() // {{{
{
	NREGDIR* sh;
        struct ns_dirent *d;
        char *path = "/HKEY_LOCAL_MACHINE/software";

        if( NULL == ( sh = ns_opendir( path) ) ) {
		fprintf(stderr, "Error opening key: %s, %s\n", path, uerr_str( uerr_get() ) );
                return -1;
        }

        while( ( d = ns_readdir( sh ) ) ) {
                printf("%s\n", d->filename);
        }

        ns_closedir( sh );

	return 0;
} // }}}
int list_users() // {{{
{
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int *ridlist;
        int i;

        if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error reading user RID list: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

        ridlist = (int*)vb->b;

        for( i=0; i<vb->dlen; i++ ) {
                if( 0 == sam_get_user_name( nbuf, ridlist[i] ) ) {
                        printf("%08x %s\n", ridlist[i], nbuf->b );
                }
                else {
                        fprintf(stderr, "Error reading data for RID %x: %s\n", ridlist[i], uerr_str( uerr_get() ) );
                }
        }

	return 0;
} // }}}
int list_users_long() // {{{
{

	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int *ridlist;
        int i;

        if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error reading user RID list: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

        ridlist = (int*)vb->b;

	printf("Excessivly verbose user list:\n");
        for( i=0; i<vb->dlen; i++ ) {
		//TODO: return value
		nbuf->dlen = 0;
		if( 0 != sam_print_user( nbuf, ridlist[i] ) ) {
			fprintf(stderr, "Error printing user data for RID %x: %s\n", ridlist[i], uerr_str( uerr_get() ) );
			return -1;
		}
		printf("------------------------------------------------\n%s", nbuf->b );
        }

	return 0;
} // }}}

// vim600: set foldlevel=0 foldmethod=marker:
