/* {{{ preamble
 * nstdreg.h
 *
 * NonStandardRegistry
 *
 * POSIX-like routines that provide access to the registry via rlReg* functions
 *
 * Authors:	Sean Loaring
 *
 * Copyright (c) 2005-2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#ifndef NSTDREG_H
#define NSTDREG_H

#include "../lib/libhive.h"

#define PATH_MAX 1024

#define RTYPE_VALUE     0
#define RTYPE_KEY       1
#define RTYPE_NOT_FOUND	2
#define RTYPE_FAKEKEY	3

#define NSR_FLAG_PARENT_LAST 	1
#define NSR_FLAG_DITCH_ON_FAIL  2

#define DOCHECK
//TODO:move internal functions to the ifdef DOCHECK block and then define only when doing test stuff

struct ns_dirent {
	char filename[PATH_MAX];
	unsigned char type;
};
//#define PATH_RELATIVE 1
//#define PATH_ABSOLUTE 2
//#define PATH_MAX 256
//
//struct parsdpath {
//	char type;      //PATH_RELATIVE || PATH_ABSOLUTE
//	int n;          //no. of elements
//	char *e[PATH_MAX] //path elements
//};*

extern int ns_errno;

typedef struct ns_dir_handle {
	//struct parsdpath *p;
	//char path[PATH_MAX];
	vbuf *vb_path;
	int rdidx; //index for ns_readdir
	int rd_donekeys;
	rl_hkey hk;
} NREGDIR;


typedef struct ns_handle {
	//struct parsdpath *p;
	char path[PATH_MAX];
	ulong type;
	ulong blen;
	char *buf;
	rl_hkey parent;
} NREGH;

void ns_init();

char *ns_dirnameA( char *p, char delim );
void ns_vdirname( vbuf *v, char *p );
void ns_vbasename( vbuf *vb, char *path ); 
int vb_process_path( vbuf *vb, char *path );
void ns_pairpath( vbuf *dn, vbuf *bn, char *path ); // in: a dirty path, out: dn=dirname, bn=basename 
int rfinddelim( char *p, char delim );
int ns_finddelim( char *p, char delim );

int ns_recurse( char *base, char *rel, int (*f)(char *base, char *rel, long type, void *opt ), int flags, void *opt );

int ns_regp_val( vbuf *out, char *path ); 

/* Relatives are no more
NREGH* ns_open_relative( NREGDIR* h, char *path ); //just like ns_open, but *path can be relative to h
NREGDIR* ns_opendir_relative( NREGDIR* h, char *path ); //just like ns_open, but *path can be relative to h

int ns_unlink_relative( NREGDIR* h, char *path ); // just like ns_unlink, but *path can be relative to h

int ns_set_value_relativeT( NREGDIR *h, char *path, int type, char *value, int len );

int ns_mkdir_relative( NREGDIR *h, char *path );
*/

NREGDIR* ns_opendir( char *path );	//open an ns handle
void ns_closedir( NREGDIR *h );		//close a ns handle

char *ns_getpath( NREGDIR *h );

int ns_chdir( char *path );
char *ns_getcwd();
//char *ns_getpath( NREGDIR *h );

NREGH* ns_open( char *path );	//open an ns handle
void ns_close( NREGH *h );		//close a ns handle

int ns_path_ns2regt( vbuf *dest, char *src, int t ); // TODO: for test purposes only 

int ns_path_element_reg2ns( vbuf *dest, char *src );
int ns_path_ns2regvalname( vbuf *dest, char *src ); // in: ns basename; out: reg value name
int ns_path_ns2regkeypath( vbuf *dest, char *src ); // in: ns dirname; out: reg key path
int ns_path_regkey2ns( vbuf *dest, char *str );

//(call it cwd?  It's tied to the directory.
char *ns_get_cwd( NREGDIR *h );	//get the cwd.  Statically allocated per handle, overwritten on subsequent calls

//int ns_gettype( NREGDIR *dir, char *name );
int ns_gettype( char *path );

int ns_cd( char *path );	//attempt to change cwd, will try to work with relative paths, etc.

rl_hkey ns_get_hkey( NREGDIR *h );	//return the hkey associated with cwd (if one exists)

int ns_unlink( char *path );
int ns_set_valueT( char *path, int type, char *value, int len );
int ns_mkdir( char *path, int flags );

#define MKDIR_P		1
char *ns_strerror( int no );

int ns_read_valueT( char *path, unsigned long *type, struct varbuf *vb );

//print a value to stdout in regedit style
void ns_print_val( FILE *out, char *name, unsigned long type, vbuf *val );

struct ns_dirent *ns_readdir( NREGDIR *h );

void ns_rewinddir( NREGDIR *h );

int ns_read_dword( char *path, int *out );


typedef (*ns_export_callback)( char *b, int len );
int ns_reg_export( char *path, char *cr, ns_export_callback cb );
int ns_reg_merge( char *path );

#endif


#ifdef DOCHECK
int path_end_is_delim( char *p );
#endif

// vim600: set foldlevel=0 foldmethod=marker:
