/* {{{ preamble
 * sam.c
 *
 * Structures and related functions for dealing with structures in the SAM hive
 * of NT registry
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <openssl/des.h>
#include <openssl/md4.h>
#include "sam.h"
#include "hivetools.h"
#include <errno.h>
#include "../misc/tools.h"
#include "../misc/uerr.h"
#include "../lib/regtypes.h"
#include "nstdreg.h"

char *acb_fields[16] = { // {{{
	"Disabled" ,
	"Homedir req." ,
	"Passwd not req." ,
	"Temp. duplicate" ,
	"Normal account" ,
	"NMS account" ,
	"Domain trust act." ,
	"Wks trust act." ,
	"Srv trust act" ,
	"Pwd don't expire" ,
	"Auto lockout" ,
	"(unknown 0x08)" ,
	"(unknown 0x10)" ,
	"(unknown 0x20)" ,
	"(unknown 0x40)" ,
	"(unknown 0x80)" ,
}; // }}}

char *uf_fields[] = { // {{{
	"unknown1",
	"last_login",
	"unknown2",
	"last_pw_change",
	"account_expires",
	"last_invalid_pass",
	"rid",
	"unknown3",
	"acct_flags",
	"password_flags",
	"unknown4",
	"country_code",
	"unknown5",
	"bad_login_count",
	"no_login",
	"unknown6",
	"unknown7",
	""
}; // }}}

// {{{ Registry types (REG_SZ, REG_BINARY, etc) TODO: which SHOULD NOT BE IN THIS FILE
#define REG_NONE                0       /* no type */
#define REG_SZ                  1       /* string type (ASCII) */
#define REG_EXPAND_SZ           2       /* string, includes %ENVVAR% (expanded by caller) (ASCII) */
#define REG_BINARY              3       /* binary format, callerspecific */
	/* YES, REG_DWORD == REG_DWORD_LITTLE_ENDIAN */
#define REG_DWORD               4       /* DWORD in little endian format */
#define REG_DWORD_LITTLE_ENDIAN 4       /* DWORD in little endian format */
#define REG_DWORD_BIG_ENDIAN    5       /* DWORD in big endian format  */
#define REG_LINK                6       /* symbolic link (UNICODE) */
#define REG_MULTI_SZ            7       /* multiple strings, delimited by \0, terminated by \0\0 (ASCII) */
#define REG_RESOURCE_LIST       8       /* resource list? huh? */
#define REG_FULL_RESOURCE_DESCRIPTOR    9       /* full resource descriptor? huh? */
#define REG_RESOURCE_REQUIREMENTS_LIST  10
#define REG_QWORD               11      /* QWORD in little endian format */
#define REG_QWORD_LITTLE_ENDIAN 11      /* QWORD in little endian format */
// }}}
#if 0
short load_user_F(int rid, int mode) // {{{ Read a \SAM\Domains\Account\Users\<RID>\F structure
{
	/* Try to decode and possibly change account lockout etc
	 * This is \SAM\Domains\Account\Users\<RID>\F
	 * It's size seems to always be 0x50.
	 * Params: RID - user ID, mode - 0 silent, 1 silent, 2 edit.
	 * Returns: ACB bits with high bit set if lockout count is >0
	 */
	struct user_F *f;
	char s[200];
	//char yn[10];
	//struct keyval *v;
	char buf[USER_F_SIZE];
	ulong buf_len = USER_F_SIZE;
	unsigned short acb;
	int b;
	//rl_hkey hkey;
	ulong type;

	if (H_SAM < 0) return(0);

	/* Get users F value */
	snprintf(s,180,"\\SAM\\Domains\\Account\\Users\\%08X\\F",rid);
	//v = get_val2buf(hive[H_SAM], NULL, 0, s, REG_BINARY);
	if( 0 != rlRegQueryValueEx( HKEY_LOCAL_MACHINE, s, &type, buf, &buf_len )) {
		fprintf(stderr, "load_user_F: couldn't find key %s\n", s );
		return -1;
	}

	if( type != REG_BINARY ) {
		fprintf(stderr, "load_user_F: got invalid type %ld\n", type );
		return -1;
	}

	if (buf_len != USER_F_SIZE) {
		printf("handle_F: F value is 0x%lx bytes, not 0x50, unable to check account flags!\n",buf_len);
		return(0);
	}

	f = (struct user_F *)buf;
	acb = f->ACB_bits;

	if (mode) {
		printf("Account bits: 0x%04x =\n",acb);


		for (b=0; b < 15; b++) {
			printf("[%s] %-15.15s | ",
			(acb & (1<<b)) ? "X" : " ", acb_fields[b] );
			if (b%3 == 2) printf("\n");
		}

		printf("\nFailed login count: %u, while max tries is: %u\n",
		f->failedcnt,max_sam_lock);
		printf("Total  login count: %u\n",f->logins);

		/*
		if (mode > 1) {
			if ( acb & (ACB_DISABLED|ACB_AUTOLOCK) || (f->failedcnt > 0 && f->failedcnt >= max_sam_lock)  ) {
				printf("Account is %s\n",(acb & ACB_DISABLED) ? "disabled" : "probably locked out!");
				printf("Do you wish me to reset the failed count, unset disabled and lockout,\n");
				fmyinput("and set the \"password never expires\" option? (y/n) [n]",yn,2);
				if (*yn == 'y') {
					acb |= ACB_PWNOEXP;
					acb &= ~ACB_DISABLED;
					acb &= ~ACB_AUTOLOCK;
					f->ACB_bits = acb;
					f->failedcnt = 0;
					put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY);
					printf("Unlocked!\n");
				}
			}
		}
		*/
	}

	return (acb | ( (f->failedcnt > 0 && f->failedcnt >= max_sam_lock)<<15 ) |
	(acb & ACB_AUTOLOCK)<<15 | (acb & ACB_DISABLED)<<15);
} // }}}
#endif

/* THIS STUFF SHOULD BE IN SAM.{HC} */
#define USER_NAME_PATH "\\SAM\\SAM\\Domains\\Account\\Users\\Names"

#define NS_USER_NAME_PATH "/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/Account/Users/Names"
#define NS_USER_PATH "/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/Account/Users"
#define NS_PATH_MACHINE_V "/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/Account/V"


// SAM ACCESS
ulong sam_no_users() // {{{
{
	//TODO: use nstdreg for this crap
	ulong result;
	rl_hkey rl;

	NREGDIR *sh;
       
	if( NULL == ( sh = ns_opendir( NS_USER_NAME_PATH ) ) ) 
		return -1;

	/*
	WARN("Here %s", USER_NAME_PATH);
	if( rlERROR_SUCCESS != rlRegOpenKeyEx( rlHKEY_LOCAL_MACHINE, USER_NAME_PATH, 0, KEY_ALL_ACCESS, &rl) )
		return -1;
	*/

	if( NULL == ( rl = ns_get_hkey(sh) ) ) {
	}
	if( rlERROR_SUCCESS != rlRegQueryInfoKey( rl, NULL, NULL, &result, NULL, NULL, NULL, NULL, NULL, NULL, NULL) )
		return -1;

	ns_closedir( sh );

	return result;
} // }}}
rl_hkey sam_get_users_hkey() // {{{
{
	rl_hkey rl;

	if( rlERROR_SUCCESS != rlRegOpenKeyEx( rlHKEY_LOCAL_MACHINE, USER_NAME_PATH, 0, KEY_ALL_ACCESS, &rl) )
		return NULL;

	return rl;
} // }}}
int uv_get_username( vbuf *out, struct uv *userv ) // {{{
{
	VBUF_STATIC( raw, 100 );
	char *buf;
	int blen;
	if( NULL == ( buf = uv_get_field_ptr( userv, UV_USERNAME, &blen ) ) ) 
		return -1;

	vbuf_set( raw, buf, blen );
	vbuf_grow( raw, 2 );
	raw->b[ raw->dlen++ ] = '\0';
	raw->b[ raw->dlen++ ] = '\0';
	vb_utf16to8( out, raw->b, raw->dlen);

	return 0;

} // }}}
int sam_get_user_name( struct varbuf *out, int irid ) { // {{{ 
	VBUF_STATIC( uvbuf, 100 );
	VBUF_STATIC( uname, 100 );
	int blen;
	struct uv *userv;

	if( NULL == ( userv = sam_get_user_V( irid, uvbuf ) ) ) return -1;
	if( 0 != uv_get_username( out, userv ) ) return -1;

	return 0;

//	int index;
//	ulong type;
//	char name[SZ_MAX];
//	ulong len = SZ_MAX - 1;
//	char keystr[9];
//
//	snprintf(keystr, 9, "%08d", irid );
//
//	rl_hkey rl;
//
//	rl = sam_get_users_hkey();
//
//	for(	index = 0;
//		(rlERROR_SUCCESS == rlRegEnumKeyEx(rl, index, name, &len, NULL, NULL, NULL));
//		index++
//	   ) {
//
//		ulong rid;
//		ulong rid_len = sizeof( rid );
//
//		ASSERT( rlERROR_SUCCESS == rlRegQueryValueEx( rl, name, &type, (void*)&rid, &rid_len ) , "rlRegQueryValueEx error.");
//		ASSERT( rid_len == 4, "sam_get_user_name(): rid_len is looking strange." );
//
//		//fprintf(stderr, "RIDIN: %x, RID: %lx\n", irid, rid );
//
//		if( irid == rid ) {
//			vbuf_resize( vb, len );
//			strncpy( vb->b, name, len);
//		}
//
//		len = SZ_MAX - 1;
//	}
//
//	return 0;
} // }}}
int sam_user_listA( struct varbuf *vb ) { // {{{ populate vb with array of RIDs (dword user IDs)
	int index;
	ulong type;
	char name[SZ_MAX];
	rl_hkey rl;
	int no_users;
	int *ridlist;

	NREGDIR *dir = NULL;

	if( NULL == ( dir = ns_opendir( NS_USER_NAME_PATH ) ) ) {
		return -1;
	}

	if( 0 > ( no_users = sam_no_users() ) ) {
		return -1;
	}

	vbuf_resize( vb, no_users * sizeof( int ) );

	vb->dlen = no_users;

	ridlist = (int*)vb->b;

	ulong len = SZ_MAX - 1;

	if( NULL == (rl = sam_get_users_hkey() ) )
		return -1;

	for(	index = 0;
		(rlERROR_SUCCESS == rlRegEnumKeyEx(rl, index, name, &len, NULL, NULL, NULL));
		index++
	   ) {

		ulong rid;
		ulong rid_len = sizeof( rid );

		ASSERT( index < no_users, "sam_user_list(): index is too big." );

		ASSERT( rlERROR_SUCCESS == rlRegQueryValueEx( rl, name, &type, (void*)&rid, &rid_len ) , "rlRegQueryValueEx error.");
		ASSERT( rid_len == 4, "sam_user_list(): rid_len is looking strange." );

		ridlist[index] = rid;

		len = SZ_MAX - 1;
	}

	return 0;
} // }}}
int sam_user_list( struct varbuf *vb ) { // {{{ populate vb with array of RIDs (dword user IDs)
	int index;
	ulong type;
	rl_hkey rl;
	int no_users;
	ulong *ridlist;
	VBUF_STATIC( path, 100 );

	NREGDIR *dir = NULL;
	struct ns_dirent *de;

	//WARN("I am here.");
	if( 0 > ( no_users = sam_no_users() ) ) {

		WARN("I am here2.");
		return -1;
	}

	vbuf_resize( vb, no_users * sizeof( int ) );
	vb->dlen = no_users;

	ridlist = (ulong*)vb->b;

	ulong len = SZ_MAX - 1;

	//WARN("Cont...");
	if( NULL == ( dir = ns_opendir( NS_USER_NAME_PATH ) ) )
		return -1;

	index = 0;
	while( de = ns_readdir( dir ) ) {
		int type;

		//WARN(" Cont...");
		if( de->type != RTYPE_KEY ) {
			WARN( "Invalid type reading user rid list." );
			continue;
		}
		//TODO: It might be good to check the RID validity here. :P
		vbuf_printf( path, "%s/%s/\\0", NS_USER_NAME_PATH, de->filename );
		if( 0 != ns_read_valueT( path->b, &ridlist[index], NULL ) ) return -1;
		index++;
		if( index >= no_users ) break;
	}

	if( index < no_users - 1 ) {
		ERR_UNIX( EINVAL, "Couldn't not load correct number of users.");
		return -1;
	}

	return 0;
} // }}}
struct samkeyf *sam_get_samkeyf( struct varbuf *vb ) // {{{
{
	unsigned long type;
	if( 0 != ns_read_valueT("/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/Account/F", &type, vb ) ) return NULL;

	if( type != REG_BINARY ) {
		ERR_UNIX(EILSEQ, "sam f key is not REG_BINARY type...?");
		return NULL;
	}

	return (struct samkeyf*)vb->b;
} // }}}
struct secpoldata *sam_get_secpoldata( struct varbuf *vb ) // {{{
{
	DIE("STUB");
	return NULL;
} // }}}
int sam_set_lm_password( int rid, char *pass ) // {{{
{
	char newlandes[20];
	int len = strlen( pass );
	int blen;
	char *lmdata;

	struct uv *userv;
	static struct varbuf *uvbuf;
	if( NULL == uvbuf) uvbuf = vbuf_alloc(10);
	if( NULL == ( userv = sam_get_user_V( rid, uvbuf ) ) ) return -1;
	if( NULL == ( lmdata = uv_get_field_ptr( userv, UV_LMPW, &blen ) ) ) return -1;

	if( 0 == len ) {
		if( 0 != uv_set_field( uvbuf, UV_LMPW, NULL, 0 ) ) return -1;
	}
	else {
		if( 0 != lmhash_make( rid, newlandes, pass ) ) return -1;
		if( 0 != uv_set_field( uvbuf, UV_LMPW, newlandes, 16 ) ) return -1;
	}

	if( 0 != sam_set_user_V( rid, uvbuf ) ) return -1;

	return 0;
} // }}}
int sam_set_nt_password( int rid, char *pass ) // {{{
{
	char newntdes[20];
	int len = strlen( pass );
	int blen;
	char *ntdata;

	static struct varbuf *strbuf;
	if( NULL == strbuf) strbuf = vbuf_alloc(10);

	struct uv *userv;
	static struct varbuf *uvbuf;
	if( NULL == uvbuf) uvbuf = vbuf_alloc(10);
	if( NULL == ( userv = sam_get_user_V( rid, uvbuf ) ) ) return -1;
	if( NULL == ( ntdata = uv_get_field_ptr( userv, UV_NTPW, &blen ) ) ) return -1;

	if( 0 == len ) {
		if( 0 != uv_set_field( uvbuf, UV_NTPW, NULL, 0 ) ) return -1;
	}
	else {
		if( 0 != nthash_make( rid, newntdes, pass ) ) return -1;
		if( 0 != uv_set_field( uvbuf, UV_NTPW, newntdes, 16 ) ) return -1;
	}

	if( 0 != sam_set_user_V( rid, uvbuf ) ) return -1;

	return 0;
} // }}}
// Misc
void samkeyf_dump( vbuf *vb, struct samkeyf *sf ) // {{{
{
	vbuf_catprintf( vb, "%-20s \n", "unknown1"); hexdump(sf->unknown1, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "unknown2"); hexdump(sf->unknown2, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "autoinc"); hexdump(sf->autoinc, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "maxage"); hexdump(sf->maxage, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "minage"); hexdump(sf->minage, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "unknown3"); hexdump(sf->unknown3, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "lockoutage"); hexdump(sf->lockoutage, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "lockoutreset"); hexdump(sf->lockoutreset, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s \n", "unknown"); hexdump(sf->unknown, 0, 0x8, 1);
	vbuf_catprintf( vb, "%-20s %x %d\n", "next_rid ", sf->next_rid , sf->next_rid );
	vbuf_catprintf( vb, "%-20s %x %d\n", "flags", sf->flags, sf->flags);
	vbuf_catprintf( vb, "%-20s %x %d\n", "min_pass", sf->min_pass, sf->min_pass);
	vbuf_catprintf( vb, "%-20s %x %d\n", "enforce_password_history", sf->enforce_password_history, sf->enforce_password_history);
	vbuf_catprintf( vb, "%-20s %x %d\n", "lockout_threshold", sf->lockout_threshold, sf->lockout_threshold);
	vbuf_catprintf( vb, "%-20s %x %d\n", "lockout_threshold", sf->lockout_threshold, sf->lockout_threshold);
	vbuf_catprintf( vb, "%-20s %x %d\n", "unknown4", sf->unknown4, sf->unknown4);
	vbuf_catprintf( vb, "%-20s \n", "unknown5"); hexdump(sf->unknown5, 0, 0x18, 1);
	vbuf_catprintf( vb, "%-20s \n", "obf"); hexdump(sf->obf, 0, 0x30, 1);

} // }}}
void secpoldata_dump( vbuf *vb, struct secpoldata *sd ) // {{{
{
	vbuf_strset(vb, "secpoldata_dump() has yet to be written.");
} // }}}
/*
int sam_getmachinesid( vbuf *out ) // {{{ in: user name (case sensitive?), out: valid RID or -1 on error
{
	int index;
	ulong type;
	char name[SZ_MAX];
	ulong len = SZ_MAX - 1;
	VBUF_STATIC( path, 10 );

	rl_hkey rl;
	ulong t;

	if( 0 != ns_read_valueT( NS_PATH_MACHINE_SID, &t, out ) ) return -1;
	return r;
} // }}}
*/
int sam_uname2rid( char *uname ) // {{{ in: user name (case sensitive?), out: valid RID or -1 on error
{
	int index;
	ulong type;
	char name[SZ_MAX];
	ulong len = SZ_MAX - 1;
	VBUF_STATIC( path, 10 );

	rl_hkey rl;
	ulong r;

	vbuf_printf( path, "%s/%s/\\0", NS_USER_NAME_PATH, uname );

	if( 0 != ns_read_valueT( path->b, &r, NULL ) ) return -1;
	return r;
} // }}}

///* 
//#define USER_NAME_PATH "\\SAM\\SAM\\Domains\\Account\\Users\\Names\\"
char *sam_rid2uname( int rid ) // {{{ return user name in static buffer
{
	static struct varbuf *vb = NULL;
	if( NULL == vb ) vb = vbuf_alloc( 10 );

	if( 0 != sam_get_user_name( vb, rid ) ) return NULL;

	return vb->b;
} // }}}
int sam_print_user( vbuf *out, int rid ) // {{{ Nicely print the information for a user
{
	//TODO: ugly and no ACB information

	struct uv *userv;
	struct uf *userf;

	VBUF_STATIC( unbuf, 10 );
	VBUF_STATIC( vb, 10 );

	if( 0 != sam_get_user_name( unbuf, rid ) ) {
		return -1;
	}
	//vbuf_catprintf(out, "------------------------------------------------------------------------------------------\n");
	vbuf_catprintf(out, "RID: %x: Username: %s\n", rid, unbuf->b);

	vbuf_catprintf(out, "  USER V STRUCT:\n");
	if( NULL == ( userv = sam_get_user_V( rid, vb ) ) ) {
		return -1;
	}
	uv_dump( out, userv );

	vbuf_catprintf(out, "  USER F STRUCT:\n");
	if( NULL == ( userf = sam_get_user_F( rid, vb ) ) ) {
		return -1;
	}
	uf_dump( out, userf );
	acb_print( out, userf->acb );

	return 0;
} // }}}
void acb_print( vbuf *out, short acb ) // {{{
{
	int b;
	vbuf_catprintf(out, "  USER F STRUCT:\n");
	vbuf_catprintf(out, "Account bits: 0x%04x =\n", acb);

	for (b=0; b < 15; b++) 
		vbuf_catprintf( out, "    %-20s: %s\n", acb_fields[b], (acb & (1<<b)) ? "Yes" : "No" );

} // }}}
ulong sam_get_max_lock() // {{{ No. of failed login attempts before account is locked out
{
	static vbuf *vb = NULL;
	struct samkeyf *f;
	if( !vb ) vb = vbuf_alloc(sizeof( struct samkeyf ) );
	f = sam_get_samkeyf( vb );

	return f->lockout_threshold;
} // }}}
ulong sam_get_pass_min() // {{{ Min pass length
{
	static vbuf *vb = NULL;
	struct samkeyf *f;
	if( !vb ) vb = vbuf_alloc(sizeof( struct samkeyf ) );
	f = sam_get_samkeyf( vb );

	return f->min_pass;
} // }}}
ulong sam_get_pass_history() // {{{ How many passwords are kept in history (???)
{
	static vbuf *vb = NULL;
	struct samkeyf *f;
	if( !vb ) vb = vbuf_alloc(sizeof( struct samkeyf ) );
	f = sam_get_samkeyf( vb );

	return f->enforce_password_history;
} // }}}

void acb_edit( struct uf *userf ) // {{{
{
	char yn[10];
	if ( userf->acb & (ACB_DISABLED|ACB_AUTOLOCK) || (userf->bad_login_count > 0 && userf->bad_login_count >= sam_get_max_lock() )  ) {
		printf("Account is %s\n",(userf->acb & ACB_DISABLED) ? "disabled" : "probably locked out!");
		printf("Do you wish me to reset the failed count, unset disabled and lockout,\n");
		fmyinput("and set the \"password never expires\" option? (y/n) [n]",yn,2);
		if (*yn == 'y') {
			userf->acb |= ACB_PWNOEXP;
			userf->acb &= ~ACB_DISABLED;
			userf->acb &= ~ACB_AUTOLOCK;
			//userf->acb = acb;
			userf->bad_login_count = 0;
			WARN("TODO: WRITE CHANGES TO REGISTRY");
			//put_buf2val(hive[H_SAM], v, 0, s, REG_BINARY);
			printf("Unlocked!\n");
		}
	}
} // }}}
//TODO: the following is commented out because it won't compile.  Fix it!
//void promote_user(int rid) // {{{
//{
//	/* Promote user into administrators group (group ID 0x220)
//	* And remove from all others...
//	* hdesc - hive
//	* rid   - users rid
//	* no returns yet
//	* THIS IS VERY HACKISH YET
//	*/
//	char s[200];
//	char g[200];
//	//int nk = 0;
//	rl_hkey hk;
//	long mblen = 0, cblen = 0;
//	//struct keyval *m = NULL, *c = NULL;
//	char *m, *c;
//	//struct keyval admember = { 4, 0x220 };
//	unsigned int *grps, *gcnts;
//	int count = 0;
//	int i, grp;
//	ulong size;
//
//	if (!rid || (H_SAM < 0)) return;
//
//	/* Get member list for user. Go for the first full SID, it's usually local computer I hope */
//	snprintf(s,180,"\\SAM\\SAM\\Domains\\Builtin\\Aliases\\Members\\S-1-5-21-\\%08X",rid);
//	/* Now, the TYPE field is the number of groups the user is member of */
//	/* Don't we just love the inconsistent use of fields!! */
//	ASSERT( rlERROR_SUCCESS == rlRegOpenKeyEx( HKEY_LOCAL_MACHINE, s, 0, 0, &hk ), "Couldn't open key." );
////	nk = get_file_offset( hive[H_SAM], trav_path(hive[H_SAM], 0, s, 0));
////	if (!nk) {
////		printf("Cannot find path <%s>\n",s);
////		return;
////	}
////	nk += 4;
////	count = get_val_type(hive[H_SAM],nk,"@");
////	if (count == -1) {
////		printf("Cannot find value <%s\\@>\n",s);
////		return;
////	}
////	printf("User is member of %d groups.\n",count);
//
//	/* This is the data size */
//	ASSERT( rlERROR_MORE_DATA == rlRegQueryValueEx( hk, "", NULL, NULL, &size ), "Another broken error message" );
//	//size = get_val_len(hive[H_SAM],nk,"@");
//
//	/* It should be 4 bytes for each group */
//	printf("Data size %ld bytes.\n",size);
//	if (size != count * 4) {
//		printf("DEBUG: Size is not 4 * count! May not matter anyway. Continuing..\n");
//	}
//
//	ASSERT( 0 < get_realloc( hk, "", NULL, &m, &mblen ), "Huh?");
//	//m = get_val2buf(hive[H_SAM], NULL, nk, "@", 0);
//	/*
//	if (!m) {
//		printf("Could not get value data! Giving up.\n");
//		return;
//	}
//	*/
//
//	printf("User was member of groups: ");
//	grps = (unsigned int *)m;
//	for (i = 0; i < count; i++) {
//		grp = grps[i];
//		printf("%08x ",grp);
//		switch (grp) {
//			case 0x220: printf("=Administrators, "); break;
//			case 0x221: printf("=Users, "); break;
//			case 0x222: printf("=Guests, "); break;
//			default: printf(", "); break;
//		}
//		snprintf(g, 180, "\\SAM\\SAM\\Domains\\Builtin\\Aliases\\%08X\\C",grp);
//		ASSERT( 0 < get_realloc( HKEY_LOCAL_MACHINE, g, NULL, &c, &cblen ), "Huh?");
////		ASSERT( rlERROR_MORE_DATA == rlRegQueryValueEx( HKEY_LOCAL_MACHINE, g, NULL, &blen ), "This is a bad error string" );
////		c = F_MALLOC( blen );
////		ASSERT( rlERROR_SUCCESS == rlRegQueryValueEx( HKEY_LOCAL_MACHINE, g, c, &blen ), "This is another bad error string" );
////		c = get_val2buf(hive[H_SAM], NULL, 0, g, 0);
//		if (c) {
//			gcnts = (unsigned int *)c;
//			gcnts[0xc]--;
//			/* Decrease members counter */
//			WARN("WRITE TO HIVE");
//			//put_buf2val(hive[H_SAM], c, 0, g, 0);
//		} else {
//			printf("Group info for %x not found!\n",grp);
//		}
//
//	}
//
//	cblen = 0;
//	FREE( c );
//#if 1
//	DIE("You need to replace this part.");
//	printf("\nDeleting user memberships\n");
//
//	//del_value(hive[H_SAM], hk, "@");
//
//	printf("Adding into only administrators:\n");
//
////	if (!add_value(hive[H_SAM], nk, "@", 1)) { /* Type is # of groups, here 1 */
////		printf("Failed to add @ value to key\n");
////	}
//#endif
//	//put_buf2val(hive[H_SAM], &admember, nk, "@", 0);
//
//	/* Now bumb up administrator groups count */
//	struct valh *vh;
//	if( NULL == (vh = vh_open( HKEY_LOCAL_MACHINE, "\\SAM\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C"))) {
//		printf("Group info for 220 (adm) not found!");
//		return;
//	}
//	//c = get_val2buf(hive[H_SAM], NULL, 0, "\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C", 0);
//	//if (!c) printf("Group info for 220 (adm) not found!\n");
//	gcnts = (unsigned int *)vh->buf;
//	gcnts[0xc]++;
//	//put_buf2val(hive[H_SAM], c, 0, "\\SAM\\Domains\\Builtin\\Aliases\\00000220\\C", 0);
//	
//	vh_close( vh );
//
//	printf("Promotion DONE!\n");
//
//} // }}}
void pass_print( vbuf *vb, int rid, struct uv *userv ) // {{{ Nicely print the password situation in uv
{
	int ntolen = uv_get_entry( userv, UV_NTPW )->len;
	char omd4[32], olanman[32];
	if ( ntolen < 16 )
		vbuf_catprintf(vb, "Note: NT password does not seem to be set or is invalid (expected 16 bytes, got %d).\n", ntolen);
	if( ntolen >= 0 ) {
		char *p = uv_get_field_ptr( userv, UV_NTPW, NULL );
		vbuf_catprintf(vb, "Crypted NT pw: ");
		vbuf_cathex( vb, p, ntolen );
		vbuf_catprintf(vb, "\n");

	}
	if( ntolen >= 16 ) {
		rid_pw_decrypt( rid, omd4, uv_get_field_ptr( userv, UV_NTPW, NULL ) );
		vbuf_catprintf(vb, "MD4 hash     : ");
		vbuf_cathex( vb, omd4, ntolen );
		vbuf_catprintf(vb, "\n");
	}

	int lmolen = uv_get_entry( userv, UV_LMPW )->len;
	if ( lmolen < 16 )
		vbuf_catprintf(vb, "Note: LM password does not seem to be set or is invalid (expected 16 bytes, got %d).\n", lmolen );
	if( lmolen >= 0 ) {
		char *p = uv_get_field_ptr( userv, UV_LMPW, NULL );
		vbuf_catprintf(vb, "Crypted LM pw: ");
		vbuf_cathex( vb, p, lmolen );
		vbuf_catprintf(vb, "\n");
	}
	if( lmolen >= 16 ) {
		rid_pw_decrypt( rid, olanman, uv_get_field_ptr( userv, UV_LMPW, NULL ) );
		vbuf_catprintf(vb, "MD4 hash     : ");
		vbuf_cathex( vb, olanman, lmolen );
		vbuf_catprintf(vb, "\n");
	}
} // }}}
//TODO: fix the following...I thought I had this done already...
//char *change_pw(int rid, int stat) // {{{
//{
//	/*
//	 * rid - the users RID, required for the DES decrypt stage
//	 * Some of this is ripped & modified from pwdump by Jeremy Allison
//	 */
//
//	int pl;
//	char newp[20];
//
//	char *username;
//	static struct varbuf *unbuf = NULL;
//	if( NULL == unbuf ) unbuf = vbuf_alloc(10);
//	sam_get_user_name( rid, unbuf );
//	username = unbuf->b;
//
//	static struct varbuf *uvbuf;
//	if( NULL == uvbuf) uvbuf = vbuf_alloc(10);
//	if( NULL == ( sam_get_user_V( rid, uvbuf ) ) ) return NULL;
//
//	struct uf *userf;
//	static struct varbuf *ufbuf;
//	if( NULL == ufbuf) ufbuf = vbuf_alloc(10);
//	if( NULL == ( userf = sam_get_user_F( rid, ufbuf ) ) ) return NULL;
//
//	if (stat) {
//		//acb = handle_F(rid,0);
//		printf("RID: %04x, Username: <%s>%s\n",
//		rid, username,  (  userf->acb & 0x8000 ? ", *disabled or locked*" : (((UV)uvbuf->b)->head.e[UV_NTPW].len  < 16) ? ", *BLANK password*" : "")  );
//		return(username);
//	}
//
//	/* TODO: this used to print, now doesn't...
//	printf("RID     : %04d [%04x]\n",rid,rid);
//	printf("Username: %s\n",username);
//	printf("fullname: %s\n",fullname);
//	printf("comment : %s\n",comment);
//	printf("homedir : %s\n\n",homedir);
//	*/
//
//	if( gverbose ) printf("Current passwords:\n");
//	pass_print( rid, (UV)uvbuf->b );
//		
//	if (!rid) {
//		printf("No RID given. Unable to change passwords..\n");
//		return(0);
//	}
//
//	if( 0 >= ( pl = user_get_password( newp, 16 ) ) ) return NULL;
//	if (!pl) { printf("Nothing changed.\n"); return(0); }
//
//	if (pl == 1 && *newp == '@') { //TODO: should this be here?
//		promote_user(rid);
//		printf("Changes saved!\n");
//		return(username);
//	}
//
//	if( pl > 0 ) {
//		if( 0 == strcmp( newp, "*" ) ) {
//			printf("Blanking password!\n");
//			sam_set_nt_password( rid, "" );
//			sam_set_lm_password( rid, "" );
//		}
//		else {
//			sam_set_nt_password( rid, newp );
//			sam_set_lm_password( rid, newp );
//		}
//
//
//		if( gverbose ) printf("New passwords:\n");
//		pass_print( rid, (struct uv*)uvbuf->b);
//	} else {
//		printf("Nothing entered, not setting passwords.\n");
//	}
//
//	printf("\n");
//	return(username);
//} // }}}

char *nt2s( struct nt_time *nt );
// This should be in libhive

int get_reg_value( rl_hkey k, char *path, char *vname, struct varbuf *vb ) // {{{ TODO: destroy this
{
	ulong needed = 0;
	rl_hkey hkv;
	ulong type;

	if( rlERROR_SUCCESS != rlRegOpenKeyEx( k, path, KEY_ALL_ACCESS, 0, &hkv) ) {
		//TODO: error
		printf("Couldn't open key, does it exist?\n");
		return -1;
	}
	ASSERT( rlERROR_MORE_DATA == rlRegQueryValueEx( hkv, vname, &type, NULL, &needed ), "get_reg_value(): A zero buffer is big enough?");
	vbuf_resize( vb, needed );
	ASSERT( rlERROR_SUCCESS == rlRegQueryValueEx( hkv, vname, &type, vb->b, &needed ) , "get_reg_value(): rlRegQueryValueEx error.");
	vb->dlen = needed;

	return 0;
} // }}}
// Generic V Structure stuff {{{
// user V fields

const char * imachinevfields[MV_FIELDS] = { // {{{ 
	"MACHINE_UNKNOWN",
	"MACHINE_UNKNOWN",
	"MACHINE_UNKNOWN",
	"MACHINE_UNKNOWN"
}; // }}}

const int imachinevtypes[MV_FIELDS] = { // {{{
	REG_NONE,
	REG_NONE,
	REG_NONE,
	REG_NONE
}; // }}}


const char * iuvfields[UV_FIELDS] = { // {{{ 
	//"UNKNOWN1",

	"USERNAME",
	"FULLNAME",
	"COMMENT",
	"USER_COMMENT",

	"UNKNOWN2",

	"HOMEDIR",
	"HOMEDIRCONNECT",
	"SCRIPTPATH",
	"PROFILEPATH",

	"WORKSTATIONS",
	"HOURS_ALLOWED",
	"UNKNOWN3",

	"LMPW",
	"NTPW",

	"UNKNOWN4",
	"UNKNOWN5"
}; // }}}
const int iuvtypes[UV_FIELDS] = { // {{{
	//REG_NONE,

	REG_SZ,
	REG_SZ,
	REG_SZ,
	REG_SZ,

	REG_NONE,

	REG_SZ,
	REG_SZ,
	REG_SZ,
	REG_SZ,

	REG_NONE,
	REG_NONE,
	REG_NONE,

	REG_NONE,
	REG_NONE,

	REG_NONE,
	REG_NONE
}; // }}}

/*
struct vdef {
	const int count;
	const int *types;
	const char **fields;
};
*/

const static struct vdef vdef_uv = { UV_FIELDS, iuvtypes, iuvfields };
const static struct vdef vdef_machinev = { MV_FIELDS, imachinevtypes, imachinevfields };

struct sam_v *sam_get_V( char *path, struct varbuf *vb ) { // {{{
	unsigned long type;

	ASSERT(vb, "vb must not be NULL" );

	if( 0 != ns_read_valueT( path, &type, vb ) ) return NULL;

	if( type != REG_BINARY ) {
		ERR_UNIX( EINVAL, "Expected binary type, got %d", type );
		return NULL;
	}

	//TODO: sanity check here

	/*
	if( vb->dlen < 400 ) { // TODO: I don't know the exact length
		ERR_UNIX( EINVAL, "Data corruption?  Length is too small." );
		return NULL;
	}
	*/

	return (struct sam_v*)vb->b;
} // }}}

struct v_lookup *sam_v_get_lookup( struct sam_v *v ) // {{{ Get attribute lookup table
{
	return (struct v_lookup*)( (char*)v + sizeof( struct v_header ) );
} // }}}
/*
int sam_v_get_lookup_count( struct sam_v *v ) // {{{ Get count of attribute lookup table
{
	return (v->head.off2 - v->head.off1)/sizeof( struct v_entry );
} // }}}
*/
struct v_entry *sam_v_get_table_entry( struct sam_v *v, int i ) // {{{ Get attribute table lookup entry
{
	return &sam_v_get_lookup( v )->e[i];
} // }}}
int sam_v_attrib_len( struct sam_v *v, int i ) // {{{
{
	sam_v_get_table_entry( v, i )->len;
} // }}}
char *sam_v_attrib_get( struct sam_v *v, char *s2, int *len, int i ) // {{{
{
	struct v_entry *e = sam_v_get_table_entry( v, i );
	*len = e->len;
	return e->off + s2;
} // }}}
int sam_v_info( vbuf *samvbuf, int *ecount, char **s2start ) // {{{ go through section 1, counting entries until we hit section 2
{
	struct sam_v *v = (struct sam_v*)samvbuf->b;
	struct v_entry *e0 = &sam_v_get_lookup( v )->e[0];
	struct v_entry *e = e0;
	if( ecount ) (*ecount) = 0;

	do {

		if( ecount )(*ecount)++;
		//TODO: Check element sanity
		if( (char*)e - (char*)v + sizeof( struct v_lookup ) > samvbuf->dlen - sizeof( struct v_lookup ) ) {

			ERR_UNIX( EINVAL, "Hit end of V structure looking for section 2" );
			return -1;
		}
	}
	while( e++ && e->off != 0x80140001 ); //TODO: find somewhere horrible for this horrible magic number

	(*s2start) = (char*)e;

	return 0;
} // }}}
/*
char *sam_v_get_field_ptr( struct sam_v *v, int eno, int *blen ) { // {{{
	struct sam_v_entry *e = sam_v_get_entry( v, eno );
	char *buf = (char*)((char*)userv + 0xCC + e->off);
	//fprintf(stderr, "pointers: uv=%p, buf=%p\n", userv, buf );
	if( NULL != blen ) *blen = e->len;
	return buf;
} // }}}
*/
int sam_v_getfield( int i, char **str, int *type, const struct vdef *vd ) // {{{
{
	if( i > vd->count ) {
		*str = "UNKNOWN";
		*type = REG_NONE;
		return 0;
	}

	*type = vd->types[i];
	*str = (char*)vd->fields[i];

	return 0;
} // }}}
int sam_v_perm_sec_count( vbuf *vb_v, int sec ) // {{{
{
	struct v_perms_ent *vperms_ent;
	struct v_perms_3 *vperms_s3;
	struct v_perms *vperms; 
	char *perms;
	if( 0 != sam_v_info( vb_v, NULL, &perms ) ) return -1;

	vperms = (struct v_perms*)perms;
	vperms_ent = (struct v_perms_ent*)((char*)perms + vperms->offs2[sec] );
	vperms_s3 = (struct v_perms_3*)vperms_ent->data;
	return vperms_s3->count;

} // }}}
struct v_perms_pent *sam_v_get_perm( vbuf *vb_v, int sec, int off ) // {{{
{
	struct v_perms_ent *vperms_ent;
	struct v_perms_3 *vperms_s3;
	struct v_perms *vperms; 
	char *perms;
	int j;
	if( 0 != sam_v_info( vb_v, NULL, &perms ) ) return NULL;

	vperms = (struct v_perms*)perms;
	vperms_ent = (struct v_perms_ent*)((char*)perms + vperms->offs2[sec] );
	vperms_s3 = (struct v_perms_3*)vperms_ent->data;

	struct v_perms_pent *ne = &vperms_s3->ents[0];
	for( j=0; j<vperms_s3->count;j++) {
		if( off == j ) return ne;
		ne = (struct v_perms_pent*)((char*)ne + ne->len);
	}

	return NULL;
} // }}}
int sam_v_perms_dump( vbuf *vbuf_v, vbuf *vbout,  char *perms ) // {{{ Dump V, section 2
{
	struct v_perms_pent *ne;
	int i;
	int j;
	/*
	struct uv_entry *e;
	int len;
	struct v_perms *vperms; 
	struct v_perms_ent *vperms_ent;
	struct v_perms_3 *vperms_s3;
	struct sam_v *v = (struct sam_v*)vbuf_v->b;
	int i,j;

	vperms = (struct v_perms*)perms;

	//e  = &vs2->head.e[0];
	//uv2 = (struct userv2*)uv_get_field_ptr( userv, 0, &len );
	hexdump( perms, 0, vbuf_v->dlen - ( perms - vbuf_v->b), 1 );
	*/
	vbuf_catprintf(vbout, "Permissions:\n");

	for( i=0; i<2; i++) {
		/*
		vperms_ent = (struct v_perms_ent*)((char*)perms + vperms->offs2[i] );
		//WARN("uv2e: type: %x, len: 0x%x", uv2e->type, uv2e->len);
		//hexdump( uv2e->data, 0, uv2e->len, 1 );
		vperms_s3 = (struct v_perms_3*)vperms_ent->data;

		vbuf_catprintf(vbout, "Sec %d [0x%x]\n", i, vperms->offs2[i] );

		struct v_perms_pent *ne = &vperms_s3->ents[0];
		*/

		vbuf_catprintf( vbout, "  sec %d\n", i);
		for( j=0; j<sam_v_perm_sec_count( vbuf_v, i );j++) {
			ne = sam_v_get_perm( vbuf_v, i, j );
			//vbuf_catprintf( vbout, "  perms item:  Type: %x, len: %x, off: %x", ne->type, ne->len, (char*)ne - (char*)vperms_ent );
			//vbuf_catprintf( vbout, "  \n    perms: " );
			vbuf_catprintf( vbout, "  ");
			vbuf_cathex( vbout, ne->perms, 4 );
			vbuf_catprintf( vbout, "  ");
			//vbuf_catprintf( vbout, "    sid: " );
			vbuf_cathex( vbout, ne->data, ne->len - 4 );
			vbuf_catprintf( vbout, "\n" );
			ne = (struct v_perms_pent*)((char*)ne + ne->len);
		}
	}

	return 0;
	
} // }}}
/*
int sam_v_perms_dump( vbuf *vbuf_v, vbuf *vbout,  char *perms ) // {{{ Dump V, section 2
{
	struct uv_entry *e;
	int len;
	struct v_perms *vperms; 
	struct v_perms_ent *vperms_ent;
	struct v_perms_3 *vperms_s3;
	struct sam_v *v = (struct sam_v*)vbuf_v->b;
	int i,j;

	vperms = (struct v_perms*)perms;

	//e  = &vs2->head.e[0];
	//uv2 = (struct userv2*)uv_get_field_ptr( userv, 0, &len );
	hexdump( perms, 0, vbuf_v->dlen - ( perms - vbuf_v->b), 1 );
	vbuf_catprintf(vbout, "Permissions:\n");

	for( i=0; i<2; i++) {
		vperms_ent = (struct v_perms_ent*)((char*)perms + vperms->offs2[i] );
		//WARN("uv2e: type: %x, len: 0x%x", uv2e->type, uv2e->len);
		//hexdump( uv2e->data, 0, uv2e->len, 1 );
		vperms_s3 = (struct v_perms_3*)vperms_ent->data;

		vbuf_catprintf(vbout, "Sec %d [0x%x]\n", i, vperms->offs2[i] );

		struct v_perms_pent *ne = &vperms_s3->ents[0];
		for( j=0; j<vperms_s3->count;j++) {
			//vbuf_catprintf( vbout, "  perms item:  Type: %x, len: %x, off: %x", ne->type, ne->len, (char*)ne - (char*)vperms_ent );
			//vbuf_catprintf( vbout, "  \n    perms: " );
			vbuf_catprintf( vbout, "  ");
			vbuf_cathex( vbout, ne->perms, 4 );
			vbuf_catprintf( vbout, "  ");
			//vbuf_catprintf( vbout, "    sid: " );
			vbuf_cathex( vbout, ne->data, ne->len - 4 );
			vbuf_catprintf( vbout, "\n" );
			ne = (struct v_perms_pent*)((char*)ne + ne->len);
		}
	}

	return 0;
	
} // }}}
*/
int sam_vmachine_dump( vbuf *v, vbuf *vb ) // {{{
{
	return sam_v_dump( v, vb, &vdef_machinev );
} // }}}
int sam_vuser_dump( vbuf *v, vbuf *vb ) // {{{
{
	return sam_v_dump( v, vb, &vdef_uv );
} // }}}
int sam_v_dump( vbuf *vb, vbuf *vbout, const struct vdef *vd ) // {{{
{
	int i;
	struct sam_v *v = (struct sam_v*)vb->b;
	char *buf;
	int len;
	char *s2;
	int count;
	VBUF_STATIC( val8buf, 10 );
	VBUF_STATIC( zbuf, 10 );

	if( 0 != sam_v_info( vb, &count, &s2 ) ) 
		return -1;

	if( 0 != sam_v_perms_dump( vbout, vbout, s2 ) )
		return -1;

	warn("Out: %s", vbout->b );

	vbuf_catprintf( vbout, "Got count %d and s2 offset %x\n", count, (char*)(s2 - vb->b) );

	for( i=0; i<count; i++ ) {
		char *fname;
		int ftype;

		sam_v_getfield( i, &fname, &ftype, vd );

		buf = sam_v_attrib_get( v, s2, &len, i );
		/*
		vbuf_catprintf( vbout, "  Element %03d [%03x]: ", i, len );
		vbuf_cathex( vbout, buf, len );
		vbuf_catprintf( vbout, "\n" );
		*/

		vbuf_catprintf(vbout, "    0x%04x 0x%03x %-20s: ", (char*)buf - (char*)v, len, fname );
		if( 0 != ftype ) {
			vbuf_set( zbuf, buf, len );
			vb_utf16_ensure_zero_termination( zbuf );
			vbprint_val8( val8buf, zbuf->b, ftype, len+2, 0 );
			vbuf_strcat( vbout, val8buf->b );
		}
		else {
			vbuf_cathex( vbout, buf, len );
			vbuf_strcat( vbout, "\n");
		}




	}

	return 0;

	/*
	WARN("off1: 0x%x", v->head.off1 );

	WARN("looklen: 0x%x", v->head.looklen );
	hexdump( vb->b, 0, vb->dlen, 1 );
	for( i=0; i<sam_v_get_lookup_count( v ); i++ ) {
		buf = sam_v_get_field_ptr( v, i, &len );
		WARN("Field pointer %d:", i );
		hexdump(buf, 0, len, 1 );
	}
	*/
	/*
	int f;
	struct v_entry *e;
	char *buf;
	int len;
	char zbuf[SZ_MAX];
	VBUF_STATIC( val8buf, 10 );

	dump_ms2( vb, userv );

	for( f=1; f<UV_FIELDS; f++ ) {
		e  = &userv->head.e[f];
		buf = uv_get_field_ptr( userv, f, &len );

		ASSERT( len - 2 < SZ_MAX, "String too big.");
		memcpy( zbuf, buf, len );
		zbuf[len] = '\0';
		zbuf[len+1] = '\0';
		//fprintf(stderr, "  %-20s %08x:%08x:%08x\n", get_uv_name( f ), e->off, e->len, e->unknown );
		vbuf_catprintf(vb, "    0x%04x 0x%03x %-20s: ", (char*)buf - (char*)userv, len, uv_get_field_name( f ) );
		if( 0 != uv_get_field_type( f ) ) {
			vbprint_val8( val8buf, zbuf, uv_get_field_type( f ), len+2, 0 );
			vbuf_strcat( vb, val8buf->b );
		}
		else {
			vbuf_cathex( vb, buf, len );
			vbuf_strcat( vb, "\n");
		}
		//hexdump( buf, 0, len, 1 );
	}
	return 0;
	*/
} // }}}
// }}}
//
// Machine V
struct machinev *sam_get_machine_V( struct varbuf *vb ) // {{{
{
	return (struct machinev*)sam_get_V( NS_PATH_MACHINE_V, vb );
} // }}}
// USER_V stuff
const char * uvfields[UV_FIELDS] = { // {{{ 
	"UNKNOWN1",

	"USERNAME",
	"FULLNAME",
	"COMMENT",
	"USER_COMMENT",

	"UNKNOWN2",

	"HOMEDIR",
	"HOMEDIRCONNECT",
	"SCRIPTPATH",
	"PROFILEPATH",

	"WORKSTATIONS",
	"HOURS_ALLOWED",
	"UNKNOWN3",

	"LMPW",
	"NTPW",

	"UNKNOWN4",
	"UNKNOWN5"
}; // }}}
const int types[UV_FIELDS] = { // {{{
	REG_NONE,

	REG_SZ,
	REG_SZ,
	REG_SZ,
	REG_SZ,

	REG_NONE,

	REG_SZ,
	REG_SZ,
	REG_SZ,
	REG_SZ,

	REG_NONE,
	REG_NONE,
	REG_NONE,

	REG_NONE,
	REG_NONE,

	REG_NONE,
	REG_NONE
}; // }}}
struct uv *sam_get_user_V( ulong rid, struct varbuf *vb ) { // {{{
	unsigned long type;

	ASSERT(vb, "vb must not be NULL" );

	VBUF_STATIC( path, 100 );

	vbuf_printf( path,"%s/%08lX/V", NS_USER_PATH, rid);

	//WARN("sam_get_user_V() attempting path %s", path->b );
	if( 0 != ns_read_valueT( path->b, &type, vb ) ) return NULL;

	if( type != REG_BINARY ) {
		ERR_UNIX( EINVAL, "Expected binary type, got %d", type );
		return NULL;
	}

	if( vb->dlen < 400 ) { // TODO: I don't know the exact length
		ERR_UNIX( EINVAL, "Data corruption?  Length is too small." );
		return NULL;
	}

	return (struct uv*)vb->b;
} // }}}
int sam_set_user_V( ulong rid, struct varbuf *vb ) { // {{{
	VBUF_STATIC( pathbuf, 10 );
	struct uv *userv = (struct uv*)vb->b;

	if( 0 != uv_check( userv, vb->dlen ) ) {
		//TODO: error
		printf("sam_set_user_V(): the user v struct seems to be corrupt.\n");
		return -1;
	}

	//TODO: verify type
	vbuf_printf( pathbuf, "/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/Account/Users/%08lX/V",rid);
	if( 0 != ns_set_valueT( pathbuf->b, REG_BINARY, vb->b, vb->dlen ) ) {
		//TODO: error
		printf("sam_set_user_V(): couldn't set value\n");
		return -1;
	}

	return 0;
	
} // }}}
int uv_set_field( struct varbuf *vb, int fno, char *buf, int blen ) // {{{ in: vb contains struct uv; out: buf data stored in field fno
{
	//struct uv *userv = (struct uv*)vb->b;
	struct uv_entry *ent = uv_get_entry( (UV)vb->b, fno );
	int olen = ent->len;
	int ldiff = blen - olen;
	int osize = vb->dlen;
	int i;

	//printf("uv_set_field() in: \n");
	//uv_dump( (UV)vb->b );

	if( ldiff > 0 ) vbuf_grow( vb, ldiff );

	vb->dlen += ldiff;

	ent = uv_get_entry( (UV)vb->b, fno );

	void *fp = uv_get_field_ptr( (UV)vb->b, fno, NULL );

	memmove( fp + blen, fp + olen, osize - ( (char*)fp - (char*)(UV)vb->b ) - olen );

	for( i=0; i<UV_FIELDS; i++ ) {
		if( i != fno ) {
			struct uv_entry *e = uv_get_entry( (UV)vb->b, i );
			if( e->off > ent->off ) e->off += ldiff;
		}
	}

	ent->len = blen;
	memcpy( fp, buf, blen );

	ASSERT( 0 == uv_check( (UV)vb->b, vb->dlen ), "uv_set_field(): I think I broke it.");

	//printf("uv_set_field() out: \n");
	//uv_d( (UV)vb->b );

	return 0;

} // }}}
int uv_check( struct uv *userv, int len ) // {{{ returns 0 if userv seems sane, -1 otherwise
{
	struct uv_entry *uve;
	int fno;
	for( fno = 0; fno < UV_FIELDS; fno ++ ) {
		uve = uv_get_entry( userv, fno );
		//fprintf(stderr, "uv_check(): checking field %s (%d, %d)\n", uvfields[fno], uve->len, uve->off );
		if( uve->len < 0 ) {
			//TODO: error
			fprintf(stderr, "User V field %s has negative length.\n", uvfields[fno] );
			return -1;
		}

		if( uve->off < 0 ) {
			//TODO: error
			fprintf(stderr, "User V field %s has negative offset.\n", uvfields[fno]);
			return -1;
		}

		if( uve->off + uve->len + 0xCC > len ) {
			//TODO: error
			fprintf(stderr, "User V field %s offset + length > total userv size.\n", uvfields[fno]);
			return -1;
		}
	}

	return 0;
} // }}}
struct uv_entry* uv_get_entry( struct uv *userv, int off ) // {{{
{
	return &userv->head.e[off];
} // }}}
char *uv_get_field_ptr( struct uv *userv, int eno, int *blen ) { // {{{
	struct uv_entry *e = uv_get_entry( userv, eno );
	char *buf = (char*)((char*)userv + 0xCC + e->off);
	//fprintf(stderr, "pointers: uv=%p, buf=%p\n", userv, buf );
	if( NULL != blen ) *blen = e->len;
	return buf;
} // }}}
struct userv2ent {
	short type;
	short len;
	char data[];
};

struct userv2pent {
	short type;
	short len;
	char perms[4];
	char data[];
};
struct userv2 {
	char u1[4];
	int offs1[2];
	int offs2[2];
};
struct userv3 {
	int count;
	struct userv2pent ents[];
};
int dump_ms2( vbuf *vb, struct uv *userv ) // {{{ Dump User V, struct 2
{
	struct uv_entry *e;
	int len;
	struct userv2 *uv2; 
	struct userv2ent *uv2e;
	struct userv3 *uv3;
	int i,j;

	e  = &userv->head.e[0];
	uv2 = (struct userv2*)uv_get_field_ptr( userv, 0, &len );
	//hexdump((void*)uv2 , 0, len, 1 );

	for( i=0; i<2; i++) {
		WARN("Off: 0x%x", uv2->offs2[i] );
		uv2e = (struct userv2ent*)((char*)uv2 + uv2->offs2[i] );
		//WARN("uv2e: type: %x, len: 0x%x", uv2e->type, uv2e->len);
		//hexdump( uv2e->data, 0, uv2e->len, 1 );
		uv3 = (struct userv3*)uv2e->data;

		struct userv2pent *ne = &uv3->ents[0];
		WARN("subcount: %d", uv3->count );
		for( j=0; j<uv3->count;j++) {
			WARN("--------------------");
			WARN("FOO Type: %x, len: %x, off: %x", ne->type, ne->len, (char*)ne - (char*)uv2e );
			hexdump(ne->perms, 0, 4, 1);
			hexdump(ne->data, 0, ne->len - 4, 1);
			ne = (struct userv2pent*)((char*)ne + ne->len);
		}
	}
	
} // }}}
int uv_dump( vbuf *vb, struct uv *userv ) // {{{
{
	int f;
	struct uv_entry *e;
	char *buf;
	int len;
	char zbuf[SZ_MAX];
	VBUF_STATIC( val8buf, 10 );

	dump_ms2( vb, userv );

	for( f=1; f<UV_FIELDS; f++ ) {
		e  = &userv->head.e[f];
		buf = uv_get_field_ptr( userv, f, &len );

		ASSERT( len - 2 < SZ_MAX, "String too big.");
		memcpy( zbuf, buf, len );
		zbuf[len] = '\0';
		zbuf[len+1] = '\0';
		//fprintf(stderr, "  %-20s %08x:%08x:%08x\n", get_uv_name( f ), e->off, e->len, e->unknown );
		vbuf_catprintf(vb, "    0x%04x 0x%03x %-20s: ", (char*)buf - (char*)userv, len, uv_get_field_name( f ) );
		if( 0 != uv_get_field_type( f ) ) {
			vbprint_val8( val8buf, zbuf, uv_get_field_type( f ), len+2, 0 );
			vbuf_strcat( vb, val8buf->b );
		}
		else {
			vbuf_cathex( vb, buf, len );
			vbuf_strcat( vb, "\n");
		}
		//hexdump( buf, 0, len, 1 );
	}
	return 0;
} // }}}
int uv_get_field_type( int fno ) // {{{
{
	return types[fno];
} // }}}
const char *uv_get_field_name( int fno ) // {{{
{
	return uvfields[fno];
} // }}}

// USER_F stuff
struct uf *sam_get_user_F( ulong rid, struct varbuf *vb ) { // {{{
	unsigned long type;
	VBUF_STATIC( path, 100 );

	vbuf_printf( path,"%s/%08lX/F", NS_USER_PATH, rid);

	if( 0 != ns_read_valueT( path->b, &type, vb ) ) return NULL;

	if( type != REG_BINARY ) {
		ERR_UNIX( EINVAL, "Expected binary type, got %d", type );
		return NULL;
	}

	if( vb->dlen != USER_F_SIZE ) { 
		ERR_UNIX( EINVAL, "Data corruption?  Length is wrong (%d != %d).",
				vb->dlen, USER_F_SIZE );
		return NULL;
	}

	return (struct uf*)vb->b;
} // }}}
struct uf *sam_get_user_FA( ulong rid, struct varbuf *vb ) // {{{
{
	//TODO: this SZ_MAX stuff is for the birds
	char s[SZ_MAX];

	snprintf(s,180,"\\SAM\\SAM\\Domains\\Account\\Users\\%08lX",rid);
	if( 0 != get_reg_value( rlHKEY_LOCAL_MACHINE, s, "F", vb ) )
		return NULL;

	if( vb->dlen != USER_F_SIZE ) {
		//TODO: error
		fprintf(stderr, "sam_get_user_F(): value length is not 0x50, corrupt?\n");
		return NULL;
	}

	return (struct uf*)vb->b;
} // }}}
const char *get_uf_name( int fno ) // {{{
{
	return uf_fields[fno];
} // }}}
int uf_dump( vbuf *vb, struct uf *userf ) // {{{
{
	/*
	VBUF_STATIC( hexbuf, 10 );
	vbuf_hexdump( hexbuf, userf->unknown1, 0, sizeof( userf->unknown1 ), 1 );
	*/
	vbuf_catprintf( vb, "    %-20s ", get_uf_name(0) );
	vbuf_cathex( vb, userf->unknown1, sizeof( userf->unknown1 ) );
	vbuf_catprintf( vb, "\n" );
	vbuf_catprintf( vb, "    %-20s %s\n", get_uf_name(1), nt2s( &userf->last_login));
	vbuf_catprintf( vb, "    %-20s %s\n", get_uf_name(2), nt2s( &userf->unknown2));
	vbuf_catprintf( vb, "    %-20s %s\n", get_uf_name(3), nt2s( &userf->last_pw_change));
	vbuf_catprintf( vb, "    %-20s %s\n", get_uf_name(4), nt2s( &userf->account_expires));
	vbuf_catprintf( vb, "    %-20s %s\n", get_uf_name(5), nt2s( &userf->last_invalid_pass));
	vbuf_catprintf( vb, "    %-20s %04x\n", get_uf_name(6), userf->rid);
	vbuf_catprintf( vb, "    %-20s %04x\n", get_uf_name(7), userf->unknown3);
	vbuf_catprintf( vb, "    %-20s %x\n", get_uf_name(8), userf->acb);
	//vbuf_catprintf( vb, "    %-20s %x\n", get_uf_name(9), userf->password_flags);
	vbuf_catprintf( vb, "    %-20s %02x\n", get_uf_name(10), userf->unknown4);
	vbuf_catprintf( vb, "    %-20s %02x\n", get_uf_name(11), userf->country_code);
	vbuf_catprintf( vb, "    %-20s %02x\n", get_uf_name(12), userf->unknown5);
	vbuf_catprintf( vb, "    %-20s %02x\n", get_uf_name(13), userf->bad_login_count);
	vbuf_catprintf( vb, "    %-20s %02x\n", get_uf_name(14), userf->no_login);
	vbuf_catprintf( vb, "    %-20s %04x\n", get_uf_name(15), userf->unknown6);
	vbuf_catprintf( vb, "    %-20s %08lx\n", get_uf_name(16), userf->unknown7);

	return 0;
} // }}}
time_t nt2timet( struct nt_time *nt ) // {{{ warning: this is broken
{
	// the nt time format is stored in ten-millionths of a second since Jan 1, 1601, 00:00:00 
	//this is the nttime for Jan 1 of 2000 at 00:00:00
	//TODO: make this 1970 and skip a step later on
	unsigned char y2kb[] = { 0x00, 0x40, 0x6D, 0x25, 0xEB, 0x53, 0xBF, 0x01 };

	time_t y2kepoc = 946713600; //seconds between the epoc and 2000
	time_t result;

	//here we have the value of an nttime 
	//unsigned char r[8];

	//int64_t years;
	int64_t *y2k = (int64_t*)y2kb;
	int64_t *nti = (int64_t*)nt->t;

	int64_t val = *nti - *y2k;

	result = val/10000000 + y2kepoc;

	return result;

} // }}}
int ntiszero( struct nt_time *nt )  // {{{ nt is zero? return 1, else return 0
{
	int i;
	for( i=0; i<8; i++ ) {
		if( nt->t[i] != 0 ) return 0;
	}

	return 1;
} // }}}
char *nt2s( struct nt_time *nt ) // {{{
{
	static char nicetime[SZ_MAX];
	time_t t = nt2timet( nt );
	struct tm *ts= localtime( &t  );
	//int64_t *nti = (int64_t*)nt->t;
	if( ntiszero( nt ) ) { return "-"; }
	strftime( nicetime, SZ_MAX, "%Y%m%d-%H%M%S", ts );

	//fprintf(stderr, "%016llx -> \"%s\"\n", *nti, nicetime);

	return nicetime;
} // }}}

// Crypto stuff
void rid_make_sched( int rid, des_key_schedule *ks ) //{{{ ks MUST be des_key_schedule[2]
{ 
	des_cblock deskey1, deskey2;

	/* Get the two decrpt keys. */
        sid_to_key1(rid,(unsigned char *)deskey1);
        des_set_key((des_cblock *)deskey1,ks[0]);
        sid_to_key2(rid,(unsigned char *)deskey2);
        des_set_key((des_cblock *)deskey2,ks[1]);
} // }}}
void rid_pw_decrypt( int rid, char *dest, char *src ) // {{{
{
	des_key_schedule ks[2];
	rid_make_sched( rid, ks );

        /* Decrypt the NT md4 password hash as two 8 byte blocks. */
        des_ecb_encrypt((des_cblock *)src,       (des_cblock *)dest,       ks[0], DES_DECRYPT);
        des_ecb_encrypt((des_cblock *)(src + 8), (des_cblock *)(dest + 8), ks[1], DES_DECRYPT);
} // }}}
void rid_pw_encrypt( int rid, char *dest, char *src ) // {{{
{
	des_key_schedule ks[2];
	rid_make_sched( rid, ks );

	des_ecb_encrypt((des_cblock *)src, (des_cblock *)dest, ks[0], DES_ENCRYPT);
	des_ecb_encrypt((des_cblock *)(src+8), (des_cblock *)(dest + 8), ks[1], DES_ENCRYPT);
} // }}}
int lmhash_make( int rid, char *dest, char *src ) // {{{ in: src = null terminated UTF8, out: dest = encrypted lanman hash
{
	int len;

	static uchar x1[] = {0x4B,0x47,0x53,0x21,0x40,0x23,0x24,0x25};

	char buf[16];
	char buf2[16];

	len = strlen( src );

	if( len >= 15 ) {
		ERR_UNIX(EINVAL, "lanman hashing works only on strings of length less than 15.");
		return -1;
	}

        make_lanmpw(src, buf, len);

        E1(buf,   x1, buf2);
        E1(buf+7, x1, buf2+8);

	rid_pw_encrypt( rid, dest, buf2 );

	return 0;

} // }}}
int nthash_make( int rid, char *dest, char *src ) // {{{ in: src = null terminated UTF8, out: dest = crypted nt hash
{
	MD4_CTX context;
	unsigned char digest[16];
	//static struct varbuf *vb = NULL;
	//int len = strlen( src );
	//int len16 = (len + 1) * 2;
	VBUF_STATIC( utf16buf, 10 );

	//if( NULL == vb ) vb = vbuf_alloc(20);

	//vbuf_resize( vb, len16 );

	//if( 0 != utf8to16( src, len+1, vb->b, len16 ) ) return -1;

	if( 0 != vb_utf8to16( utf16buf, src, strlen( src ) + 1 ) ) return -1;

	MD4Init (&context);
	MD4Update (&context, utf16buf->b, utf16buf->dlen );
	MD4Final (digest, &context);

	rid_pw_encrypt( rid, dest, digest );

	return 0;
} // }}}
void make_lanmpw(char *p, char *lm, int len) // {{{ zero lanman passwd, replace with uppercase supplied password
{
	int i;

	for (i=0; i < 15; i++) lm[i] = 0;
	for (i=0; i < len; i++) lm[i] = toupper(p[i]);
} // }}}
void str_to_key(unsigned char *str,unsigned char *key) // {{{ Convert a 7 byte array into an 8 byte des key with odd parity.
{
	int i;

	key[0] = str[0]>>1;
	key[1] = ((str[0]&0x01)<<6) | (str[1]>>2);
	key[2] = ((str[1]&0x03)<<5) | (str[2]>>3);
	key[3] = ((str[2]&0x07)<<4) | (str[3]>>4);
	key[4] = ((str[3]&0x0F)<<3) | (str[4]>>5);
	key[5] = ((str[4]&0x1F)<<2) | (str[5]>>6);
	key[6] = ((str[5]&0x3F)<<1) | (str[6]>>7);
	key[7] = str[6]&0x7F;
	for (i=0;i<8;i++) {
		key[i] = (key[i]<<1);
	}
	DES_set_odd_parity((des_cblock *)key);
} // }}}
void sid_to_key1(unsigned long sid,unsigned char deskey[8]) // {{{ Function to convert the RID to the first decrypt key.
{
	unsigned char s[7];

	s[0] = (unsigned char)(sid & 0xFF);
	s[1] = (unsigned char)((sid>>8) & 0xFF);
	s[2] = (unsigned char)((sid>>16) & 0xFF);
	s[3] = (unsigned char)((sid>>24) & 0xFF);
	s[4] = s[0];
	s[5] = s[1];
	s[6] = s[2];

	str_to_key(s,deskey);
} // }}}
void sid_to_key2(unsigned long sid,unsigned char deskey[8]) // {{{ Function to convert the RID to the second decrypt key.
{
	unsigned char s[7];
	
	s[0] = (unsigned char)((sid>>24) & 0xFF);
	s[1] = (unsigned char)(sid & 0xFF);
	s[2] = (unsigned char)((sid>>8) & 0xFF);
	s[3] = (unsigned char)((sid>>16) & 0xFF);
	s[4] = s[0];
	s[5] = s[1];
	s[6] = s[2];

	str_to_key(s,deskey);
} // }}}
void E1(uchar *k, uchar *d, uchar *out) // {{{ DES encrypt, for LANMAN 
{
	des_key_schedule ks;
	des_cblock deskey;

	str_to_key(k,(uchar *)deskey);
#ifdef __FreeBSD__
	des_set_key(&deskey,ks);
#else /* __FreeBsd__ */
	des_set_key((des_cblock *)deskey,ks);
#endif /* __FreeBsd__ */
	des_ecb_encrypt((des_cblock *)d,(des_cblock *)out, ks, DES_ENCRYPT);
} // }}}

// vim600: set foldlevel=0 foldmethod=marker:
