/* {{{ preamble
 * sam.h 
 *
 * Structures and related functions for dealing with structures in the SAM hive
 * of NT registry
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#ifndef _INCLUDE_SAM_H
#define _INCLUDE_SAM_H 1


#include "../lib/libhive.h"
#include <time.h>


/* Known bits in flags field */

#define ACF_COMPLEX    0x0001  /* Pass must meet complexity req. */
#define ACF_REVERSIBLE 0x0010  /* Store password using reversible encryption */

extern char *acb_fields[];

/* This is users F value, contains account type & state etc */

#define USER_F_PATH "\\SAM\\Domains\\Account\\Users\\%08X\\F"

// Generic V Struct Stuff {{{
// Format:
// [header]
// 	[int4] always 0
// 	[int4] size of section 2
// [entry_table]
// 	[short2] ???, always seems to be 02 00 
// 	[short2] ???, always seems to be 01 00 
// 	[entry lookups]
// [section 2]
//      see below
// [entries]
//
// [entry lookup] :=
//	[int4] entry offset relative to [entry_table] + section 2 size
//	[int4] entry size
//	[int4] unknown
struct v_entry { // {{{
	int off;    //offset to entry data relative to data start + section 2 size 
	int len;
	int unknown;
}; // }}}
struct v_header { // {{{
	int u0;    //always 0?
	int s2len; //length of section 2
	short u1;  //always 2?
	short u2;  //always 2?
}; // }}}
struct v_lookup{ // {{{
	struct v_entry e[1]; //TODO: 1 is not the correct answer
}; // }}}

//Section 2: Access control (I guess) {{{
// Format:
//

struct v_perms_ent { // {{{
        short type;
        short len;
        char data[];
}; // }}}
struct v_perms_pent { // {{{
        short type;
        short len;
        char perms[4];
        char data[];
}; // }}}
struct v_perms { // {{{
        char u1[4]; //always 01 00 14 80?
        int offs1[2];
        int offs2[2];
}; // }}}
struct v_perms_3 { // {{{
        int count;
        struct v_perms_pent ents[];
}; // }}}
// }}}

struct sam_v { // {{{
	struct v_header head;
	char data[1];
	/*
	struct v_lookup lookup;
	char data[4];
	*/
}; // }}}
struct vdef { // {{{ TODO: if this is public then the name should be changed
	const int count;
	const int *types;
	const char **fields;
}; // }}}

struct v_perms_pent *sam_v_get_perm( vbuf *vb_v, int sec, int off );

typedef struct sam_v* SAM_V;
struct v_lookup *sam_v_get_lookup( struct sam_v *v );
int sam_v_get_lookup_count( struct sam_v *v );
int sam_v_dump( vbuf *v, vbuf *vb, const struct vdef *vd ); //TODO: Should this be exposed like this?
int sam_vmachine_dump( vbuf *v, vbuf *vb ); 
int sam_vuser_dump( vbuf *v, vbuf *vb ); 
struct sam_v *sam_get_V( char *path, struct varbuf *vb );
char *sam_v_get_field_ptr( struct sam_v *v, int eno, int *blen );
// }}}

struct machinev *sam_get_machine_V( struct varbuf *vb );

#if 0
struct user_F { // {{{
	char unknown1[8];
	char t_lockout[8];  /* Time of lockout */
	char unknown2[8];
	char t_creation[8]; /* Time of account creation */
	char unknown3[8];
	char t_login[8];    /* Time of last login */
	long rid;
	char unknown4[4];
	unsigned short ACB_bits;  /* Account type and status flags */
	char unknown5[6];
	unsigned short failedcnt; /* Count of failed logins, if > than policy it is locked */
	unsigned short logins;    /* Total logins since creation */
	char unknown6 [0xc];
}; // }}}
#endif

struct nt_time {
	unsigned char t[8];
};

struct uf { // {{{
	unsigned char unknown1[8];
	struct nt_time last_login;
	struct nt_time unknown2;
	struct nt_time last_pw_change;
	struct nt_time account_expires;
	struct nt_time last_invalid_pass;
	int rid;
	int unknown3;
	short acb;
	short unknown4;
	short country_code;
	short unknown5;
	short bad_login_count;
	short no_login;
	int unknown6;
	long unknown7;
}; // }}}

#define USER_F_SIZE 0x50
// {{{ ACB Fields
#define ACB_DISABLED   0x0001  /* Act disabled */
#define ACB_HOMDIRREQ  0x0002  /* Home directory required */
#define ACB_PWNOTREQ   0x0004  /* User password not req */
#define ACB_TEMPDUP    0x0008  /* Temporary duplicate account?? */
#define ACB_NORMAL     0x0010  /* Normal user account */
#define ACB_MNS        0x0020  /* MNS logon user account */
#define ACB_DOMTRUST   0x0040  /* Interdomain trust account */
#define ACB_WSTRUST    0x0080  /* Workstation trust account */

#define ACB_SVRTRUST   0x0100  /*  Server trust account */
#define ACB_PWNOEXP    0x0200  /* User password does not expire */
/* Seems not to be used on failed console logins at least */
#define ACB_AUTOLOCK   0x0400  /* Account auto locked */
// }}}
#define UV_FIELDS 17
// {{{ UV Fields
#define UV_UNKNOWN1 		0
#define UV_USERNAME 		1
#define UV_FULLNAME 		2
#define UV_COMMENT 		3
#define UV_USER_COMMENT 	4
#define UV_UNKNOWN2 		5
#define UV_HOMEDIR 		6
#define UV_HOMEDIRCONNECT 	7
#define UV_SCRIPTPATH 		8
#define UV_PROFILEPATH 		9
#define UV_WORKSTATIONS 	10
#define UV_HOURS_ALLOWED 	11
#define UV_UNKNOWN3 		12
#define UV_LMPW 		13
#define UV_NTPW 		14
#define UV_UNKNOWN4 		15
#define UV_UNKNOWN5 		16
// }}}
//

#define MV_FIELDS 4
//
//

int sam_uname2rid( char *uname ); 
const char *get_uv_name( int fno );
int get_uv_type( int fno );

struct uv_entry { // {{{
	int off;
	int len;
	int unknown;
}; // }}}
struct uv_header { // {{{
	struct uv_entry e[UV_FIELDS];
}; // }}}
struct uv { // {{{
	struct uv_header head;
	char data[4];
}; // }}}
typedef struct uv* UV;
/* Users V data struct */
/* First 0xCC bytes is pointer & len table, rest is data which
 * the table points to
 * String data is unicode, not zero terminated (must use len)
 */

#define USER_V_SIZE 0xcc

#if 0
struct user_V { // {{{

  int unknown1_1;      /* 0x00 - always zero? */
  int unknown1_2;      /* 0x04 - points to username? */
  int unknown1_3;      /* 0x08 - always 0x02 0x00 0x01 0x00 ? */

  int username_ofs;    /* 0x0c */
  int username_len;    /* 0x10 */
  int unknown2_1;      /* 0x14 - always zero? */

  int fullname_ofs;    /* 0x18 */
  int fullname_len;    /* 0x1c */
  int unknown3_1;      /* 0x20 - always zero? */

  int comment_ofs;     /* 0x24 */
  int comment_len;     /* 0x28 */
  int unknown4_1;      /* 0x2c - alway zero? */

  int unknown4_2;      /* 0x30 - points 4 or 8 byte field before hashes */
  int unknown4_3;      /* 0x34 - zero? or size? */
  int unknown4_4;      /* 0x38 - zero? */

  int unknown4_5;      /* 0x3c - to field 8 bytes before hashes */
  int unknown4_6;      /* 0x40 - zero? or size of above? */
  int unknown4_7;      /* 0x44 - zero? */

  int homedir_ofs;     /* 0x48 */
  int homedir_len;     /* 0x4c */
  int unknown5_1;      /* 0x50 - zero? */

  int drvletter_ofs;   /* 0x54 - drive letter for home dir */
  int drvletter_len;   /* 0x58 - len of above, usually 4   */
  int unknown6_1;      /* 0x5c - zero? */

  int logonscr_ofs;    /* 0x60 - users logon script path */
  int logonscr_len;    /* 0x64 - length of string */
  int unknown7_1;      /* 0x68 - zero? */

  int profilep_ofs;    /* 0x6c - profile path string */
  int profilep_len;    /* 0x70 - profile path stringlen */

  char unknown7[0x90-0x74]; /* 0x74 */

  int unknown8_1;      /* 0x90 - pointer to some place before hashes, after comments */
  int unknown8_2;      /* 0x94 - size of above? */
  int unknown8_3;      /* 0x98 - unknown? always 1? */

  int lmpw_ofs;        /* 0x9c */
  int lmpw_len;        /* 0xa0 */
  int unknown9_1;      /* 0xa4 - zero? */

  int ntpw_ofs;        /* 0xa8 */
  int ntpw_len;        /* 0xac */
  int unknowna_1;      /* 0xb0 */

  int unknowna_2;      /* 0xb4 - points to field after hashes */
  int unknowna_3;      /* 0xb8 - size of above field */
  int unknowna_4;      /* 0xbc - zero? */
  int unknowna_5;      /* 0xc0 - points to field after that */
  int unknowna_6;      /* 0xc4 - size of above */
  int unknowna_7;      /* 0xc8 - zero ? */

  char data[4];        /* Data starts here. All pointers above is relative to this,
			  that is V + 0xCC */
}; // }}}
#endif

/* This contains some policy settings for the account database */
//TODO: this has been replaced with samkeyf stuff
//#define ACCOUNTDB_F_PATH "\\SAM\\Domains\\Account\\F"
//struct accountdb_F { // {{{
//	char unknown1[8];		/* 0 */
//	char unknown2[8];		/* 8 */
//	unsigned long updatecnt;	/* 10 Number of times policy data updated */
//	char unknown3[4];		/* 14 */
//	char t_maxpwage[8];		/* 18 Maximum password age, GUI shows only days */
//	char t_minpwage[8];		/* 20 Minimum password age, GUI shows only days */
//	char unknown4[8];		/* 28 */
//	char t_lockdur[8];		/*  30 Account lockout duration, GUI shows minutes */
//	char t_lockrel[8];		/*  38 Release account lockout after, GUI show minutes */
//	char unknown5[8];		/*  40 */
//	unsigned long rid;		/*  48 RID of user doing last edit? */
//	unsigned long flags;		/* 4c Some flags & options, see below */
//	unsigned short minpwlen;	/* 50 Minimum password lenght */
//	unsigned short numhist;		/* 52 How many passwords to keep in history */
//	unsigned short locklimit;	/*54  How many tries before lockout */
//	char unknown6[0x9a];		/* Rest is unknown */
//}; // }}}                        	/* Total size 0xF0 bytes, seems to be constant */
struct samkeyf { // {{{
/* This is \SAM\Domains\Account\F */
	char unknown1[0x8];       /* 0x0000 - Unknown. May be machine SID */
	char unknown2[0x8];       /* 0x0000 - Unknown. May be machine SID */
	char autoinc[0x8];        /* auto increment (??) */
	char maxage[0x8];        /* max password age */
	char minage[0x8];       /* min password age */
	char unknown3[0x8];       
	char lockoutage[0x8];       /* account lockout duration */
	char lockoutreset[0x8];       /* account lockout duration */
	char unknown[0x8];       /* account lockout duration */
	int  next_rid;		/* the next RID that will be assigned */
	int flags;
	short min_pass;
	short enforce_password_history;
	short lockout_threshold;
	short unknown4;
	char unknown5[0x18];
	char obf[0x30];     /* 0x0070 - This may very well be the obfuscated syskey */
};    // There may be more, usually 8 null-bytes? }}}
struct secpoldata { // {{{
/* Security\Policy\SecretEncryptionKey\@, only on NT5 */
/* Probably contains some keyinfo for syskey. Second DWORD seems to be syskeymode */
	int  unknown1;             /* Some kind of flag? usually 1 */
	int  syskeymode;           /* Is this what we're looking for? */
	int  unknown2;             /* Usually 0? */
	char keydata[0x40];        /* Some kind of scrambled keydata? */
}; // }}}

char *sam_rid2uname( int rid );
int sam_print_user( vbuf *vb, int rid );
void acb_print( vbuf *out, short acb );
void acb_edit( struct uf *userf );
void promote_user(int rid);
//char *change_pw(char *buf, int rid, int vlen, int stat) // {{{}}}
//
void pass_print( vbuf *vb, int rid, struct uv *userv );

//Pull data
ulong sam_no_users();
rl_hkey sam_get_users_hkey();
ulong sam_get_max_lock();
ulong sam_get_pass_min();
ulong sam_get_pass_history();

//Pull/write structs
int sam_get_user_name( vbuf *vb, int irid ); // 
int sam_user_list( struct varbuf *vb ); // populates vb with a list of user RIDs
struct samkeyf *sam_get_samkeyf( struct varbuf *vb );
struct secpoldata *sam_get_secpoldata( struct varbuf *vb );
struct uv *sam_get_user_V( ulong rid, struct varbuf *vb );
int sam_set_user_V( ulong rid, struct varbuf *vb );
struct uf *sam_get_user_F( ulong rid, struct varbuf *vb );

//Misc
void samkeyf_dump( vbuf *vb, struct samkeyf *sf );
void secpoldata_dump( vbuf *vb, struct secpoldata *sd );
time_t nt2timet( struct nt_time *nt );

//User F
int uf_dump( vbuf *vb, struct uf *userf );

//User V
char *uv_get_field_ptr( struct uv *userv, int eno, int *blen );
int uv_dump( vbuf *vb, struct uv *userv );
//int uv_d( struct uv *userv );
int uv_check( struct uv *userv, int len ); // returns 0 if userv seems sane, -1 otherwise
struct uv_entry* uv_get_entry( struct uv *userv, int off );
int uv_get_field_type( int fno );
const char *uv_get_field_name( int fno );
int uv_set_field( struct varbuf *vb, int fno, char *buf, int blen ); //in: vb contains struct uv; out: buf data stored in field fno

//Crypto
int sam_set_lm_password( int rid, char *pass );
int sam_set_nt_password( int rid, char *pass );

int nthash_make( int rid, char *dest, char *src ); // turn a password (src) into pre-syskey nt hash (as far as we can go)
int lmhash_make( int rid, char *dest, char *src ); // turn a password (src) into pre-syskey lm hash (as far as we can go)
void rid_pw_decrypt( int rid, char *dest, char *src );
void rid_pw_encrypt( int rid, char *dest, char *src );
void make_lanmpw(char *p, char *lm, int len); // zero lanman passwd, replace with uppercase supplied password
void str_to_key(unsigned char *str,unsigned char *key); // Convert a 7 byte array into an 8 byte des key with odd parity.
void sid_to_key1(unsigned long sid,unsigned char deskey[8]); // Function to convert the RID to the first decrypt key.
void sid_to_key2(unsigned long sid,unsigned char deskey[8]); // Function to convert the RID to the second decrypt key.
void E1(u_char *k, u_char *d, u_char *out); // DES encrypt, for LANMAN 
#endif

// vim600: set foldlevel=0 foldmethod=marker:
