/* {{{ preamble
 * libhive.h
 *
 * Registry Hive access library API
 *
 * Authors:	Sean Loaring
 *
 * Copyright (c) 2005-2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#ifndef _INCLUDE_HIVELIB_H
#define _INCLUDE_HIVELIB_H 1

#include <wchar.h>
#include <stdarg.h>
#include "wintypes.h"
#include "regstructs.h"
#include "defines.h"
#include "winerr.h"
#include "regtypes.h"

#ifdef WITH_WIN32_PASSTHROUGH
#include <windows.h>
#else
#define MAX_PATH 260
#endif
// {{{ Hive type enum defines
//#define HKEY_CLASSES_ROOT       ((HKEY) 0x80000000)
//#define HKEY_CURRENT_USER       ((HKEY) 0x80000001)
//#define HKEY_LOCAL_MACHINE      ((HKEY) 0x80000002)
//#define HKEY_USERS              ((HKEY) 0x80000003)
//#define HKEY_PERFORMANCE_DATA   ((HKEY) 0x80000004)
//#define HKEY_CURRENT_CONFIG     ((HKEY) 0x80000005)
//#define HKEY_DYN_DATA           ((HKEY) 0x80000006)
// }}}
//

// {{{ Typedefs & Structs

#ifndef KEY_ALL_ACCESS
#define KEY_ALL_ACCESS 0xF003F
#endif

#ifndef KEY_CREATE_LINK
#define KEY_CREATE_LINK	0x0020
#endif

#ifndef KEY_CREATE_SUB_KEY
#define KEY_CREATE_SUB_KEY 0x0004
#endif

#ifndef KEY_ENUMERATE_SUB_KEYS
#define KEY_ENUMERATE_SUB_KEYS 0x0008
#endif

#ifndef KEY_EXECUTE
#define KEY_EXECUTE 0x20019
#endif

#ifndef KEY_NOTIFY
#define KEY_NOTIFY 0x0010
#endif

#ifndef KEY_QUERY_VALUE
#define KEY_QUERY_VALUE 0x0001
#endif

#ifndef KEY_READ
#define KEY_READ 0x20019
#endif

#ifndef KEY_SET_VALUE
#define KEY_SET_VALUE 0x0002
#endif

#ifndef KEY_WOW64_32KEY
#define KEY_WOW64_32KEY 0x0200
#endif

#ifndef KEY_WRITE
#define KEY_WRITE 0x20006
#endif


#ifndef WITH_WIN32_PASSTHROUGH
typedef struct value_entW { //{{{
	wchar_t       *ve_valuename;
	unsigned long ve_valuelen;
	unsigned long *ve_valueptr;
	unsigned long ve_type;
} VALENTW, *PVALENTW; // }}}

typedef unsigned long REGSAM;
typedef void SECURITY_ATTRIBUTES;
typedef long FILETIME; //long is just a guess, I'm ignoring time for now...

//#define KEY_ALL_ACCESS 1

#endif

/*
#define NK_NORMAL 0
#define NK_META   1

#define META_NORMAL 0
#define META_ROOT   1
#define META_NATIVE 2
*/

struct nk_desc {
	int type;
	char *path; //only used in meta
	int id; //only used in meta TODO: or not at all
	int flag;
#ifdef WITH_WIN32_PASSTHROUGH
	HKEY native;
#endif
	struct hive *hd;
	struct nk_key *nk;
};

typedef struct nk_desc *rl_hkey;

// }}}
//// {{{ REG_* type defines
//#define REG_NONE                0       /* no type */
//#define REG_SZ                  1       /* string type (ASCII) */
//#define REG_EXPAND_SZ           2       /* string, includes %ENVVAR% (expanded by caller) (ASCII) */
//#define REG_BINARY              3       /* binary format, callerspecific */
///* YES, REG_DWORD == REG_DWORD_LITTLE_ENDIAN */
//#define REG_DWORD               4       /* DWORD in little endian format */
//#define REG_DWORD_LITTLE_ENDIAN 4       /* DWORD in little endian format */
//#define REG_DWORD_BIG_ENDIAN    5       /* DWORD in big endian format  */
//#define REG_LINK                6       /* symbolic link (UNICODE) */
//#define REG_MULTI_SZ            7       /* multiple strings, delimited by \0, terminated by \0\0 (ASCII) */
//#define REG_RESOURCE_LIST       8       /* resource list? huh? */
//#define REG_FULL_RESOURCE_DESCRIPTOR    9       /* full resource descriptor? huh? */
//#define REG_RESOURCE_REQUIREMENTS_LIST  10
//#define REG_QWORD               11      /* QWORD in little endian format */
//#define REG_QWORD_LITTLE_ENDIAN 11      /* QWORD in little endian format */
//// }}}

extern rl_hkey rlHKEY_CLASSES_ROOT;
extern rl_hkey rlHKEY_CURRENT_USER;
extern rl_hkey rlHKEY_LOCAL_MACHINE;
extern rl_hkey rlHKEY_USERS;
extern rl_hkey rlHKEY_PERFORMANCE_DATA;
extern rl_hkey rlHKEY_CURRENT_CONFIG;
extern rl_hkey rlHKEY_DYN_DATA;

void libhive_init(); // You have to call this first

//void normalize_path( char *p ); //in: p = "/a/broken\\\\path///with\\\\stuff"; out: p = "\\a\\broken\\path\\with\\stuff"

#if 0 // {{{ this stuff now lives in winerr.h
// {{{ windows-style errors (this is idiotic)
#define rlERROR_SUCCESS       		0
#define rlERROR_MORE_DATA     		1
#define rlERROR_NO_MORE_ITEMS 		2
#define rlERROR_BROKEN_ERROR  		3 //KLUDGE: this is for all of the undocumented crap that might get returned  
#define rlERROR_NOT_FOUND		4
#define rlERROR_OPEN_FAILED		5
// }}}

#define MAX_WIN_ERROR			6


//this is for horrible FormatMessage()
#define rlFORMAT_MESSAGE_ALLOCATE_BUFFER 	0x00000100
#define rlFORMAT_MESSAGE_FROM_SYSTEM		0x00001000

extern char *win_errstr[];

#endif // }}}

// {{{
//typedef unsigned long DWORD;
//typedef const char *LPCVOID;
//typedef char *LPTSTR;
// }}}
/* IN USE */

struct valh { // {{{
	rl_hkey hk;
	char *name;
	long type;
	char *buf;
	long blen;
}; // }}}

char *get_type_str( ulong type );

/* Helper */
struct valh *vh_open( rl_hkey hk, char *path );
int vh_close( struct valh *vh );
int vh_drop( struct valh *vh ); // Don't commit anything, just free()
int vh_set_buf( struct valh *vh, void *buf, long len ); // resize vh buffer and copy buf into it

ulong vh_get_dword( struct valh *vh );
int vh_set_dword( struct valh *vh, ulong val );

/* This shouldn't be here, it just walks past the whole API */
int hkey_get_path(rl_hkey hk, char *path, int maxlen); 

void hl_close_all();
long load_hive( const char *fname, int mode );
int hl_unload_hive( int hno );
void flush_hives();

char *hl_mount_help();
int hl_mount_list( int argc, char **argv );
int hl_hive_add( struct hive *h, int htype );
struct hive *hl_hive_remove( int hno );
int hl_mount_count();
int hl_mount_get_point( vbuf *vb, int no );
rl_hkey hl_mount_get_hkey( int no );

void rlhkey_dump(rl_hkey hkey);

long rlRegOpenHiveFile(rl_hkey *result, const char *fname, const char *keypath, int mode );
long rlRegOpenKeyEx(rl_hkey hkey, const char *skname, ulong options, REGSAM, rl_hkey *result);
long rlRegQueryValueEx(rl_hkey key, const char *vname, unsigned long *type,
		char *buf, unsigned long *blen);

long rlRegEnumKeyEx(rl_hkey, unsigned long index, char *name, unsigned long *len, char *cname, 
                                   unsigned long *cnlen, FILETIME *ft);

long rlRegEnumValue(rl_hkey key, unsigned long index, char *vname, unsigned
		long *vnlen,  unsigned long *type, char *buf, unsigned long
		*blen);

/* CURRENT */
long rlRegSetValueEx( rl_hkey key, const char *vname, ulong reserved, ulong type, const char* buf, ulong blen);

//This works with rlFORMAT_MESSAGE_FROM_SYSTEM and rlFORMAT_MESSAGE_ALLOCATE_BUFFER, otherwise it doesn't.
//DWORD rlFormatMessage( DWORD dwFlags, LPCVOID lpSource, DWORD dwMessageId, DWORD dwLanguageId, LPTSTR lpBuffer, DWORD nSize, va_list* Arguments);


/* ELSE */
// {{{ REGSAM crap (which is all ignored for now)

// }}}

long rlRegCreateKeyEx(rl_hkey hkey, const char *skname, const char *r_class, unsigned long options, REGSAM desired, 
                                     SECURITY_ATTRIBUTES *security, rl_hkey *result, unsigned long *disposition);

//long rlRegOpenUserClassesRoot(HANDLE, unsigned long, REGSAM, rl_hkey *result);
long rlRegQueryInfoKey(rl_hkey key, char *clname, unsigned long *cllen,
		unsigned long *subkeys, unsigned long *maxsklen, unsigned long
		*maxcllen, unsigned long *nvalues, unsigned long
		*maxvalnamelen, unsigned long *maxvallen, unsigned long
		*securitydesc, FILETIME *lastwrite);

long rlRegQueryMultipleValues(rl_hkey, VALENTW *val_list, unsigned long nvals,
		char *buf, unsigned long *totsize);

/* Declarations for functions that are the same in Win16 and Win32 */

long rlRegCloseKey(rl_hkey key);
long rlRegFlushKey(rl_hkey key);

long rlRegDeleteKey(rl_hkey key, const char *skname);
long rlRegDeleteValue(rl_hkey key, const char *vname);


//These depend on some sort of registry patch files (same as output from regedit?) {{{

//long rlRegReplaceKey(rl_hkey, const wchar_t *, const wchar_t *, const wchar_t *);
//long rlRegRestoreKey(rl_hkey, const wchar_t *, unsigned long);
//long rlRegLoadKey(rl_hkey, const wchar_t *, const wchar_t *);
//long rlRegSaveKey(rl_hkey, const wchar_t *, LPSECURITY_ATTRIBUTES);

//}}}

//I'll do these later (maybe) {{{

//long rlRegGetKeySecurity(rl_hkey, SECURITY_INFORMATION, PSECURITY_DESCRIPTOR, LPunsigned long);
//long rlRegUnLoadKey(rl_hkey, const wchar_t *);
//long rlRegSetKeySecurity(rl_hkey, SECURITY_INFORMATION, PSECURITY_DESCRIPTOR);
//long rlRegConnectRegistry(const wchar_t *, rl_hkey, PHKEY);
//long rlRegOpenCurrentUser(REGSAM, PHKEY);
// }}}

//these functions have Ex counterparts.  I'm going to ignore these for now... {{{

//long rlRegEnumKey(rl_hkey, unsigned long, wchar_t *, unsigned long);
//long rlRegCreateKey(rl_hkey key, const wchar_t *, PHKEY);
//long rlRegOpenKey(rl_hkey key, const wchar_t *skname, rl_hkey *result);
//long rlRegQueryValue(rl_hkey, const wchar_t *, wchar_t *, long *);
//long rlRegSetValue(rl_hkey, const wchar_t *, unsigned long, const wchar_t *, unsigned long);
//}}}
// }}}

#endif

// vim600: set foldlevel=0 foldmethod=marker:
