/* {{{ preamble
 * ntreg.h
 *
 * NT Registry Hive access library, constants & structures
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#ifndef _INCLUDE_NTREG_H
#define _INCLUDE_NTREG_H 1

#include <stdio.h>
#include "../misc/tools.h"

#include "defines.h"
#include "regstructs.h"

#define ABSPATHLEN 2048

#define HBB_AVAIL_MAX  0x1000

/* {{{ Datatypes of the values in the registry */

#define REG_NONE                    0  /* No value type */
#define REG_SZ                      1  /* Unicode nul terminated string */
#define REG_EXPAND_SZ               2  /* Unicode nul terminated string + env */
#define REG_BINARY                  3  /* Free form binary */
#define REG_DWORD                   4  /* 32-bit number */
#define REG_DWORD_BIG_ENDIAN        5  /* 32-bit number */
#define REG_LINK                    6  /* Symbolic Link (unicode) */
#define REG_MULTI_SZ                7  /* Multiple Unicode strings */
#define REG_RESOURCE_LIST           8  /* Resource list in the resource map */
#define REG_FULL_RESOURCE_DESCRIPTOR 9 /* Resource list in the hardware description */
#define REG_RESOURCE_REQUIREMENTS_LIST 10

#define REG_MAX 10
// }}}
// {{{ typedefs
typedef long heapoff; //offset within the heap of the hive file
typedef long fileoff; //offset within the hive file
// }}}
/* {{{ Structure defines for my routines */

struct ex_data {
	int nkoffs;
	struct nk_key *nk;
	char *name;
};

struct vex_data {
	int vkoffs;
	struct vk_key *vk;
	int type;       /* Value type REG_??? */
	int size;       /* Values size (normalized, inline accounted for) */
	void *val;        /* WAS: Actual value itself if type==REG_DWORD, is now pointer */
	int dwbuf;
	char *name;
};

struct keyval {
	int len;      /* Length of databuffer */
	int data;    /* Data. Goes on for length of value */
};

struct keyvala {
	int len;      /* Length of databuffer */
	int data[1];    /* Data. Goes on for length of value */
};
// }}}
/* {{{ Suggested type of hive loaded, guessed by library, but not used by it */
#define HTYPE_UNKNOWN   0
#define HTYPE_SAM       1
#define HTYPE_SYSTEM    2
#define HTYPE_SECURITY  3
#define HTYPE_SOFTWARE  4
#define HTYPE_USER	5
#define HTYPE_USERDIFF  6
#define HTYPE_USRCLASS	7
#define HTYPE_DEFAULT	8

#define MAX_HTYPE 9

/* These MUST be in the same order as HTYPE_*, they are accessed thusly: hivepath[ HTYPE_TYPE ] */
extern char *hivepath[];
// }}}
/* {{{ Hive descriptor structs, allocated by openHive(), dealloc by closeHive()
 * contains state data, must be passed in all functions
 */
struct hive {
	char *filename;		/* Hives filename */
	int  filedesc;		/* File descriptor (only valid if state == OPEN) */
	int  state;		/* Current state of hive */
	int  type;		/* Suggested type of hive. NOTE: Library will
				   guess when it loads it, but application may
				   change it if needed */
	//int  pages;		/* Number of pages, total */
	//int  useblk;		/* Total # of used blocks */
	//int  unuseblk;		/* Total # of unused blocks */
	//int  usetot;		/* total # of bytes in useblk */
	//int  unusetot;		/* total # of bytes in unuseblk */
	int  size;		/* Hives size (filesise) in bytes */
	//  int  rootofs;	/* Offset of root-node */
	struct nk_key *root_key;
	struct hbin_page **hbb_avail; /* quick 'n dirty free hbinb lookup table */
	char *buf;		/* Files raw contents */
	struct hbin_block *root_block;
	int alloc_max;		// If alloc(size) fails then alloc_max = size, hbinb_free() clears
	struct regf_header *head;     /* regf Header */
	char *heap;		/* offset of data past header */
	int refcount;		/* incremented on each open, decremented on close */
};
// }}}
//
extern char *val_types[];

/* {{{ **** Function prototypes */

char *str_dup( const char *str );
int fmyinput(char *prmpt, char *ibuf, int maxlen);
//void hexprnt(char *s, unsigned char *bytes, int len);
//void hexdump(const char *hbuf, int start, int stop, int ascii);
//void winhex( FILE *stream, char *hbuf, int start, int stop );
int find_in_buf(char *buf, char *what, int sz, int len, int start);
int get_int( char *array );
void cheap_uni2ascii(char *src, char *dest, int l);
void cheap_ascii2uni(char *src, char *dest, int l);
void skipspace(char **c);
int gethex(char **c);
int gethexorstr(char **c, char *wb);
//int debugit(char *buf, int sz);
char *regf_get_name( struct regf_header *head ); 
int parse_block(struct hive *hdesc, int vofs,int verbose);
int parse_nk_nice( struct hive *hdesc, struct nk_key *nk );
int ex_next_n(struct hive *hdesc, fileoff nkofs, int *count, int *countri, struct ex_data *sptr);
int ex_next_v(struct hive *hdesc, int nkofs, int *count, struct vex_data *sptr);
int get_abs_path(struct hive *hdesc, int nkofs, char *path, int maxlen);
struct hbin_block *trav_path(struct hive *hdesc, int vofs, char *path, int type);
struct nk_key *nk_trav_path(struct hive *hdesc, struct nk_key *nk, const char *path);
struct vk_key *vk_trav_path(struct hive *hdesc, struct nk_key *nk, const char *path);
int get_val_type(struct hive *hdesc, int vofs, char *path);
int get_val_len(struct hive *hdesc, int vofs, char *path);

//void *get_key_data( struct hive *hdesc, struct vk_key *vkkey );
void *get_val_data(struct hive *hdesc, int vofs, char *path, int val_type);

struct keyval *get_val2buf(struct hive *hdesc, struct keyval *kv,
			   int vofs, char *path, int type );
int get_dword(struct hive *hdesc, int vofs, char *path);
int put_buf2val(struct hive *hdesc, struct keyval *kv,
		int vofs, char *path, int type );
int put_dword(struct hive *hdesc, int vofs, char *path, int dword);
void closeHive(struct hive *hdesc);
int writeHive(struct hive *hdesc);
struct hive *openHive(const char *filename, int mode);

int nk_scan( struct hive *hdesc, struct nk_key *nk );

void nk_ls(struct hive *hdesc, char *path, int vofs, int type);
//int nk_delete_value(struct hive *hdesc, struct nk_key *nk, const char *name);
int nk_delete_value(struct hive *hdesc, struct nk_key *nk, int val_no);
struct vk_key *add_value(struct hive *hdesc, int nkofs, char *name, int type);
void del_allvalues(struct hive *hdesc, int nkofs);
int del_value(struct hive *hdesc, int nkofs, char *name);
struct nk_key *add_key(struct hive *hdesc, int nkofs, char *name);
//int del_key(struct hive *hdesc, int nkofs, char *name);
long get_file_offset( struct hive *hdesc, void *p );
long get_heap_offset( struct hive *hdesc, void *p );

void *get_ptr_off_f( struct hive *hdesc, long off );
void *get_ptr_off_h( struct hive *hdesc, long off );

int vk_get_name_len( struct vk_key *vk ); // return the length of the vk name were it encoded in UTF8

//NK methods
//
int nk_get_path(struct hive *hdesc, struct nk_key *nk, char *path, int maxlen); // {{{ traceback - trace nk's back to root, APPENDS TO *path!!!
struct nk_key* check_nk_ptr( struct hive *hdesc, void *p );
struct nk_key* nk_get_subkey( struct hive *hdesc, struct nk_key *nk, int off );
struct vk_key* nk_get_value( struct hive *hdesc, struct nk_key *nk, int off );
int nk_find_value( struct hive *hdesc, struct nk_key *nk, const char *name );
int nk_find_subkey( struct hive *hdesc, struct nk_key *nk, const char *name );
int nk_delete_subkey( struct hive *hdesc,  struct nk_key *parent, const char *name );
//int nk_get_name( struct nk_key *nk, char *buf, int n);
int nk_vb_get_name( struct nk_key *nk, vbuf *out );
int nk_get_classname( struct hive *hdesc, struct nk_key *nk, char *buf, int n);
long nk_get_class_len( struct nk_key *nk );
long nk_get_name_len( struct nk_key *nk );
long nk_get_no_subkeys( struct nk_key *nk );
struct nk_key *nk_create( struct hive *hdesc, const char *name, struct nk_key *nkparent );
long nk_seek_vk( struct hive *hdesc, struct nk_key *nk, const char *name );
int nk_set_value(struct hive *hdesc, struct nk_key *nk, const char *name, int type, const char *buf, ulong blen );



//VK methods
struct vk_key* check_vk_ptr( struct hive *hdesc, void *p );
int vk_get_nameT( struct vk_key *vk, char *buf, int n ); //old
void vb_vk_get_name( vbuf *vb, struct vk_key *vk ); // Places name of vk in vb in utf8
void *vk_get_val( struct hive *hdesc, struct vk_key *vk );
int vk_get_val2( struct hive *hdesc, struct vk_key *vk, vbuf *vb );

int vk_get_length( struct vk_key *vk );
int vk_get_length2( struct hive *hdesc, struct vk_key *vk );

ulong vk_get_type( struct hive *hdesc, struct vk_key *vk );

struct nk_key* get_root_key( struct hive *hdesc );

struct hbin_page* check_hbin_page( struct hive *hdesc, void *p );

struct hbin_page* hive_get_first_hbinp( struct hive *hdesc );
struct hbin_block* hbinp_first_hbinb( struct hbin_page *p );
struct hbin_page* hbinp_get_next_hbinp( struct hive *hdesc, struct hbin_page *p );
struct hbin_block* hbinb_get_next( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b );

void hbinb_vdump( vbuf *vb, struct hive *hdesc, struct hbin_block *b );
void hbinbh_vdump( vbuf *vb, struct hive *hdesc, struct hbb_head *hbbh );

long hbinb_size( struct hbin_block *p );

void hive_scan( struct hive *hdesc );

char *id_to_str( unsigned long id );
// }}}
#endif


// vim600: set foldlevel=0 foldmethod=marker:
