/* {{{ preamble
 * regstructs.c
 *
 * Functions to process primative registry data and .reg file contents.
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "regstructs.h"
#include "regtypes.h"
#include "libhive.h"
#include "../misc/uerr.h"
#include "../misc/tools.h"
#include "../misc/globals.h"

char typebuf[TYPE_BUF_LEN];

char *get_hextype( int type );

void print_val( char *name_in, unsigned long name_in_len, unsigned long type, char *val, long val_len, char *path ); 
int wr_fprint_val( FILE* stream, void *data, int type, int len, int loff );

void vb_escape_evil( vbuf *vb, unsigned char *string ) // {{{ escape things like '\' and '"'
{
	int i=0;

	vbuf_strset( vb, "" );

	for( i=0; i < strlen( string ); i++ )
	{
		switch ( string[i] ) {
			case '\\':
			case '"':
				vbuf_charcat( vb, '\\' );
				vbuf_charcat( vb, string[i] );
				break;
			case '\r':
				vbuf_charcat( vb, '\\' );
				vbuf_charcat( vb, 'r' );
				break;
			case '\n':
				vbuf_charcat( vb, '\\' );
				vbuf_charcat( vb, 'n' );
				break;
			default:
				vbuf_charcat( vb, string[i] );
				break;
		}

		/* I don't know what this is about, winreg outputs a space after a 0x19 char */
		if( 0x19 == string[i] ) {
			//memmove( string + i + 2, string + i + 1, strlen( string + i + 1) + 1 );

			vbuf_charcat( vb, ' ' );
			//string[i + 1] = ' ';
			//i++;
		}
	}

	//fprintf(stderr, "escape_evil() out: \"%s\"\n", string );
} // }}}
void escape_evil( unsigned char *string, int length ) // {{{ escape things like '\' and '"'
{
	int i=0;

	//fprintf(stderr, "escape_evil() in: \"%s\"\n", string );
	for( i=0; i <= strlen( string ); i++ )
	{
		if(
			'\\' == string[i] ||
			'\"' == string[i] 
		) {
			memmove( string + i + 1, string + i, strlen( string + i ) + 1 );

			string[i] = '\\';
			i++;
		}

		/*
		if( ! isalpha( string[i] )  )
			printf("> %c == %x == %d < ", string[i], string[i], string[i] );
			*/


		/* I don't know what this is about, winreg outputs a space after a 0x19 char */
		if( 0x19 == string[i] ) {
			memmove( string + i + 2, string + i + 1, strlen( string + i + 1) + 1 );

			string[i + 1] = ' ';
			i++;
		}
	}

	//fprintf(stderr, "escape_evil() out: \"%s\"\n", string );
}
// }}}

//char *pv_string_buf = NULL;
//int pv_buf_len=0;
//int winout = 1;
#define PV_INIT_SIZE 100
struct varbuf *pv_buf = NULL;
/*
void conv_sz_utf8( struct varbuf *b, void *data, int len ) // {{{ Convert a SZ string to normal utf8, dumps in b
	//TODO: This is kindof duplicated in vb_utf16_to_8 EXCEPT escape_evil() isn't called from other
{
	//cheap_uni2ascii(data, string, len);
	DEBUG(
		fprintf(stderr, "fprint_val@REG_SZ: len=%d\n", len );
		hexdump( data, 0, len, 1 );
	     );
	if( 0 == len ) {
		vbuf_set( b, "\"\"", 3 );
		return;
	}

	if( b->blen < len ) {
		fprintf(stderr, "fprint_val(): buffer too small, calling realloc for size %d\n", len );
		vbuf_resize( b, len );
	}

	utf16to8( data, b->b, len );
	escape_evil( b->b, b->blen );

	b->dlen = b->blen;

	return;
	//utf16_fprintf( stderr, "\"%x\"%s", string, stupid_cr );
} // }}}
*/

void fprint_val_nice( FILE* stream, void *data, int type, int len, int loff ) /* {{{ print a value to a stream - doesn't conform to anything specific */
{
	VBUF_STATIC( buf8, 100 );

	//TODO: this looks dumpy
	if( pv_buf == NULL ) pv_buf = vbuf_alloc( PV_INIT_SIZE );

	switch( type ) {
		case REG_DWORD:
			fflush( stdout );
			fprintf(stream, "%08x",*(unsigned int *)data);
			break;

		case REG_EXPAND_SZ:
		//case REG_BINARY:

		case REG_SZ:
			//conv_sz_utf8( pv_buf, data, len );
			/*
			if( 0 == vb_utf16to8( buf8, data, len ) ) {
				fprintf( stream, "\"%s\"", buf8->b );
				break;
			} else {
				WARN("vb_utf16to8(): %s", uerr_str( uerr_get() ) );
				hexdump( data, 0, len, 1 );
			}
			*/

			fprintf( stream, "\"%s\"", data );
			break;

		//anything we don't know we print in hex (regedit does)
		default:
			hexprnt( stdout, (char*)data, len );
			break;

	}
}
// }}}
void fprint_val( FILE* stream, void *data, int type, int len, int loff ) /* {{{ print a value to a stream - conforms to regedit output*/
{
	DIE("Here.");
	fprint_val8( stream, data, type, len, loff );
	/*
	VBUF_STATIC( buf8, 100 );

	//TODO: this looks dumpy
	if( pv_buf == NULL ) pv_buf = vbuf_alloc( PV_INIT_SIZE );

	switch( type ) {
		case REG_DWORD:
			fflush( stdout );
			fprintf(stream, "%08x",*(unsigned int *)data);
			break;

		//case REG_EXPAND_SZ:
		//case REG_BINARY:

		case REG_SZ:
			//conv_sz_utf8( pv_buf, data, len );
			if( 0 == vb_utf16to8( buf8, data, len ) ) {
				fprintf( stream, "\"%s\"", buf8->b );
				break;
			} else {
				WARN("vb_utf16to8(): %s", uerr_str( uerr_get() ) );
				hexdump( data, 0, len, 1 );
			}

		//anything we don't know we print in hex (regedit does)
		default:
			hexprnt( stdout, (char*)data, len );
			break;
	}
	*/
}
// }}}
/*
int print_typed_winhex( FILE *stream, int type, char *data, int length ) // {{{
{
	int loff = 0;
	loff += utf16_fprintf( stream, get_hextype( type ) );
	winhex( stream, (char *)data, 0, length, loff);
	return loff;
} // }}}
*/
int unbreak_sz( vbuf *out, char *in, int len ) // {{{ IN: utf16 is valid | not-zero-terminated | with trailing garbage, OUT: valid utf16 (we hope)
{ //TODO: This should probably be in regtypes
        int i;
        for( i=0; i<=len-2; i+= 2) { //look for alligned zero-termination
                if( '\0' == in[i] && '\0' == in[i+1] ) {
                        if( len == i+2 ) { // It's a good string
                                vbuf_set( out, in, len );
                        } else {
                                vbuf_set( out, in, i+2 );
                        }
                        return 0;
                }
        }

        //didn't find zero-termination

        if( len % 2 == 0 ) {
                vbuf_set( out, in, len );
                return 0;
        }

        // Odd number of bytes in the input data and no zero termination, don't
        // know how to make UTF-16 out of this
        return -1;
} // }}}
int vb_print_sz( vbuf *out, char *data, int len ) // {{{ Convert a SZ value as regedit does using data right from the registry
{
	VBUF_STATIC( buf8, 10 );
	VBUF_STATIC( sanesz, 10 );
	VBUF_STATIC( eevil, 10 );
	int res;

	if( 0 == len ) {
		vbuf_strset( out, "\"\"");
		return 0;
	}

	vbuf_set( sanesz, data, len );

	/*
	if( 0 != unbreak_sz( sanesz, data, len ) ) { // || -1 == vb_utf16to8( buf8, sanesz->b, sanesz->dlen) ) 
		warn("wtf?");
		return -1;
	}
	*/


	vb_escape_evil( eevil, sanesz->b );
	//vb_escape_evil( eevil, buf8->b );
	//utf16_fprintf( stream, "\"%s\"", eevil->b );

	vbuf_printf( out, "\"%s\"", eevil->b );
	return 0;

} // }}}
int wr_fprint_val( FILE* stream, void *data, int type, int len, int loff ) /* {{{ print a value to a stream*/
{
	VBUF_STATIC( buf, 10 );
	VBUF_STATIC( utf16buf, 10 );
	vbprint_val8( buf, data, type, len, loff );
	if( -1 == vb_utf8to16( utf16buf, buf->b, buf->dlen ) ) return -1;
	return fwrite( utf16buf->b, utf16buf->dlen, 1, stream );
	//utf16_fprintf( stream, "%s", buf->b );

	/*
	//int i=0;

	VBUF_STATIC( buf16, 100 );
	VBUF_STATIC( bufsz, 10 );
	if( pv_buf == NULL ) pv_buf = vbuf_alloc( PV_INIT_SIZE );

	switch( type ) {
		case REG_SZ:
			if( 0 != vb_print_sz( bufsz, data, len ) )
				utf16_fprintf( stream, "\"%s\"", bufsz->b );
			else //fallback on hex output
				print_typed_winhex( stream, type, data, len );

			break;
		case REG_DWORD:
			fflush( stdout );
			utf16_fprintf(stream, "dword:%08x",*(unsigned int *)data);
			break;

		//anything we don't know we print in hex (regedit does)
		default:
			print_typed_winhex( stream, type, data, len );
			break;

	}


	utf16_fprintf( stream, "%s", stupid_cr );

	return 0;
	*/
}
// }}}
void fprint_val8( FILE* stream, void *data, int type, int len, int loff ) /* {{{ print a value to a stream (UTF8)*/
{
	VBUF_STATIC( buf, 10 );
	DIE("Here.");
	vbprint_val8( buf, data, type, len, loff );
	fprintf( stream, "%s", buf->b );
	/*
	VBUF_STATIC( bufsz, 10 );

	static struct varbuf *vb = NULL;
	if( vb == NULL ) vb = vbuf_alloc( 10 );

	vbuf_resize( vb, len );

	switch( type ) {
		case REG_SZ:
			if( 0 != vb_print_sz( bufsz, data, len ) )
				fprintf( stream, "\"%s\"", bufsz->b );
			else //fallback on hex output
				print_typed_winhex( stream, type, data, len );

			break;
		case REG_DWORD:
			fflush( stdout );
			fprintf(stream, "dword:%08x",*(unsigned int *)data);
			break;

		//anything we don't know we print in hex (regedit does)
		default:
			loff += fprintf( stream, get_wintype( type ) );
			winhex8( stream, (char *)data, 0, len, loff);
			break;
	}


	fprintf( stream, "%s", stupid_cr );
	*/
} // }}}

void vbprint_val8( vbuf *out, void *data, int type, int len, int loff ) /* {{{ print a value to a vbuf in UTF8*/
{
	//int i=0;

	/*
	static struct varbuf *vb = NULL;
	if( vb == NULL ) vb = vbuf_alloc( 10 );

	vbuf_resize( vb, len );
	*/
	vbuf_strset( out, "" );

	switch( type ) {
		case REG_SZ:
			if( 0 != vb_print_sz( out, data, len ) ) {
				WARN("SZ FAIL. %d", len);
				hexdump( data, 0, len, 1 );
				loff += vbuf_catprintf( out, get_wintype( type ) );
				vbwinhex8( out, (char *)data, 0, len, loff);
			}

			break;
			/*
			conv_sz_utf8( vb, data, len );
			vbuf_catprintf( out, "\"%s\"", vb->b );
			*/

		case REG_DWORD:
			vbuf_catprintf( out, "dword:%08x",*(unsigned int *)data);
			break;

		//anything we don't know we print in hex (regedit does)
		default:
			loff += vbuf_catprintf( out, get_wintype( type ) );
			vbwinhex8( out, (char *)data, 0, len, loff);
			break;
	}

	vbuf_catprintf( out, "%s", stupid_cr );
} // }}}
/*
void print_vk( struct hive *hdesc, char *path, struct vk_key *vk ) // {{{
{
	char *tstr;
	char name[SZ_MAX];
	int i;
	int loff=0;
//  tstr = (vex->type < REG_MAX)? val_types[vex->type] : "(unknown)";
	void *val = vk_get_val( hdesc, vk );

	vk_get_name( vk, name, SZ_MAX );

	if( gverbose ) {
		fprintf(stderr, "print_vk() starting with %s\n", path );
		fflush( stderr );
	}

	//print value name,type,size
	if( winout ) {
		escape_evil( name, SZ_MAX );
		if( 0 == strcmp( name, "" ) )
			loff+=utf16_fprintf(stdout, "@=");
		else
			loff+=utf16_fprintf(stdout,"\"%s\"=", name);


		fprint_val( stdout, val, vk->val_type, vk_get_length( vk ), loff );
	}
	else
	{
		printf("V %s\\%s\t0x%lX:%s\t0x%lX", path, name, vk->val_type, tstr, vk->len_data );

		if( vk_get_val( hdesc, vk ) == NULL )
		{
			fprintf(stderr, "NULL POINTER ERROR (%s) !!!!\n", name);
			exit(1);
		}
		else
		{
			//hex dump value data
			printf("\t0x");
			for( i=0; i<vk->len_data; i+=2) { printf("%X", (int)*((char*)val + i) ); }

			//print value data
			printf("\t");
			fprint_val( stdout, val, vk->val_type, vk->len_data, 0 );

		}

		printf("\n");
	}

}
// }}}
*/
/*
void wr_print_val( char *name_in, unsigned long name_in_len, unsigned long type, char *val, long val_len, char *path ) // {{{ Replacement with new API
{
	char *tstr;
	int i;
	int loff=0;
	char *name=NULL;
	unsigned long name_len=name_in_len * 2 + 1; //worst case we escape everything 

	if( gverbose ) fprintf(stderr, "print_val() starting with %s\n", path );

	name = (char*)F_MALLOC( name_len ); 

	strcpy( name, name_in );

	//print value name,type,size
	if( winout ) {
		if( 0 == strcmp( name, "" ) )
			loff+=utf16_fprintf(stdout, "@=");
		else {
			escape_evil( name, name_len );
			loff+=utf16_fprintf(stdout,"\"%s\"=", name);
		}

		wr_fprint_val( stdout, val, type, val_len, loff );
	}
	else {
		printf("V %s\\%s\t0x%lX:%s\t0x%lX", path, name, type, tstr, val_len );

		//hex dump value data
		printf("\t0x");
		for( i=0; val_len; i+=2) { printf("%X", (int)*((char*)val + i) ); }

		//print value data
		printf("\t");
		//fprint_val( stdout, val, type, val_len, 0 );

		printf("\n");
	}

} // }}}
*/
char *get_hextype( int type ) // {{{
{
	snprintf( typebuf, TYPE_BUF_LEN - 1, "hex(%x):", type );  
	return typebuf;
}
char *get_wintype( int type ) // {{{
{
	switch( type )
	{
		case REG_SZ: return "";
		case REG_DWORD: return "dword:";
		case REG_BINARY: return "hex:";
		default:
				 return get_hextype( type );
		return typebuf;
		//      case REG_EXPAND_SZ: return "hex(2):";
		//      case REG_MULTI_SZ:
		//      default:
		//      {
		//	fprintf(stderr, "Don't know how to handle type %d!\n", type);
		//	
		//	exit(1);
		//      }
	}
}
// }}}

char *valstr[] = {
	 "NONE", 
	 "SZ",
	 "EXPAND_SZ",
	 "BINARY",
	 "DWORD",
	 "DWORD_BIG_ENDIAN",
	 "LINK",
	 "MULTI_SZ",
	 "RESOURCE_LIST",
	 "FULL_RESOURCE_DESCRIPTOR",
	 "RESOURCE_REQUIREMENTS_LIST",
	 "QWORD",
	 ""
};

int str2valtype( char *type ) // {{{ "BINARY" -> 3 (REG_BINARY)
{
	int i;
	for( i=0; valstr[i][0] != '\0'; i++ ) {
		if( 0 == strcmp( type, valstr[i] ) ) return i;
	}

	return -1;
} // }}}

char *val_type_str( int type ) // {{{ returns "BINARY", "SZ" and the like
{
	ASSERT( type < REG_NO_TYPES, "val_type_str(): type out of bounds");
	return valstr[type];

} // }}}
char *val_type_desc( int type ) // {{{ returns "DWORD in little endian format, etc"
{
	switch( type ) {
		case REG_NONE: return "No associated type (data entered in hex)";
		case REG_SZ: return "String";
		case REG_EXPAND_SZ: return "A string with %ENVVAR%";
		case REG_BINARY: return "Binary";
		case REG_DWORD: return "little endian dword";
		case REG_DWORD_BIG_ENDIAN: return "big endian dword";
		case REG_LINK: return "Symbolic link stored in UTF16";
		case REG_MULTI_SZ: return "Multiple strings";
		case REG_RESOURCE_LIST: return "Unknown";
		case REG_FULL_RESOURCE_DESCRIPTOR: return "Unknown";
		case REG_RESOURCE_REQUIREMENTS_LIST: return "Unknown";
		case REG_QWORD: return "QWORD in little endian format";
		default: return "Unknown type.";
	}

} // }}}
int val_encode_user( char *udat, int type, struct varbuf *vbuf ) // {{{ IN: udat = ascii user input; OUT: vbuf
{
	switch( type ) {
		case REG_NONE: 
		case REG_SZ: 
		case REG_EXPAND_SZ: 
		case REG_BINARY: 
		case REG_DWORD: 
		case REG_DWORD_BIG_ENDIAN: 
		case REG_LINK: 
		case REG_MULTI_SZ: 
		case REG_RESOURCE_LIST: 
		case REG_FULL_RESOURCE_DESCRIPTOR: 
		case REG_RESOURCE_REQUIREMENTS_LIST: 
		case REG_QWORD: 
			WARN("val_encode_user(): I don't know how to encode that type yet.");
		default: 
			DIE("val_encode_user(): got unknown type.");
	}

	return -1;
} // }}}

int read_hex_int( char *b, int len, unsigned int *result ) // {{{ in: 0 .. ffff; out: *result = value; returns: 0 on okay, -1 on error
{
	ASSERT( len > 0, "read_hex_int(): len is not > 0" );
	int i;
	int v = 0;
	*result = 0;
	for( i=0; i<len; i++ ) {
		*result = *result << 4;
		*result += v;

		v = 0;
		if( islower( b[i] ) ) {
			v = b[i] - 'a' + 10;
			continue;
		}

		if( isupper( b[i] ) ) {
			v = b[i] - 'A' + 10;
			continue;
		}

		if( isdigit( b[i] ) ) {
			v = b[i] - '0';
			continue;
		}

		ERR_UNIX( EINVAL, "read_hex_int(): passed a bad character.");
		return -1;
	}

	*result = *result << 4;
	*result += v;

	return 0;
} // }}}
int regp_val_type( int *type, int *format, vbuf *line ) // {{{
{
	//int slen = strlen( str );
	if( line->b[0] == '"' ) {
		*type = REG_SZ;
		*format = REG_SZ;
		return 0;
	}

	if( 0 == strncmp( line->b, "hex", 3 ) ) {
		vbuf_skip(line,3);
		//int off = 3;
		char *eb;
		*format = REG_BINARY;
		if( line->b[0] == '(' ) {
			eb = index( line->b, ')' );
			if( NULL == eb ) {
				ERR_UNIX( EINVAL, "val_read_type_str(): unmatched opening parenthesis.");
				return -1;
			}

			if( -1 == read_hex_int( line->b + 1, eb - line->b - 1, type ) ) 
				return -1;

			vbuf_skip(line, eb - line->b + 2 );
			return 0;
			//return eb + 2;
		}
		else {
			*type = REG_BINARY;
			vbuf_skip( line, 1 );
			//return str + off + 1;
			return 0;
		}
	}

	if( 0 == strncmp( line->b, "dword:", 6 ) ) {
		*type = REG_DWORD;
		*format = REG_DWORD;
		//return str + 6;
		vbuf_skip( line, 6 );
		return 0;
	}

	ERR_UNIX( EINVAL, "regp_val_type(): unknown type string %s", line->b );
	return -1;
	//DIE("val_read_type_str(): unknown type: %s.", str);
} // }}}
int parse_winhex_data( struct varbuf *vb, char *str, int maxlen ) // {{{
{
	char *p = str;

	int v;

	vbuf_clear( vb );

	while( '\0' != *p && ( maxlen < 0 || p - str < maxlen ) ) {
		if( isxdigit( *p ) ) {
			if( isxdigit( *(p+1) ) ) {
				if( -1 == read_hex_int( p, 2, &v ) ) return -1;
				vbuf_append( vb, &v, 1 );
			}
			else {
				printf("parse_winhex_data(): found single hex character (they must come in 2s)");
				return -1;
			}
			p+=2;
			continue;
		}

		if( isspace( *p ) || *p == ',' ) {
			p++;
			continue;
		}

		//TODO: error
		printf("parse_winhex_data(): parse error: %s\n", str );
		return -1;

	}

	return 0;
} // }}}
int parse_quoted_string( struct varbuf *vb, char *p, int len ) // {{{ in: p = "\"Hi\tthere\"", out: vb->b = "HiTABthere" (TAB is a tab)
{
	//static struct varbuf *ucvb = NULL; //place to dump horrible utf16
	//if( NULL == ucvb ) ucvb = vbuf_alloc( 0 );

	ASSERT( p[0] == '\"' && p[len-1] == '\"', "parse_quoted_string(): I only deal with quoted strings." );

	char* p2 = p + 1;
	int len2 = len - 2;

	if( -1 == parse_escaped_string( vb, p2, len2 ) ) return -1;

	//fprintf(stderr, "parse_quoted_string: %s -> %s\n", p, vb->b );

	if( strstr( vb->b, "ImageFolder\"=" ) ) {
		DIE("Closer still.");
	}

	//if( 0 != vb_utf8to16( vb->b, vb->dlen, ucvb ) ) return -1;

	//vbuf_overwrite( vb, ucvb );

	return 0;
} // }}}

struct regfile *regfile_open( char *path ) // {{{ Open a .reg file for reading
{
	struct regfile *result = F_MALLOC( sizeof( struct regfile ) );

	result->got_header = 0;
	result->ln = 0;

	if( NULL == (result->ur = ugly_open( path ) ) ) {
		//TODO: error
		free( result );
		return NULL;
	}

	result->cl = vbuf_alloc(10);

	return result;
} // }}}
void regfile_close( struct regfile *rf ) // {{{ Close a regfile
{
	ugly_close( rf->ur );
	vbuf_free( rf->cl );
	free( rf );
} // }}}
int regfile_eof( struct regfile *rf ) // {{{
{
	if( ugly_eof( rf->ur ) && ( rf->cl->dlen == 0 || rf->cl->b[0] == '\0') ) return 1;
	return 0;
} // }}}
int regfile_readrec( struct varbuf *rec, struct regfile *rf ) // {{{
{
	static struct varbuf *line = NULL;
	if( NULL == line ) line = vbuf_alloc(10);

	//vbuf_resize( rec, 5000000 );
	vbuf_strset( rec, "" );

	while(! regfile_eof( rf ) ) {
		if( 0 != ugly_readline( line, rf->ur ) ) return -1;

		//printf("PreA:\n");
		//hexdump( rec->b, 0, rec->dlen, 1 );

		//printf("PreB:\n");
		//hexdump( line->b, 0, line->dlen, 1 );

		vbuf_strncat( rec, line->b, line->dlen );

		//printf("Post:\n");
		//hexdump( rec->b, 0, rec->dlen, 1 );

		if( rec->dlen >=2 && rec->b[rec->dlen-2] == '\\' ) {
			if( 0 == line->dlen && ugly_eof( rf->ur ) ) {
				//TODO: error
				printf("regfile_readrec(): Escaped newline at end of input.\n");
				return -1;
			}
			rec->b[rec->dlen-2] = '\0';
			rec->dlen--;
		}
		else {
			break;
		}
	} 

	vbuf_skipws( rec );

	return 0;
} // }}}
int regfile_next( struct varbuf *name, struct varbuf *val, int *type, struct regfile *rf ) // {{{ populates name,val,type and returns RTK_KEY, RTK_COM, RTK_VAL
{
	/* [regfile] := [header][reglines]
	 * [header] := see reg_header()
	 * [reglines] := [keyline]|[valline]|[comment]
	 *
	 * [record] := [keyline][ncomment] || [valline][ncomment] || [comment]
	 *
	 * [ncomment] := null || '; stuff'
	 */

	int tokno = -1;
	//static struct varbuf *rec = NULL;
	//if( NULL == rec ) rec = vbuf_alloc(10);

	do {
		if( regfile_eof( rf ) && tokno == RTK_EMPTY ) {
			return RTK_EOF;
		}
		if( ! rf->got_header ) { 
			if( 0 != regfile_readrec( rf->cl, rf ) ) return -1;
			if( 0 != regp_header( rf->cl ) ) return -1;

			rf->got_header = 1;
		}

		//vbuf_clear( rf->cl );

		if( 0 != regfile_readrec( rf->cl, rf ) ) return -1;
		tokno = regp_tok( name, val, type, rf->cl );
	} while ( tokno == RTK_COMMENT || tokno == RTK_EMPTY);

	return tokno;
} // }}}

int regp_tok( vbuf *name, vbuf *val, int *type, vbuf *line ) // {{{ Tokenize a .reg file line
{
	ASSERT( line->dlen > 0, "regp_tok() got empty line.");
	switch( line->b[0] ) {
		case '\0':
			//EMPTY
			return RTK_EMPTY;
		case '[':
			//KEY NAME
			if( 0 != regp_key_line( name, line ) ) {
				//TODO: error
				printf("Error parsing key line: %s\n", line->b );
				return -1;
			}
			return RTK_KEYNAME;

		case ';':
			//COMMENT
			regp_ncomment( line );
			return RTK_COMMENT;

		default:
			//VAL LINE
			if( 0 != regp_val_line( val, name, type, line ) ) {
				return -1;
			}
			return RTK_VAL;
	}

	DIE("This should not execute.");
	return -1;
} // }}}
int regp_ncomment( struct varbuf *vb ) // {{{ Make sure that s contains either nothing or a valid comment
{
	char *s = vb->b;
	while( s[0] && isspace( s[0] ) ) s++;
	if( ';' == s[0] || '\0' == s[0] ) {
		//found a comment or EOL
		vbuf_clear( vb );
		return 0;
	}
	//TODO: error
	DIE("Here we are.\n");
	printf("regp_ncomment: this -->%s<-- must be either nothing or a comment.\n", s);
	return -1;
} // }}}
int regp_key_line( vbuf *name, vbuf *line ) // {{{ Decodes a .reg file key line
{
	char *idx;
	if( line->b[0] != '[' ) {
		//TODO: error
		printf("Invalid key line, no opening square parenthesis.\n");
		return -1;
	}
	if( NULL == ( idx = index(line->b, ']') ) ) {
		//TODO: error
		printf("Invalid key line, no closing square parenthesis.\n");
		return -1;
	}

	vbuf_strnset( name, line->b + 1, idx - line->b - 1 );
	//name->b[name->dlen] = '\0';

	vbuf_skip( line, (idx+1) - line->b );

	//if( 0 != regp_ncomment( idx + 1 ) ) return -1;

	return 0;
} // }}}
int regp_val_line( struct varbuf *val, struct varbuf* name, int *type, struct varbuf *line ) // {{{ Decodes .reg file value lines
{
	/* [val_line] := [val_name]=[data_line]
	 *
	 * [val_name] := @ || [esc_str]
	 *
	 * Examples:
	 *
	 * @="Default Beep"
	 * @=hex(2):25,00,ff
	 * "C:\\backslash\\delimiters\\are\\retarded"=dword:00000001
	 */

	switch( line->b[0] ) {
		case '@':
			vbuf_strset( name, "@" );
			vbuf_skip( line, 2 );
			if( 0 != regp_val_data( val, type, line ) ) return -1;
			break;
		case '"':
			if( 0 != regp_quoted_string( name, line ) ) return -1;
			vbuf_skipws(line);
			if( line->b[0] != '=' ) {
				ERR_UNIX( EINVAL, "regp_val_line(): expecting an \"=\", got %s", line->b );
				return -1;
			}
			vbuf_skip(line,1);
			if( 0 != regp_val_data( val, type, line ) ) return -1;
			break;
		default:
			ERR_UNIX( EINVAL, "regp_val_line(): Parse error, unexpected character at start of value line %s", line->b );
			return -1;
	}

	return 0;
} // }}}
int regp_udata( struct varbuf *vb, int format, struct varbuf *line ) // {{{ same as regp_data, but strings are not quoted
{
	static struct varbuf *ucvb = NULL; //place to dump horrible utf16
	if( NULL == ucvb ) ucvb = vbuf_alloc( 0 );
	vbuf_skipws( line );
	//int slen = line->dlen - 1;

	//ASSERT( slen >= 2, "regp_data(): string length < 2 and min string is two double-quotes.");

	switch( format ) {
		case REG_SZ:

			if( 0 != parse_escaped_string( vb, line->b, line->dlen - 1 ) ) return -1;
			//if( 0 != regp_quoted_string( vb, line ) ) return -1;

			//if( 0 != vb_utf8to16( ucvb, vb->b, vb->dlen ) ) return -1;

			//vbuf_overwrite( vb, vb );

			break;
		case REG_BINARY:
			if( 0 != regp_winhex( vb, line ) ) return -1;
			break;
		case REG_DWORD:
			if( 0 != regp_dword( vb, line ) ) return -1;

			break;

		default:
			DIE("regp_data(): unknown type");
	}

	return 0;
} // }}}
int regp_data( struct varbuf *vb, int format, struct varbuf *line ) // {{{ convert from extended regedit data format to raw, registry binary form
{
	static struct varbuf *ucvb = NULL; //place to dump horrible utf16
	if( NULL == ucvb ) ucvb = vbuf_alloc( 0 );
	vbuf_skipws( line );
	//int slen = line->dlen - 1;

	//ASSERT( slen >= 2, "regp_data(): string length < 2 and min string is two double-quotes.");

	switch( format ) {
		case REG_SZ:

			if( 0 != regp_quoted_string( vb, line ) ) return -1;

//			if( 0 != vb_utf8to16( ucvb, vb->b, vb->dlen ) ) return -1;
//			vbuf_overwrite( vb, ucvb );

			break;
		case REG_BINARY:
			if( 0 != regp_winhex( vb, line ) ) return -1;
			break;
		case REG_DWORD:
			if( 0 != regp_dword( vb, line ) ) return -1;

			break;

		default:
			DIE("regp_data(): unknown type");
	}

	return 0;
} // }}}
int regp_dword( struct varbuf *out, struct varbuf *line ) // {{{
{
	long dwval;
	if( line->dlen < 8 ) {
		printf("regp_dword(): length is less than 8 characters.\n");
		return -1;
	}
	if( -1 == parse_winhex_data( out, line->b, 8 ) ) return -1;
	vbuf_skip( line, 8 );

	regp_ncomment( line );

	dwval = out->b[0];
	out->b[0] = out->b[3];
	out->b[3] = dwval;

	dwval = out->b[1];
	out->b[1] = out->b[2];
	out->b[2] = dwval;

	return 0;
} // }}}
int regp_winhex( struct varbuf *out, struct varbuf *line ) // {{{
{
	char *le = index( line->b, ';' );
	int l = (NULL!=le)?(le - line->b):line->dlen-1;

	if( -1 == parse_winhex_data( out, line->b, l ) ) return -1;

	vbuf_skip( line, l );

	return 0;
} // }}}
int regp_quoted_string( struct varbuf *out, struct varbuf *line ) // {{{
{
	char *p = line->b;
	int endl;
	if( *p != '"' ) {
		ERR_UNIX( EINVAL, "regp_quoted_string(): expected '\"' at: <%s>", p );
		return -1;
	}
	p++;
	p = esc_index( p, '"' );
	/*
	do {
		p = index( p+1, '"' );
	} while(
			NULL != p &&
			(
				(p-1)[0] == '\\' &&
				!( p-2 >= line->b && (p-2)[0] == '\\')
			)
	); //TODO: this is so ugly
	*/

	if( NULL == p ) {
		ERR_UNIX( EINVAL, "regp_quoted_string(): Unterminated double-quote string: %s", line->b );
		DIE("splat");
		return -1;
	}
	endl = p - line->b;
	parse_quoted_string( out, line->b, endl + 1 );

	if( strstr( out->b, "ImageFolder\"=" ) ) {
		DIE("First bad one.");
	}
	//fprintf(stderr, "regp_quoted_string() %s -> %s\n", line->b, out->b );
	vbuf_skip( line, endl + 1 );
	return 0;
} // }}}
int regp_header( struct varbuf *line ) // {{{ returns 0 if p is a valid .reg header
{
	//TODO: I think that there are more valid headers
	char *s1 = "Windows Registry Editor Version 5.00";

	if( 0 == strcmp( line->b, s1 ) ) {
		vbuf_skip( line, strlen( s1 ) );
		if( 0 != regp_ncomment( line ) ) return -1;
		return 0;
	}

	ERR_UNIX( EINVAL, "Unrecognized registry header: %s", line->b );
	return -1;
} // }}}
int regp_val_data( struct varbuf *vb, int *type, vbuf *line ) // {{{ in: str := [data_line] ( see below ); out: data in var_buf, type in *type
{
	/*
	 * [data_line] := 	\"[string_data]\" |
	 * 			dword:[dword_data] |
	 * 			hex[hex_type]:[hex_data]
	 *
	 * [string_data] := 	TODO: define this.  Mostly normal strings but they can have escapes in them
	 *
	 * [dword_data] :=	[octet][octet][octet][octet]
	 *
	 * [hex_type] :=	null || ([plain_hex])
	 *
	 * [hex_data] :=	[hex_data_element][hex_data] || null
	 * [hex_data_element] := [octet] || , || [white_space] 
	 *
	 * [octet] :=		[hex_digit][hex_digit]
	 * [hex_digit] :=	0-9a-fA-F
	 * 			
	 */

	int format;
	//int slen = strlen( str );

	//char *p = val_read_type_str( line->b, type, &format );

	//if( NULL == p ) return -1;

	if( 0 != regp_val_type( type, &format, line ) ) return -1;
	if( 0 != regp_data( vb, format, line ) ) return -1;

	return 0;
} // }}}
int val_can_encode( int type ) // {{{ OUT: 1 == can encode type, 0 == can't
{
	switch( type ) {
		case REG_NONE: 
		case REG_SZ: 
		case REG_EXPAND_SZ: 
		case REG_BINARY: 
		case REG_DWORD: 
		case REG_DWORD_BIG_ENDIAN: 
		case REG_LINK: 
		case REG_MULTI_SZ: 
		case REG_RESOURCE_LIST: 
		case REG_FULL_RESOURCE_DESCRIPTOR: 
		case REG_RESOURCE_REQUIREMENTS_LIST: 
		case REG_QWORD: 
			return 0;
		default: 
			DIE("val_can_encode(): got unknown type.");
	}

	DIE("HOW DID WE GET HERE?");
	return -1;
} // }}}
int val_get_type( char *tstr ) // {{{ IN: tstr == "REG_SZ" || "SZ", etc; OUT: REG_SZ || -1
{
	int i;
	char *s = tstr;

	if( 0 == strncmp( s, "REG_", 4 ) ) {
		s = s + 4;
	}

	for( i=0; i<REG_NO_TYPES; i++ ) {
		if( 0 == strcmp( s, valstr[i] ) ) return i;
	}

	return -1;
} // }}}
int vb_dword2int( vbuf *b, int *out ) // {{{
{
	if( b->dlen != 4 ) {
		ERR_UNIX(EINVAL, "dword length seems incorrect, 4 != %d\n", b->dlen );
		return -1;
	}

	*out = dword2int( b->b );

	return 0;
} // }}}
int dword2int( void *p ) // {{{
{
	return *(int*)p;
} // }}}
