/* {{{ preamble
 * regstructs.h
 *
 * Functions to process primative registry data and .reg file contents.
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#ifndef REGTYPES_H
#define REGTYPES_H 1
#include <stdio.h>
//#include "regstructs.h"
#include "libhive.h"
#include "../misc/tools.h"

// {{{ REG_* type defines
#define REG_NONE                0       /* no type */
#define REG_SZ                  1       /* string type (ASCII) */
#define REG_EXPAND_SZ           2       /* string, includes %ENVVAR% (expanded by caller) (ASCII) */
#define REG_BINARY              3       /* binary format, callerspecific */
/* YES, REG_DWORD == REG_DWORD_LITTLE_ENDIAN */
#define REG_DWORD               4       /* DWORD in little endian format */
//#define REG_DWORD_LITTLE_ENDIAN 4       /* DWORD in little endian format */
#define REG_DWORD_BIG_ENDIAN    5       /* DWORD in big endian format  */
#define REG_LINK                6       /* symbolic link (UNICODE) */
#define REG_MULTI_SZ            7       /* multiple strings, delimited by \0, terminated by \0\0 (ASCII) */
#define REG_RESOURCE_LIST       8       /* resource list? huh? */
#define REG_FULL_RESOURCE_DESCRIPTOR    9       /* full resource descriptor? huh? */
#define REG_RESOURCE_REQUIREMENTS_LIST  10
#define REG_QWORD               11      /* QWORD in little endian format */
//#define REG_QWORD_LITTLE_ENDIAN 11      /* QWORD in little endian format */

#define REG_NO_TYPES	12
// }}}

//cruft
#define TYPE_BUF_LEN 30
#define SCAN_BUF 1024

/* TODO: This should be elsewhere and shouldn't be so ugly
 *
 * I thought that writing a recursive-descent-type grammar for .reg files would
 * be relatively easy.  It seems to look like roadkill.  Bad roadkill.
 *
 * [regfile] := [header][record]...[record]
 * [header] := [header_str][ncomment]
 * [record] := [keyrec] | [valrec]
 * [keyrec] := '['[str]']'[ncomment]
 * [valrec] := [valname]'='[valdata]
 * [valname] := '@' | [escstr]
 * [valdata] := [valstr] | [valdword] | [valhex]
 * [valstr] := [escstr][ncomment]
 * [valdword] := "dword:\d{8}"[ncomment]
 * [valhex] := [hexmess][ncomment]
 * [hexmess] := 'xx,xx,xx'...sortof
 */

struct regfile {
	struct uglyread *ur;
	int ln; //line number
	int got_header; //has the header been read yet?
	struct varbuf *cl; //the current line
};

#define RTK_EMPTY 1
#define RTK_KEYNAME 2
#define RTK_VAL 3
#define RTK_COMMENT 4
#define RTK_EOF 5


//TODO: below are in the include for testing purposes only
int val_encode_generic( struct varbuf *vb, int *type, char *str );
int reg_val_line( struct varbuf *val, struct varbuf* name, int *type, char *str ); 
char *val_read_type_str( char *str, int *type, int *format );

struct regfile *regfile_open( char *path );
void regfile_close( struct regfile *rf );
int regfile_next( struct varbuf *name, struct varbuf *val, int *type, struct regfile *rf );

int regp_tok( vbuf *name, vbuf *val, int *type, vbuf *line);
int regp_key_line( vbuf *name, vbuf *str ); 
int regp_val_line( struct varbuf *val, struct varbuf* name, int *type, vbuf *line ); 
int regp_val_data( struct varbuf *vb, int *type, vbuf *line ); 
int regp_quoted_string( struct varbuf *out, struct varbuf *line );
int regp_header( struct varbuf *line ); 
int regp_ncomment( vbuf *s );
int regp_udata( struct varbuf *vb, int format, struct varbuf *line );
int regp_data( struct varbuf *vb, int format, struct varbuf *line );
int regp_winhex( struct varbuf *out, struct varbuf *line );
int regp_dword( struct varbuf *out, struct varbuf *line );
int regp_val_type( int *type, int *format, vbuf *line ); 

void fprint_val( FILE* stream, void *data, int type, int len, int loff ); /* print a value to a stream*/
void wr_print_val( char *name_in, unsigned long name_in_len, unsigned long type, char *val, long val_len, char *path ); 
int wr_fprint_val( FILE* stream, void *data, int type, int len, int loff );
void fprint_val8( FILE* stream, void *data, int type, int len, int loff );
void vbprint_val8( vbuf *vb, void *data, int type, int len, int loff );
void escape_evil( unsigned char *string, int length ); 

char *get_wintype( int type );

char *val_type_str( int type ); // returns "BINARY", "SZ" and the like
char *val_type_desc( int type ); // returns "DWORD in little endian format, etc"
int val_encode_user( char *udat, int type, struct varbuf *vbuf ); //IN: udat = ascii user input; OUT: vbuf
int val_can_encode( int type ); //OUT: 1 == can encode type, 0 == can't
int val_get_type( char *tstr ); //IN: tstr == "REG_SZ" || "SZ", etc; OUT: REG_SZ || -1
int read_hex_int( char *b, int len, unsigned int *result ); 

int str2valtype( char *type ); // "BINARY" -> 3 (REG_BINARY)
int dword2int( void *p ); // converts a dword stored at p to an int
int vb_dword2int( vbuf *b, int *out ); 
#endif


// vim600: set foldlevel=0 foldmethod=marker:
