/* {{{ preamble
 * winerr.h
 *
 * Emulates only the most absolutely necessary parts of the win32 error system.
 *
 * Authors:	Sean Loaring
 *
 * Copyright (c) 2005-2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

#ifndef WINERR_H
#define WINERR_H

#include "wintypes.h"
#include <stdarg.h>
#include "../misc/uerr.h"

#ifdef WITH_WIN32_PASSTHROUGH
#include <windows.h>
#endif


// {{{ windows-style (idiotic) errors 
#ifdef WITH_WIN32_PASSTHROUGH

#define rlFORMAT_MESSAGE_ALLOCATE_BUFFER        FORMAT_MESSAGE_ALLOCATE_BUFFER
#define rlFORMAT_MESSAGE_FROM_SYSTEM            FORMAT_MESSAGE_FROM_SYSTEM

#define rlERROR_SUCCESS       		ERROR_SUCCESS
#define rlERROR_MORE_DATA     		ERROR_MORE_DATA
#define rlERROR_NO_MORE_ITEMS 		ERROR_NO_MORE_ITEMS
#define rlERROR_BROKEN_ERROR  		3 //KLUDGE: this is for all of the undocumented crap that might get returned  
#define rlERROR_NOT_FOUND		ERROR_NOT_FOUND
#define rlERROR_OPEN_FAILED		ERROR_OPEN_FAILED

#define rlLocalFree(x)			LocalFree(x)

#else

//TODO: this stuff should have the same values as whatever windows uses

#define rlFORMAT_MESSAGE_ALLOCATE_BUFFER        0x00000100
#define rlFORMAT_MESSAGE_FROM_SYSTEM            0x00001000

#define rlERROR_SUCCESS       		0
#define rlERROR_MORE_DATA     		1
#define rlERROR_NO_MORE_ITEMS 		2
#define rlERROR_BROKEN_ERROR  		3 //KLUDGE: this is for all of the undocumented crap that might get returned  
#define rlERROR_NOT_FOUND		4
#define rlERROR_OPEN_FAILED		5

#define rlLocalFree(x)			free(x)

#endif
// }}}

#define MAX_WIN_ERROR			6

#define ERR_WIN( myerrno, ... ) eprintf( UERR_WIN, myerrno, winstrerror( myerrno ), __VA_ARGS__ );

char *winstrerror( DWORD myerrno );
DWORD rlFormatMessage( DWORD dwFlags, LPCVOID lpSource, DWORD dwMessageId, DWORD dwLanguageId, LPTSTR lpBuffer, DWORD nSize, va_list* Arguments);

#endif

// vim600: set foldlevel=0 foldmethod=marker:
