#include <stdlib.h>
#include <check.h>
#include <string.h>
#include "../hivetools/nstdreg.h"

START_TEST( test_path_stuff ) // {{{
{
	vbuf *v = vbuf_alloc(0);

	//ns_vdirname // {{{
	ns_vdirname(v,"");
	fail_unless( 0 == strcmp( ".", v->b ), "invalid.");

	ns_vdirname(v,"foo");
	fail_unless( 0 == strcmp( ".", v->b ), "invalid.");

	ns_vdirname(v,"foo/");
	fail_unless( 0 == strcmp( ".", v->b ), "invalid.");

	ns_vdirname(v,"foo/foo");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vdirname(v,"foo/foo/");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vdirname(v,"/foo/");
	fail_unless( 0 == strcmp( "/", v->b ), "invalid.");

	ns_vdirname(v,"/foo/foo");
	printf("What is this: %s\n", v->b );
	fail_unless( 0 == strcmp( "/foo", v->b ), "invalid.");

	ns_vdirname(v,"/foo/foo/");
	fail_unless( 0 == strcmp( "/foo", v->b ), "invalid.");

	ns_vdirname(v,"/HKEY_LOCAL_MACHINE/software/\\0");
	fail_unless( 0 == strcmp( "/HKEY_LOCAL_MACHINE/software", v->b ), "invalid.");

	ns_vdirname(v,"/\\//foo/");
	fail_unless( 0 == strcmp( "/\\/", v->b ), "invalid.");
	// }}}

	//ns_vbasename {{{
	ns_vbasename(v,"");
	printf("wtf: <%s>\n", v->b);
	fail_unless( 0 == strcmp( "", v->b ), "invalid.");

	ns_vbasename(v,"foo");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"foo/");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"foo//");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"foo//////");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"foo/bar");
	fail_unless( 0 == strcmp( "bar", v->b ), "invalid.");

	ns_vbasename(v,"foo/bar/");
	fail_unless( 0 == strcmp( "bar", v->b ), "invalid.");

	ns_vbasename(v,"foo/bar//");
	fail_unless( 0 == strcmp( "bar", v->b ), "invalid.");

	ns_vbasename(v,"/foo/");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"/foo/foo");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"/foo/foo/");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	ns_vbasename(v,"/\\/\\//");
	fail_unless( 0 == strcmp( "\\/\\/", v->b ), "invalid.");

	ns_vbasename(v,"/\\/foo/");
	fail_unless( 0 == strcmp( "\\/foo", v->b ), "invalid.");


	ns_vbasename(v,"/\\\\/foo/");
	fail_unless( 0 == strcmp( "foo", v->b ), "invalid.");

	// }}}
	vbuf_strset(v, "");
	vbuf_charcat(v, 'x');
	fail_unless( 0 == strcmp(v->b, "x"));
	// ns_path_element_reg2ns && ns_path_ns2reg {{{
	fail_unless( 0 == ns_path_element_reg2ns( v, "." ) );
	fail_unless( 0 == strcmp( v->b, "...") );
	fail_unless( 0 == ns_path_ns2regt( v, "...", -1 ) );
	fail_unless( 0 == strcmp( v->b, ".") );

	fail_unless( 0 == ns_path_element_reg2ns( v, ".." ) );
	fail_unless( 0 == strcmp( v->b, "....") );
	fail_unless( 0 == ns_path_ns2regt( v, "....", -1 ) );
	fail_unless( 0 == strcmp( v->b, "..") );

	fail_unless( 0 == ns_path_element_reg2ns( v, "..." ) );
	fail_unless( 0 == strcmp( v->b, ".....") );
	fail_unless( 0 == ns_path_ns2regt( v, ".....", -1) );
	fail_unless( 0 == strcmp( v->b, "...") );

	fail_unless( 0 == ns_path_element_reg2ns( v, "" ) );
	fail_unless( 0 == strcmp( v->b, "\\0") );
	fail_unless( 0 == ns_path_ns2regt( v, "\\0", RTYPE_VALUE) );
	fail_unless( 0 == strcmp( v->b, "") );

	fail_unless( 0 == ns_path_element_reg2ns( v, "\\" ) );
	fail_unless( 0 == strcmp( v->b, "\\\\") );
	fail_unless( 0 == ns_path_element_reg2ns( v, "\\/\\" ) );
	fail_unless( 0 == strcmp( v->b, "\\\\\\/\\\\") );

	fail_unless( 0 != ns_path_ns2regkeypath( v, "\\\\") );
	fail_unless( 0 == ns_path_ns2regvalname( v, "\\\\") );
	fail_unless( 0 == ns_path_ns2regvalname( v, "a\\\\") );
	fail_unless( 0 == strcmp( v->b, "a\\") );

	fail_unless( 0 == ns_path_ns2regvalname( v, "a\\\\b") );
	fail_unless( 0 == strcmp( v->b, "a\\b") );
	// }}}

	// {{{ vb_process_path
	vb_process_path( v, "");
	fail_unless( 0 == strcmp( v->b, "") );

	vb_process_path( v, "/");
	fail_unless( 0 == strcmp( v->b, "/") );

	vb_process_path( v, "//");
	fail_unless( 0 == strcmp( v->b, "/") );

	vb_process_path( v, "/foo/..");
	fail_unless( 0 == strcmp( v->b, "/") );

	vb_process_path( v, "/foo/foo/..");
	WARN("I got this: %s", v->b );
	fail_unless( 0 == strcmp( v->b, "/foo/") );

	vb_process_path( v, "/foo/foo/foo/foo/foo" );
	fail_unless( 0 == strcmp( v->b, "/foo/foo/foo/foo/foo") );

	vb_process_path( v, "/././././." );
	printf("S: <%s>\n", v->b );
	fail_unless( 0 == strcmp( v->b, "/") );
	//fail_unless( 0 == strcmp( v->b, "/.") );

	vb_process_path( v, ".");
	fail_unless( 0 == strcmp( v->b, ".") );

	vb_process_path( v, "..");
	fail_unless( 0 == strcmp( v->b, "..") );

	vb_process_path( v, "/../../..");
	fail_unless( 0 == strcmp( v->b, "/") );

	vb_process_path( v, "./../../..");
	fail_unless( 0 == strcmp( v->b, "./../../..") );

	vb_process_path( v, "/meow\\//bark");
	fail_unless( 0 == strcmp( v->b, "/meow\\//bark") );

	vb_process_path( v, "meow\\//bark");
	WARN("ThisX: >%s<", v->b );
	fail_unless( 0 == strcmp( v->b, "meow\\//bark") );

	vb_process_path( v, "/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/\\0" );
	WARN("This: %s", v->b );
	fail_unless( 0 == strcmp( v->b,  "/HKEY_LOCAL_MACHINE/SAM/SAM/Domains/\\0") );

	//vb_process_path( v, "/meow\\//");
	//fail_unless( 0 == strcmp( v->b, "/meow\\//") );

	//empty should not be allowed
	//fail_unless( 0 != ns_path_regkey2ns( v, "" ) );

	// }}}
	fail_unless( 0 != ns_path_regkey2ns( v, "\\HKEY_LOCAL_MACHINE\\foo") );
	fail_unless( 0 == ns_path_regkey2ns( v, "HKEY_LOCAL_MACHINE\\foo") );
	fail_unless( 0 == strcmp( v->b, "/HKEY_LOCAL_MACHINE/foo") );

	fail_unless( 0 == ns_path_regkey2ns( v, "HKEY_LOCAL_MACHINE\\fo/o") );
	fail_unless( 0 == strcmp( v->b, "/HKEY_LOCAL_MACHINE/fo\\/o") );

	fail_unless( path_end_is_delim( "/" ) );
	fail_unless( path_end_is_delim( "blar/" ) );

	fail_unless( !path_end_is_delim( "foo" ) );
	fail_unless( !path_end_is_delim( "\\/" ) );
	fail_unless( !path_end_is_delim( "blar\\/" ) );
} 
END_TEST // }}}

Suite* regtypes_suite( void ) // {{{
{
	Suite *s = suite_create("Tools");
	TCase *tc_core = tcase_create("Core");

	suite_add_tcase (s, tc_core);
	tcase_add_test(tc_core, test_path_stuff );

	return s;
} // }}}

int main(void) // {{{
{
	int nf;
	Suite *s = regtypes_suite();
	SRunner *sr = srunner_create(s);
	srunner_set_fork_status( sr, CK_NOFORK );
	srunner_run_all(sr, CK_NORMAL);
	nf = srunner_ntests_failed(sr);
	srunner_free(sr);
	return (nf == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
} // }}}
